#!/usr/bin/perl

########################################################################
# install.pl 2.9
#   install script for metaf2xml
#
# copyright (c) 2011-2019 metaf2xml
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
########################################################################

########################################################################
# some things strictly Perl
########################################################################
use warnings;
use 5.014_004;

use Getopt::Std;
use Pod::Usage;

use File::Path qw(mkpath);
use File::Copy qw(copy);

=encoding utf8

=head1 NAME

install.pl - install metaf2xml from the sources

=head1 SYNOPSIS

 install.pl [OPTION]... {all|parser|xsl|ui} [<CONFIG>=<VALUE> ...]

=head1 DESCRIPTION

This script installs metaf2xml from the sources.

=over

=item Note: Alternatives to manual installation

=over

=item Use installable packages

Various installable packages are available from the
L<File Release System|https://sourceforge.net/projects/metaf2xml/files/2.9/>
at SourceForge.

=item Use without installation

metaf2xml can be used without any installation. Just extract one of the
source archives or check out the files from the Git repository as
described in step 2 below.

The Perl scripts of metaf2xml can be used from the command line without
modification. The Perl modules can also be used by other Perl scripts;
just use S<C<use lib ...>> in the scripts, or include the path to the
module directory in the environment variable C<PERL5LIB>.
To use C<metaf.pl> as a web interface without installation, configure the
web server to map the URL of the CGI script (e.g. C</cgi-bin/metaf.pl>)
and the XSL files (e.g. C</metaf.xsl>, C</metaf-lang.xml> etc.) to the
location of the files.

=back

=back

=head2 Installation steps

=over 2

=item 1. Check that you have Perl and the required Perl modules

Perl version 5.14.4 or higher is required for the installation and at runtime.
To check, execute S<C<perl -v>>.
If this does not execute the Perl binary you want to use, please use the full
path to the correct Perl binary (e.g. F</usr/local/bin/perl>) instead of
just F<perl> for this check and also use it as value for C<PERL> (see
L<Miscellaneous settings|/Miscellaneous settings for the installation>).

There are the following additional dependencies:

=over

=item Perl module C<CGI>

required by C<metaf.pl>

=item Perl modules C<DBI> and C<DBD>, programs C<curl> and C<xsltproc>

optionally required by C<metaf.pl>

=item Perl module C<Geo::BUFR>

optionally required by C<metaf2xml::src2raw>

=back

=item 2. Get metaf2xml sources

The preferred method is to download the latest source archive from the
L<File Release System|https://sourceforge.net/projects/metaf2xml/files/2.9/>
at SourceForge.
Download either F<metaf2xml-2.9.zip> or F<metaf2xml-2.9.tgz>, depending on
which programs for unpacking are available to you and which you are most
comfortable with.

SourceForge also offers the creation of ZIP archives on the fly via the web
interface to the Git repository, e.g. go to the
L<latest release|https://sourceforge.net/p/metaf2xml/code/ci/r2.9/tree/>
or
L<development snapshot|https://sourceforge.net/p/metaf2xml/code/ci/unstable/tree/>
and then click on the S<"Download Snapshot"> button.
You could even download each file separately via these links.

Alternatively, you could clone the Git repository to your machine
(L<Git|https://git-scm.com/> must be installed, see the
L<SourceForge Git Wiki|https://sourceforge.net/p/forge/documentation/Git/>),
and check out the files, e.g. S<C<git checkout r2.9>> for the latest
release, or S<C<git checkout unstable>> for the development snapshot.

=item 3. Extract the files

First choose a (temporary) directory to which to extract files. It must not be
one of the intended installation directories or a parent or subdirectory. All
files in the archive will be extracted to a subdirectory named F<metaf2xml-2.9>
below the chosen directory.

Then use the command appropriate for the file type with the proper options
for the archive file name and the directory to which to extract files.

=item 4. Install the extracted files

First change to the directory with the extracted files, i.e. the subdirectory
F<metaf2xml-2.9> below the directory chosen in step 3.

Then execute S<C<perl install.pl>> (using the Perl binary from step 1) with the
L<appropriate argument|/ARGUMENTS> for the desired installation target.
If you want installation settings which differ from the default, specify them
as described in the L<next section|/Installation settings>.

=item 5. Test the installation

Now the installation can be tested using the example commands in the manuals for
the scripts C<metaf2xml.pl>, C<metafsrc2raw.pl> and C<metaf.pl>.

=item 6. Clean up the temporary files

After the installation, the archive file and the extracted files can be deleted.

=back

=head1 DEPENDENCIES

Perl 5.14.4+ is required.

=cut

my (%opts, $mkpath, $copy);
our $VERSION = '2.9';

############################################
# functions
############################################
sub usage {
    my $msg = shift;

    if ($msg eq '') {
        pod2usage -exitval => 'NOEXIT', -output => \*STDERR, -noperldoc => 1;
        print {*STDERR} <<"EOF";
Options may be merged together.  -- stops processing of options.
Space is not required between options and their arguments.
EOF
    } else {
        print {*STDERR} $msg;
    }

    print {*STDERR} <<"EOF";

For more details run
        perldoc -F $0
EOF
    exit 1;
}

sub getconfvar {
    my ($name, $default) = @_;
    my $val;

    $val = defined ($_ = (grep { /^$name=/ } @ARGV[1..$#ARGV])[0])
            ? (/=(.+)/ ? $1 : $default)
            : exists $ENV{$name} && $ENV{$name} ne '' ? $ENV{$name} : $default;
    usage "ERROR: value for parameter $name must not contain quotes\n"
        if index($val, q{'}) > -1 || index($val, q{"}) > -1;
    return $val;
}

sub do_copy {
    my ($to, @files) = @_;

    for (@files) {
        if ($to =~ m{/$}) {
            m{([^/]*)$};
            $copy->($_, "$to$1")
                or usage "ERROR: copy '$_': $!\n";
        } else {
            $copy->($_, $to)
                or usage "ERROR: copy '$_': $!\n";
        }
    }
    return;
}

sub do_copy_subst {
    my ($from, $to_dir, $mode, $subst) = @_;
    my ($to, $from_fh, $to_fh);

    if ($opts{n}) {
        printf 'do_copy_subst("%s", "%s", %o)' . "\n", $from, $to_dir, $mode;
        return;
    }

    $from =~ m{([^/]*)$};
    $to = "$to_dir/$1";
    open $from_fh, '<', $from
        or usage "ERROR: open '$from': $!\n";
    open $to_fh, '>', $to
        or usage "ERROR: open '$to': $!\n";
    while (<$from_fh>) {
        &$subst;
        print {$to_fh} $_
            or usage "ERROR: write '$to': $!\n";
    }
    close $from_fh
        or usage "ERROR: close '$from': $!\n";
    close $to_fh
        or usage "ERROR: close '$to': $!\n";
    if ($mode) {
        chmod $mode, "$to"
            or usage "ERROR: chmod '$to': $!\n";
    }
    return;
}

############################################
# installation settings
############################################

=head2 Installation settings

The default settings will install files according to the FHS but below the top
level directory F</opt/metaf2xml-2.9>.

Deviations from the default installation settings can be specified via
environment variables or as arguments to the installation command (after the
argument for the installation target) in the form C<CONFIG=VALUE>. Arguments
have priority over environment variables.

Note that all paths must be specified as absolute paths.

Also note: If cmd.exe is used in MS Windows, the leading and tailing double
quotation mark (e.g. added by directory name completion) must be omitted when
specifying paths, e.g. use:

 set METAF2XML=c:\Program Files\metaf2xml-2.9

=head3 Settings to change the destination directories for files to be installed

=over

=item METAF2XML (default: F</opt/metaf2xml-2.9>)

top level directory used as default for several other directories below

=cut

my $METAF2XML = getconfvar 'METAF2XML', '/opt/metaf2xml-2.9';

=item METAF2XML_BIN (targets: parser, ui. default: F<$METAF2XML/bin>)

destination directory for the Perl scripts metaf2xml.pl (target parser) and
metaf.pl, metafsrc2raw.pl (target ui)

=cut

my $METAF2XML_BIN = getconfvar 'METAF2XML_BIN', "$METAF2XML/bin";

=item METAF2XML_MAN (targets: parser, ui. default: F<$METAF2XML/share/man>)

destination directory for the manuals for parser.pm, metaf2xml.pl (target
parser) and src2raw.pm, metaf.pl, metafsrc2raw.pl (target ui)

=cut

my $METAF2XML_MAN = getconfvar 'METAF2XML_MAN', "$METAF2XML/share/man";

=item METAF2XML_LIB (targets: parser, ui. default: F<$METAF2XML/lib>)

destination directory for the Perl modules parser.pm (target parser) and
src2raw.pm (target ui)

=cut

my $METAF2XML_LIB = getconfvar 'METAF2XML_LIB', "$METAF2XML/lib";

=item METAF2XML_DOC (targets: parser. default: F<$METAF2XML/share/doc>)

destination directory for documentation files COPYING, ChangeLog, README

=cut

my $METAF2XML_DOC = getconfvar 'METAF2XML_DOC', "$METAF2XML/share/doc";

=item METAF2XML_SHARE (default: F<$METAF2XML/share>)

destination directory for metaf.dtd (target parser), metaf*.xsl,
metaf-lang*.xml, stations.xml (target xsl), and metaf2xml.css (target ui)

=cut

my $METAF2XML_SHARE = getconfvar 'METAF2XML_SHARE', "$METAF2XML/share";

=item CONF_DIR (targets: ui. default: F<$METAF2XML/etc>)

destination directory for the example configuration file for Apache

=cut

my $CONF_DIR = getconfvar 'CONF_DIR', "$METAF2XML/etc";

=back

=head3 Settings to change the runtime behaviour

The following settings determine the runtime behaviour of L<metaf|metaf>(1).
They can also be changed after the installation by changing the assignment of
the identically named variables in the installed file.

=over

=item CGI_DATA_DIR (default: F</var/opt/metaf2xml>)

directory for files with pre-processed METAR/TAF/SYNOP/BUOY/AMDAR messages

=cut

my $CGI_DATA_DIR = getconfvar 'CGI_DATA_DIR', '/var/opt/metaf2xml';

=item CGI_LOG_DIR (default: F</var/opt/metaf2xml>)

used to compose the value of CGI_LOG_FILE

=cut

my $CGI_LOG_DIR = getconfvar 'CGI_LOG_DIR', '/var/opt/metaf2xml';

=item CGI_LOG_NAME (default: F<metaf.log>)

used to compose the value of CGI_LOG_FILE

=cut

my $CGI_LOG_NAME = getconfvar 'CGI_LOG_NAME', 'metaf.log';

=item CGI_LOG_FILE (default: F<$CGI_LOG_DIR/$CGI_LOG_NAME>)

complete path for log file

=cut

my $CGI_LOG_FILE = getconfvar 'CGI_LOG_FILE', "$CGI_LOG_DIR/$CGI_LOG_NAME";

=item CURL_PATH (default: C<curl>)

full path for C<curl>

=cut

my $CURL_PATH = getconfvar 'CURL_PATH', 'curl';

=item CURL_OPTS (default: S<C<-sSf>>)

options for C<curl>

=cut

my $CURL_OPTS = getconfvar 'CURL_OPTS', '-sSf';

=item CURL (default: C<$CURL_PATH $CURL_OPTS>)

path and options for C<curl>

=cut

my $CURL = getconfvar 'CURL', "$CURL_PATH $CURL_OPTS";

=item CURL_PROXY_OPTS (default: empty)

options for C<curl> if an HTTP proxy should be used

=cut

my $CURL_PROXY_OPTS = getconfvar 'CURL_PROXY_OPTS', '';

=item XSLTPROC_PATH (default: C<xsltproc>)

full path for C<xsltproc>

=cut

my $XSLTPROC_PATH = getconfvar 'XSLTPROC_PATH', 'xsltproc';

=item DB_DATA_SOURCES (default: empty)

space delimited strings to describe possible data sources

=back

=cut

my $DB_DATA_SOURCES = getconfvar 'DB_DATA_SOURCES', '';

=head3 Miscellaneous settings for the installation

The following settings rarely need to be touched.

=over

=item PERL (default: F</usr/bin/perl>)

full path for the Perl binary to be used if the Perl scripts are started via
their name (e.g. S<C<./metaf2xml.pl ...>>)

=cut

my $PERL = getconfvar 'PERL', '/usr/bin/perl';

=item BUILD_ROOT (default: empty)

additional parent directory for all destination directories, to (temporarily)
install all files there instead of the intended final destination directory,
e.g. to build an RPM

=cut

my $BUILD_ROOT = getconfvar 'BUILD_ROOT', '';

=back

=cut

# files for xsl
my @METAF2XML_XSL_FILES = qw(
    metaf-fullhtml.xsl
    metaf-sum-ui.xsl
    metaf-html.xsl
    metaf-text.xsl
    metaf-sum-html.xsl
    metaf-sum-text.xsl
    metaf-lang.xml
    metaf-lang-de.xml
    metaf-lang-en.xml
    metaf-lang-es.xml
    metaf-sum.xsl
    metaf-ui.xsl
    metaf.xsl
    stations.xml
);

# files for doc
my @METAF2XML_DOC_FILES = qw(
    COPYING
    ChangeLog
    README
);

sub HELP_MESSAGE { usage ''; return; } # for Getopt::Std

# don't run if loaded as a library
return 1
    if caller;

=head1 OPTIONS

=over

=item -v

print version of install.pl and exit

=item -n

do not install anything, just print what would be done

=item -P

remove S<C<use lib ...>> from scripts. Use only when modules are installed to a
default Perl installation directory (site lib, vendor lib)

=back

=cut

# get and check options
$Getopt::Std::STANDARD_HELP_VERSION = 1;
getopts('vnP', \%opts)
    or usage '';
if ($opts{v}) {
    print "metaf2xml.pl version $VERSION\n" or exit 1;
    exit 0;
}

$mkpath  = $opts{n}
            ? sub { print 'mkpath("', join('", "', @{$_[0]}), '")', "\n" }
            : \&mkpath;
$copy    = $opts{n}
            ? sub { print 'copy("', join('", "', @_), '")', "\n" }
            : \&copy;

############################################
# targets
############################################

=head1 ARGUMENTS

The installation target must be specified. It can be one of the following:

=cut

my $cmd = $#ARGV >= 0 ? $ARGV[0] : '';

usage ''
    unless { ui => 1, parser => 1, xsl => 1, all => 1 }->{$cmd};

my $man3_sep =
      { VMS => 1, dos => 1 }->{$^O}                                  ? q{__}
    : { map { $_ => 1 } qw(cygwin os2 uwin uwin-nt MSWin32) }->{$^O} ? q{.}
    :                                                                  q{::}
    ;

=over

=item parser

Installs the files required to parse messages and create the XML.

=cut

if ($cmd eq 'all' || $cmd eq 'parser') {
    $mkpath->([ "$BUILD_ROOT$METAF2XML_BIN",
                "$BUILD_ROOT$METAF2XML_LIB/metaf2xml",
                "$BUILD_ROOT$METAF2XML_DOC",
                "$BUILD_ROOT$METAF2XML_MAN/man1",
                "$BUILD_ROOT$METAF2XML_MAN/man3",
                "$BUILD_ROOT$METAF2XML_SHARE"
              ]);

    do_copy "$BUILD_ROOT$METAF2XML_LIB/metaf2xml/",
            'metaf2xml/parser.pm', 'metaf2xml/bufr.pm';
    do_copy "$BUILD_ROOT$METAF2XML_MAN/man3/metaf2xml${man3_sep}parser.3pm",
            'metaf2xml/parser.3pm';
    do_copy "$BUILD_ROOT$METAF2XML_MAN/man3/metaf2xml${man3_sep}bufr.3pm",
            'metaf2xml/bufr.3pm';

    do_copy_subst 'metaf2xml.pl', "$BUILD_ROOT$METAF2XML_BIN", oct(755),
        sub {
              s{^#!/usr/bin/perl$}{#!$PERL}o;
              s{(?=^use File::Basename )}{# };
              $opts{P}
                ? s{(?=^use lib )}{# }
                : s{^use lib \K.*(?=; # METAF2XML_LIB\r?$)}{'$METAF2XML_LIB'}o;
            };
    do_copy "$BUILD_ROOT$METAF2XML_MAN/man1/", 'metaf2xml.1';

    do_copy "$BUILD_ROOT$METAF2XML_DOC/", @METAF2XML_DOC_FILES;
    do_copy "$BUILD_ROOT$METAF2XML_SHARE/", 'metaf.dtd';
}

=item xsl

Installs the XSL example files required to transform the XML to HTML and text.

=cut

if ($cmd eq 'all' || $cmd eq 'xsl') {
    $mkpath->(["$BUILD_ROOT$METAF2XML_SHARE"]);
    do_copy "$BUILD_ROOT$METAF2XML_SHARE/", @METAF2XML_XSL_FILES;
}

=item ui

Installs the files required to download and display weather data.

Requires the installation of the targets parser and xsl.

If the web interface is to be used, a web server like Apache is required. To
use Apache for the web interface, include the file
F<$CONF_DIR/metaf2xml-2.9-apache-2.X.conf> in your Apache configuration (e.g.
link or copy it to F</etc/apache/conf.d/>) and reload or restart Apache. Then
use the URL L<http://localhost/metaf2xml-2.9/cgi-bin/metaf.pl>.

=item all

Installs the targets parser, xsl, and ui.

=back

=cut

if ($cmd eq 'all' || $cmd eq 'ui') {
    $mkpath->([ "$BUILD_ROOT$METAF2XML_BIN",
                "$BUILD_ROOT$METAF2XML_LIB/metaf2xml",
                "$BUILD_ROOT$CONF_DIR",
                "$BUILD_ROOT$METAF2XML_MAN/man1",
                "$BUILD_ROOT$METAF2XML_MAN/man3"
              ]);
    do_copy_subst 'metaf.pl', "$BUILD_ROOT$METAF2XML_BIN", oct(755),
        sub {
              s{^#!/usr/bin/perl$}{#!$PERL}o;
              s{(?=^use File::Basename )}{# };
              $opts{P}
                ? s{(?=^use lib )}{# }
                : s{^use lib \K.*(?=; # METAF2XML_LIB\r?$)}{'$METAF2XML_LIB'}o;
              s{^my .METAF2XML_SHARE = \K.*(?=;\r?$)}{'$METAF2XML_SHARE'}o;
              s{^my .CGI_DATA_DIR = .\K.*(?=.;\r?$)}{$CGI_DATA_DIR}o;
              s{^my .CGI_LOG_FILE = .\K.*(?=.;\r?$)}{$CGI_LOG_FILE}o;
              s{^my .CURL_PATH = .\K.*(?=.;\r?$)}{$CURL_PATH}o;
              s{^my .CURL_OPTS = .\K.*(?=.;\r?$)}{$CURL_OPTS}o;
              s{^my .CURL = .\K.*(?=.;\r?$)}{$CURL}o;
              s{^my .CURL_PROXY_OPTS = .\K.*(?=.;\r?$)}{$CURL_PROXY_OPTS}o;
              s{^my .XSLTPROC_PATH = .\K.*(?=.;\r?$)}{$XSLTPROC_PATH}o;
              s{^my .DB_DATA_SOURCES = .\K.*(?=.;\r?$)}{$DB_DATA_SOURCES}o;
            };
    do_copy "$BUILD_ROOT$METAF2XML_MAN/man1/", 'metaf.1';

    do_copy_subst 'metafsrc2raw.pl', "$BUILD_ROOT$METAF2XML_BIN", oct(755),
        sub {
              s{^#!/usr/bin/perl$}{#!$PERL}o;
              s{(?=^use File::Basename )}{# };
              $opts{P}
                ? s{(?=^use lib .* METAF2XML_LIB)}{# }
                : s{^use lib \K.*(?=; # METAF2XML_LIB\r?$)}{'$METAF2XML_LIB'}o;
            };
    do_copy "$BUILD_ROOT$METAF2XML_MAN/man1/", 'metafsrc2raw.1';

    do_copy "$BUILD_ROOT$METAF2XML_LIB/metaf2xml/", 'metaf2xml/src2raw.pm';
    do_copy "$BUILD_ROOT$METAF2XML_MAN/man3/metaf2xml${man3_sep}src2raw.3pm",
            'metaf2xml/src2raw.3pm';

    do_copy "$BUILD_ROOT$METAF2XML_SHARE/", 'metaf2xml.css';

    for my $v (qw(2.2 2.4)) {
        do_copy_subst 'metaf2xml-2.9-apache-' . $v . '.conf',
            "$BUILD_ROOT$CONF_DIR", 0,
            sub { s{METAF2XML_SHARE}{$METAF2XML_SHARE}o;
                  s{METAF2XML_BIN}{$METAF2XML_BIN}o;
                };
    }
}

exit 0;

=head1 EXIT STATUS

If an invalid option was provided or an error occurs during the installation,
the script will exit with status 1, otherwise it will exit with status 0.

=head1 EXAMPLES

Install all files in the default destination:

 perl install.pl all

Install all files required to parse messages in the structure defined by the FHS
but below the directory F</usr/local>:

 perl install.pl parser METAF2XML=/usr/local

or via prefixed assignment of environment variables in (any variant of) UNIX
C<sh>:

 METAF2XML=/usr/local ./install.pl parser

=head1 SEE ALSO

L<metaf2xml::src2raw|metaf2xml::src2raw>(3pm),
L<metaf2xml|metaf2xml>(1), L<metafsrc2raw|metafsrc2raw>(1), L<metaf|metaf>(1),
L<https://metaf2xml.sourceforge.io/>

=head1 COPYRIGHT and LICENSE

copyright (c) 2011-2019 metaf2xml @ L<https://metaf2xml.sourceforge.io/>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see L<https://www.gnu.org/licenses/>.

=cut
