#!/usr/bin/perl

########################################################################
# metaf.pl 2.9
#   (CGI) script to download and display METAR/TAF/SYNOP/BUOY/AMDAR messages
#
# copyright (c) 2006-2023 metaf2xml
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
########################################################################

########################################################################
# some things strictly Perl
########################################################################
use warnings;
use 5.014_004;

=encoding utf8

=head1 NAME

metaf.pl - download and display METAR/TAF/SYNOP/BUOY/AMDAR messages

=head1 SYNOPSIS

metaf.pl can be run from the command line like this:

 metaf.pl 'type_metaf=...&type_synop=...&type_buoy=...&type_amdar=...&lang=...&unit_temp=...&format=...&src_metaf=...&src_synop=...&src_buoy=...&src_amdar=...&mode=...&hours=...&msg_metaf=...&msg_synop=...&msg_buoy=...&msg_amdar=...'

or like this:

 metaf.pl type_metaf=... type_synop=... type_buoy=... type_amdar=... lang=... unit_temp=... format=... src_metaf=... src_synop=... src_buoy=... src_amdar=... mode=... hours=... msg_metaf='...' msg_synop='...' msg_buoy='...' msg_amdar='...'

It can also be used as a CGI script via an HTML form like this:

 <form method=GET action=metaf.pl>
  <input name=type_metaf value=... >
  <input name=type_synop value=... >
  <input name=type_buoy value=... >
  <input name=type_amdar value=... >
  <input name=lang value=... >
  <input name=unit_temp value=... >
  <input name=format value=... >
  <input name=src_metaf value=... >
  <input name=src_synop value=... >
  <input name=src_buoy value=... >
  <input name=src_amdar value=... >
  <input name=mode value=... >
  <input name=hours value=... >
  <input name=msg_metaf value="..." >
  <input name=msg_synop value="..." >
  <input name=msg_buoy value="..." >
  <input name=msg_amdar value="..." >
 </form>

=head1 DESCRIPTION

This script can download and display in plain language up-to-date or historical
weather information (METAR, TAF, SYNOP, BUOY or AMDAR messages) from Internet
servers, an SQL database, or a local file.
It can be used from the command line, or executed as a CGI script by a web
server with HTTP GET requests via an HTML form.

If the script is executed by a web server and the file $CGI_LOG_FILE is
writable by the user running the web server (e.g. "wwwrun"), metaf.pl will log
the requests there.

=head1 DEPENDENCIES

The Perl modules C<CGI>, C<metaf2xml::parser> and C<metaf2xml::src2raw>
are required.

The Perl module C<DBI> and the module C<DBD> (for the configured type of data
source) are required if an SQL database is used as the data source.

The program C<curl> is required to download weather messages from Internet
servers.

The program C<xsltproc> is required if the script is to convert the XML to HMTL
or text.

=head1 ABBREVIATIONS

=over 7

=item noaa

National Oceanic and Atmospheric Administration

=item nws

the National Weather Service department of the NOAA

=item iws

the Internet Weather Source of the NWS

=item addsds

the Text Data Server API service of the Aviation Weather Center of the NOAA

=item cod

College of DuPage

=back

=head1 OPTIONS

=over

=item -v

print version of metaf.pl and exit

=back

=head1 ARGUMENTS

The following variables can be set from CGI parameters or on the command line:

=over

=item type_metaf

type of content in C<msg_metaf>:

=over 6

=item B<metar>

METAR/SPECI/SAO message (default)

=item B<taf>

TAF message

=item B<icao>

ICAO id(s)

=back

If C<msg_metaf> starts with C<METAR>, C<SPECI> or C<TAF>, C<type_metaf> is set
to B<metar> or B<taf>, respectively.

If C<msg_metaf> is not a list of ICAO id(s), C<type_metaf> is set to B<metar>.

If C<type_metaf> is B<icao> and one or more ICAO ids are given, messages for
the specified airports are fetched from C<src_metaf>.

=item type_synop

type of content in C<msg_synop>:

=over 6

=item B<synop>

SYNOP message (default)

=item B<wmo>

WMO SYNOP id(s)

=item B<nat>

national id(s) I<STATION>-I<COUNTRY>

=item B<ship>

SYNOP SHIP id(s)

=item B<mobil>

SYNOP MOBIL id(s)

=back

If C<msg_synop> starts with C<AAXX>, C<BBXX> or C<OOXX> or is not a list of
SYNOP ids, C<type_synop> is set to B<synop>.

If C<type_synop> is B<wmo>, B<ship>, B<mobil> or B<nat> and one or more SYNOP
ids are given, messages for the specified stations are fetched from
C<src_synop>.

=item type_buoy

type of content in C<msg_buoy>:

=over 5

=item B<buoy>

BUOY message (default)

=item B<wmo>

WMO buoy id(s)

=back

If C<msg_buoy> is not a list of WMO buoy ids, C<type_buoy> is set to B<buoy>.

If C<type_buoy> is B<wmo> and one or more WMO buoy ids are given, messages for
the specified buoys are fetched from C<src_buoy>.

=item type_amdar

type of content in C<msg_amdar>:

=over 6

=item B<amdar>

AMDAR message (default)

=item B<ac>

aircraft id(s)

=back

If C<msg_amdar> starts with C<AMDAR> or C<ARP>, or is not a list of aircraft
ids, C<type_amdar> is set to B<amdar>.

If C<type_amdar> is B<ac> and one or more aircraft ids are given, messages for
the specified aircrafts are fetched from C<src_amdar>.

=item lang

language for output of script and display of parsed message:
B<de> (default),
B<en>,
B<es>

If C<lang> is not set the first allowed value of the HTTP header
C<Accept-Language> is used.

=item unit_temp

unit for temperatures in summary mode:
B<C> (default),
B<F>

=item format

output format:
B<html> (default),
B<text>,
B<xml> (with style sheet reference)

=item src_metaf

source for METAR/TAF messages to be fetched:
B<noaa> (NOAA/IWS)[1] (default),
B<addsds>[2],
B<ogimet>[2],
B<nws>[1],
B<cod>[1] (no TAF).
It is ignored if a source in C<DB_DATA_SOURCES>[3] could be opened or
F<$CGI_DATA_DIR/metaf.txt>[4] exists.

Note: B<cod> is not available in the web-based interfaces.

=item src_synop

source for SYNOP messages to be fetched:
B<nws>[1] (default),
B<ogimet>[2],
B<cod>[1].
It is ignored if a source in C<DB_DATA_SOURCES>[3] could be opened or
F<$CGI_DATA_DIR/synop.txt>[4] exists.

=item src_buoy

source for BUOY messages to be fetched:
B<nws>[1] (default),
B<ogimet>[2].
It is ignored if a source in C<DB_DATA_SOURCES>[3] could be opened or
F<$CGI_DATA_DIR/buoy.txt>[4] exists.

=item src_amdar

source for AMDAR messages to be fetched:
B<nws>[1] (default).
It is ignored if a source in C<DB_DATA_SOURCES>[3] could be opened or
F<$CGI_DATA_DIR/amdar.txt>[4] exists.

=item mode

which messages to show and how to show them:

=over

=item B<summary>

show all messages in selected time period, show one row per message (TAF:
one row per forecast period, AMDAR: one row per observation)

=item any other value (or omitted)

show only latest message for each station in selected time period, show
values in detail

=back

=item hours

for mode B<summary>: number of hours to fetch messages for (0..72, default: 24)

Note: The exact time range for which data can be fetched depends on the source
of messages.

=item msg_metaf

METAR/TAF message to be processed, or ICAO id(s)

=item msg_synop

SYNOP message to be processed, or WMO id(s)

=item msg_buoy

BUOY message to be processed, or WMO id(s)

=item msg_amdar

AMDAR message to be processed, or aircraft id(s)

=back

[1]: Files will be fetched from the server and searched.
Note: The parameter B<hours> is used to select files by their time stamp or
name; the selected files do not necessarily contain only messages for the
chosen B<hours>.

[2]: the HTML form provided by the server will be used

[3]: For the description of the configuration variable $DB_DATA_SOURCES see the
I<USAGE NOTES> for C<DB_DATA_SOURCES> below.

[4]: For the format required for the files in F<$CGI_DATA_DIR> see the I<USAGE
NOTES> for C<CGI_DATA_DIR> below.

=cut

# all times are UTC (esp. required for mktime())
$ENV{TZ} = 'UTC'; ## no critic (RequireLocalizedPunctuationVars)

use Getopt::Std;
use Pod::Usage 'pod2usage';
use File::Basename 'dirname';
use POSIX qw(mktime WIFEXITED WEXITSTATUS WTERMSIG);
use CGI qw(charset :cgi);

our $VERSION = '2.9';

# path will be set by install
use lib dirname $0; # METAF2XML_LIB
use metaf2xml::parser 2.009 qw(start_xml parse_report finish);
use metaf2xml::src2raw 2.009 qw(start_file next_line);

#
# restrict use of CGI:
#

# uploads are not allowed
$CGI::DISABLE_UPLOADS = 1;

=head1 USAGE NOTES

The length of the body of the POST request (not the length of query string of
the URL) is restricted to 600 bytes.

=cut

$CGI::POST_MAX = 600;

########################################################################
# config section, possibly changed by install
########################################################################

=pod

The following variables can be adapted to local needs; this is typically done
during the installation:

=over

=item METAF2XML_SHARE

directory for the XSL style sheets

=cut

# path will be set by install
my $METAF2XML_SHARE = dirname $0;

=item CGI_DATA_DIR

directory for files with pre-processed METAR/TAF/SYNOP/BUOY/AMDAR messages

The messages in these files must have the format specified by the I<WMO Manual
No. 306>, without modifications due to distribution like providing the initial
part of messages only once for several messages or appending an "=" (equal sign)
to terminate a message, with the data type C<T1T2> prepended. This format can
be obtained using:

 metafsrc2raw.pl -F<FORMAT> -H1 [FILE] ...

=cut

my $CGI_DATA_DIR = '/var/opt/metaf2xml';

=item CGI_LOG_FILE

complete path for log file

=cut

my $CGI_LOG_FILE = '/var/opt/metaf2xml/metaf.log';

=item CURL_PATH

full path for C<curl>

=cut

my $CURL_PATH = 'curl';

=item CURL_OPTS

options for C<curl>

=cut

my $CURL_OPTS = '-sSf';

=item CURL

path and options for C<curl>

=cut

my $CURL = "$CURL_PATH $CURL_OPTS";

=item CURL_PROXY_OPTS

options for C<curl> if an HTTP proxy should be used

=cut

my $CURL_PROXY_OPTS = '';

=item XSLTPROC_PATH

full path for C<xsltproc>

=cut

my $XSLTPROC_PATH = 'xsltproc';

=item DB_DATA_SOURCES

space delimited strings to describe possible data sources

If the string C<sqlite> is in the list, the file F<$CGI_DATA_DIR/sqlite.db> is
used as database.
For other strings, the file F<$CGI_DATA_DIR/>C<string>F<.db> should contain the
Data Source Name (C<DBI:>...) in the first line. If the second line starts with
C<user=>, the rest of the line is used as user name for the authentication. If
a user name was given and the third line starts with C<pass=>, the rest of the
line is used as password for the authentication.

The first data source which can be opened is used. If the last data source fails
to open or C<DB_DATA_SOURCES> is empty, the parameters C<src_>* are used.

However, if a string ends in C</f>, this is the last data source to try and the
parameters C<src_>* are not used, either.
If this source fails to open, an error will be reported.

The Perl module C<DBI> and the module C<DBD> for the configured data source are
required if a database is to be used.

=cut

my $DB_DATA_SOURCES = '';

=back

=cut

#
# define some regular expressions
#
my $re_report_type_synop = '(?:I?S[IMN])';
my $re_report_type_metaf = '(?:S[AP]|F[CT])';
my $re_report_type_buoy = '(?:SS|IOB)';
my $re_report_type_amdar = '(?:UD|IUA)';
my $re_header_synop = '(?:AAXX .{5}|(?:BB|OO)XX)';
my $re_header_buoy = '(?:ZZYY)';
my $re_header_amdar = '(?:(?:AMDAR|ARP) \d{4})';
my $re_ICAO = '(?:[A-Z][A-Z\d]{3})';
my $re_day = '(?:[0-2]\d|3[01])';
my $re_hour = '(?:[01]\d|2[0-3])';
my $re_min  = '[0-5]\d';
my $re_header_metaf = '(?:COR |AMD |METAR |TAF |SPECI )*';
my $re_header_sao = "([A-Z]{3,4}) (S[AP]|RS)(?: COR)? $re_hour$re_min (?:AUTO|AWOS|SAWR)";
my $re_aircraft = '[A-Z]{2}[A-Z0-9]{2,6}';

my $is_cmdline = server_software() eq 'cmdline' ? 1 : 0;

#
# make a list of strings unique
#
sub mk_uniq { ## no critic (RequireArgUnpacking)
    my $uniq = '';
    for (@_) {
        if ($uniq eq '') {
            $uniq = $_;
        } elsif (index(" $uniq ", " $_ ") == -1) {
            $uniq .= " $_";
        }
    }
    return $uniq;
}

sub error {
    print {*STDERR} shift;
    return '';
}

#
# provide info for processes with errors
#
sub exit_info {
    my ($error, $rc, $msg) = @_;
    my $wifexited;

    $wifexited = eval { WIFEXITED($rc) };

    return error 'ERROR'
           . ($error ? " $error" : '') . ': '
           . (defined $wifexited
                             ? ($wifexited ? 'exit(' . WEXITSTATUS($rc) . ')'
                                           : 'signal(' . WTERMSIG($rc) . ')')
                             : "retval($rc)")
           . ": $msg\n";
}

#
# provide pipe open depending on OS
#
sub pipe_open {
    my ($fh, $is_read, @args) = @_;
    my $pid;

    if ({ map { $_ => 1 } qw(dos MSWin32 NetWare MacOS riscos) }->{$^O}) {
        my ($at_beginning, $at_end, $rc);

        ($at_beginning, $at_end) = $is_read ? ('', '|') : ('|', '');

        # quote spaces in args
        $rc = open $$fh,   $at_beginning ## no critic (ProhibitTwoArgOpen)
                         . join(' ', map { / / ? "\"$_\"" : $_ } @args)
                         . $at_end;
        if ($is_read && defined $rc) {
            binmode $$fh, ':bytes'
                or return error "ERROR binmode input: $!.\n"
        }
        return $rc;
    }
    $pid = open $$fh, $is_read ? '-|' : '|-'; ## no critic (ProhibitTwoArgOpen)
    return error "ERROR starting '@args': $!.\n"
        unless defined $pid;     # fork failed?
    if ($pid > 0) {              # parent. check $? later
        if ($is_read) {
            binmode $$fh, ':bytes'
                or return error "ERROR binmode input: $!.\n";
        }
        return 1;
    }
    open STDERR, '>&STDOUT'      # child ...
        if !$is_cmdline;
    exec @args or exit 1;
}

#
# determine max. available hours
#
sub max_hours {
    my ($max_hours, $hours, $now, $end_date) = @_;

    $end_date = $now
        if !defined $end_date || $end_date > $now;

    if ($max_hours > 0) {
        my ($hh_start_avail, $hh_end_date, $hh_avail);

        $hh_start_avail = $now - $max_hours * 3600;
        $hh_start_avail -= $hh_start_avail % 3600;
        $hh_end_date    = $end_date - $end_date % 3600;
        $hh_avail       = ($hh_end_date - $hh_start_avail) / 3600;
        $hours = $hh_avail
            if $hours > $hh_avail;
    }

    return ($hours, $end_date);
}

# for testing
BEGIN {
    *add_searched_file = sub { return; }
        unless defined &add_searched_file;
    *gettime = sub { return time; }
        unless defined &gettime;
}

my $data_url;
sub set_data_url {
    $CGI_DATA_DIR = shift;
    $data_url = 'file:' . $CGI_DATA_DIR;
    return;
}
set_data_url $ENV{M2X_TEST_DATA}
    if $ENV{M2X_TEST_DATA};

sub fmtTime {
    my ($time, $format) = @_;
    my @t;

    # gmtime returns struct tm. copy year, month, day, hour, minute
    @t = (gmtime $time)[5, 4, 3, 2, 1];
    # year starts from 1900, month from 0
    return sprintf $format, $t[0] + 1900, $t[1] + 1, @t[2 .. 4];
}

sub mkTime {
    my @t = @_; # year, month, day, hour, minute, second

    # year starts from 1900, month from 0
    return mktime @t[5, 4, 3, 2], $t[1] - 1, $t[0] - 1900, -1, -1, -1;
}

#
# get the contents of the directory on the NOAA/NWS server from the server
#   type: metar, taf, land, ship, buoy, amdar
sub files_nws {
    my ($type, $hours, $end_date) = @_;
    my ($url, $cmd, $fh, $response, @files, $oldest_file, $now);

    # fetch directory listing
    # available also via FTP
    $url = ($data_url // 'https://tgftp.nws.noaa.gov/SL.us008001/DF.an')
           . '/'
           . { metar => 'DC.sflnd/DS.metar',
               taf   => 'DC.sflnd/DS.tafst',
               land  => 'DC.sflnd/DS.synop',
               ship  => 'DC.sfmar/DS.ships',
               buoy  => 'DC.sfmar/DS.dbuoy',
               amdar => 'DC.sluan/DS.amdar',
             }->{$type};
    $cmd = "$CURL $CURL_PROXY_OPTS $url/ls-lt";
    pipe_open \$fh, 1, split(' ', "$CURL $CURL_PROXY_OPTS"), "$url/ls-lt"
        or return '';

    {
        local $/;
        $response = <$fh>;
    }
    if (!close $fh || $?) {
        $response =~ s/[\n\r]+/ /g;
        return exit_info $!, $?, $cmd . ($response ? ":\n$response" : '.');
    }

    @files = $response =~ / (sn.\d{4}.txt)/g;

    # don't use the oldest file, it may have/may shortly become the newest
    $oldest_file = $#files - 1;

    # ignore older files if less hours are requested
    # (one file every 5 min. = 12 per hour)
    $oldest_file = $hours * 12
        if $hours * 12 < $oldest_file;

    # ignore newer files if end date is before now
    $now = gettime;
    $end_date = $now
        if !defined $end_date || $end_date > $now;
    # round down to next 5 min.
    $now      -= $now % 300;
    $end_date -= $end_date % 300;

    return $url, @files[($now - $end_date) / 300 .. $oldest_file];
}

#
# determine file names for data from NOAA/IWS (cycle) files
#   type: metar, taf
sub files_cycle_iws {
    my ($type, $icao, $hours, $now, $end_date) = @_;
    my ($offset, $newest_hh, $end_hh, $oldest_hh, $need_fc, $need_ft);

    # METAR cycle files start 15 min. before the hour
    # TODO: TAF cycle files start 60 min. before the hour?!?
    $offset = $type eq 'metar' ? 15 * 60 : 0;

    # newest file:
    $newest_hh = $now + $offset;
    $newest_hh -= $newest_hh % 3600;
    $newest_hh /= 3600;

    # file requested for end_date:
    $end_date = $now
        if !defined $end_date || $end_date > $now;
    $end_hh = $end_date + $offset;
    $end_hh -= $end_hh % 3600;
    $end_hh /= 3600;

    # oldest file available:
    $oldest_hh = $newest_hh - ($type eq 'metar' ? 23 : ($newest_hh % 6) + 18);
    $oldest_hh = $end_hh - $hours + 1
        if $oldest_hh < $end_hh - $hours + 1;

    return   ($data_url // 'https://tgftp.nws.noaa.gov/data')
           . '/observations/metar/cycles',
           map { sprintf '%02dZ.TXT', $_ % 24 } reverse $oldest_hh .. $end_hh
        if $type eq 'metar';

    # round down to next 6 h
    $end_hh -= $end_hh % 6;
    $oldest_hh -= $oldest_hh % 6;

    $need_fc = " $icao" =~ / (?:FC-)?$re_ICAO\b/o;
    $need_ft = " $icao" =~ / (?:FT-)?$re_ICAO\b/o;

    return ($data_url // 'https://tgftp.nws.noaa.gov/data'),
           map {(  $need_fc
                 ? sprintf('forecasts/shorttaf/cycles/%02dZ.TXT', ($_ * 6) % 24)
                 : (),
                   $need_ft
                 ? sprintf('forecasts/taf/cycles/%02dZ.TXT', ($_ * 6) % 24)
                 : ())
               } reverse $oldest_hh / 6 .. $end_hh / 6;
}

#
# determine file names for hourly files
#
sub files_hourly {
    my ($max_hours, $hours, $end_date, $format, @ids) = @_;
    my @files;

    ($hours, $end_date) = max_hours $max_hours, $hours, gettime, $end_date;

    # one file every hour
    for (0 .. $hours - 1) {
        my @t;

        # gmtime returns struct tm. copy year, year, month, day, hour
        @t = (gmtime $end_date - $_ * 60 * 60)[5, 5, 4, 3, 2];
        $t[0] %= 100;  # year starts from 1900, format as YY
        $t[1] += 1900; # year starts from 1900
        $t[2]++;       # month starts from 0
        push @files, sprintf $format, @t[@ids];
    }
    return @files;
}

#
# determine file names for data from College of DuPage
#   type: syn, sao
sub files_cod {
    my ($type, $hours, $end_date) = @_;

    return $data_url // "https://weather.cod.edu/digatmos/$type",
           files_hourly 60, $hours, $end_date,
                        "%02d%02d%02d%02d.$type", 0, 2 .. 4;
}

#
# determine file names for data from Iowa State University
#  - historical text files since 2005-07-25 (except for 2011 .. 2013)
#  - provided at 08:10 for the previous day
#  - METAR: selected stations, only
#  - TAF: none
#  - SYNOP: not all stations for all observation times, but some with nat. data
#
#   type: syn, boy, sao, upa
sub files_isu {
    my ($type, $hours, $end_date) = @_;

    # also:  ftp://mtarchive.geol.iastate.edu/data
    return $data_url // 'http://mtarchive.geol.iastate.edu',
           files_hourly -1, $hours, $end_date,
                        "%d/%02d/%02d/text/$type/%d%02d%02d%02d_$type.txt",
                        1 .. 3, 1 .. 4;
}

#
# handle data from opened files
#
{
    my @latest;

    sub start_fh {
        @latest = ();
        return;
    }

    sub next_fh {
        my $fh = shift;

        # line(s) from previous call?
        return pop @latest
            if $#latest > -1;

        while (1) {
            # get one line from input and parse
            $_ = <$fh>;
            @latest = next_line $_;

            # return EOF on error
            last
                if $#latest == 0 && !defined $latest[0];
            # have line(s)? return first one, others ones are for next call(s)
            return pop @latest
                if $#latest > -1;
            # found no lines and input has EOF? return EOF
            last
                unless defined;
        }
        return;
    }
}

#
# scan files with SYNOP data for $wmoid
#
sub search_files_synop {
    my ($wmoid, $mode, $src, $hours, $type_synop, $end_date) = @_;
    my ($url, @files, $no_wmoid, @prev_msgs);

    # these files have no BUFR
    $wmoid =~ s/ IS[IMN]-[^ ]+|IS[IMN]-[^ ]+ ?//g;
    return
        if $wmoid eq '';

    $no_wmoid = $wmoid =~ tr/ // + 1;

    $hours = 7 unless $mode eq 'summary'; # overwrite for mode latest
    ($url, @files) =   $src eq 'local' ? ('', "$CGI_DATA_DIR/synop.txt")
                     : $src eq 'cod'   ? files_cod('syn', $hours, $end_date)
                     : $src eq 'isu'   ? files_isu('syn', $hours, $end_date)
                     : $src eq 'nws'   ? files_nws($type_synop eq 'ship'
                                                              ? 'ship' : 'land',
                                                   $hours, $end_date)
                     :                   ();
    return
        unless defined $url;
    return ''
        if $src eq 'nws' && $url eq '';

    @prev_msgs = ();
    for my $file (@files) {
        my ($cmd, $fh, @curr_msgs, %msg_found, $msg_seq);

        if ($url) {
            add_searched_file "$url/$file";
            $cmd = "$CURL $CURL_PROXY_OPTS $url/$file";
            start_file("synop_$src", 1, 1);
            start_fh;
            pipe_open \$fh, 1, split(' ', "$CURL $CURL_PROXY_OPTS"),"$url/$file"
                or return '';
        } else {
            add_searched_file $file;
            $cmd = $file;
            open $fh, '<:bytes', $file ## no critic (RequireBriefOpen)
                or return error "ERROR: open '$cmd' failed: $!.\n";
        }

        @curr_msgs = ();
        $msg_seq  = '';
        while (my $line = $url ? next_fh $fh : <$fh>) {
            my ($report_type, $report_type_allowed, $msg, $id);
            my ($report_type_id, $report_type_matches, $report_type_any);

            chomp $line
                unless $url;

            ($report_type, $msg, $id) = $line =~ m{^($re_report_type_synop) ($re_header_synop ([A-Z\d/]{3,}).*)}o;
            next
                unless defined $id;

            $report_type_id = "$report_type-$id";
            $report_type_matches = index(" $wmoid ", " $report_type_id ") > -1;
            $report_type_any     = index(" $wmoid ", " $id ") > -1;

            # - skip if not required
            # - prepare check if message was in previous (= more recent) files
            #   for mode=latest
            # looking for messages with this id?
            if ($report_type_matches) {
                # skip if report_type_id was in previous files
                $report_type_allowed = $report_type;
            } elsif ($report_type_any) {
                # skip if any type was in previous files
                $report_type_allowed = $re_report_type_synop;
            } else {
                next;
            }

            # include station
            $line = "$report_type $id $msg";

            # looking for all messages (with correct id)?
            if ($mode eq 'summary') {
                # prepend unless it already exists
                unshift @curr_msgs, $line
                    unless grep { /^\Q$line\E$/ } (@prev_msgs, @curr_msgs);
                next;
            }

            # skip if message was in previous (= more recent) files
            next
                if grep { /^$report_type_allowed $id / } @prev_msgs;

            # request  : input             => output
            # id       : Sy-id,Sx-id,Sz-id => Sz-id
            # Sx-id    : Sy-id,Sx-id,Sz-id => Sx-id
            # Sx-id,id : Sx-id,Sy-id       => Sx-id,Sy-id
            # Sx-id,id : Sy-id,Sx-id       => Sx-id
            # looking for messages with the report type of this message?
            if ($report_type_matches) {
                if ($msg_found{$report_type_id}) {  # record type found already?
                    $msg_seq =~ s/ $report_type_id //;   # remove existing entry
                } else {
                    $no_wmoid--;                    # mark record type as found
                }
                $msg_seq = " $report_type_id " . $msg_seq;   # prepend new entry
                $msg_found{$report_type_id} = $line;         # save new value
            }

            # looking for messages with any report type?
            if ($report_type_any) {
                if ($msg_found{$id}) {
                    $msg_seq =~ s/ $id //;
                } else {
                    $no_wmoid--;
                }
                # prepend only if not done already above
                $msg_seq = " $id " . $msg_seq
                    unless $report_type_matches;
                $msg_found{$id} = $line;
            }
        }
        (close $fh && !$?)
            or return exit_info $!, $?, "$wmoid - $cmd";

        # fill curr_msgs if not done yet
        @curr_msgs = map { $msg_found{$_} } $msg_seq =~ /[^ ]+/g
            if $mode ne 'summary';

        # append data of this file to data of previous (= more recent) files
        @prev_msgs = (@prev_msgs, @curr_msgs);

        last if $mode eq 'latest' && $no_wmoid == 0;
    }

    return @prev_msgs;
}

#
# scan files with BUOY data for $wmoid
#
sub search_files_buoy {
    my ($wmoid, $mode, $src, $hours, $end_date) = @_;
    my ($url, @files, $no_wmoid, @prev_msgs);

    # these files have no BUFR, and SS-X is the same as X
    $wmoid = mk_uniq $wmoid =~ /\G(?:IOB-[^ ]+ ?)*+(?:SS-)?([^ ]+) ?/g;
    return
        if $wmoid eq '';

    $no_wmoid = $wmoid =~ tr/ // + 1;

    $hours = 7 unless $mode eq 'summary'; # overwrite for mode latest
    ($url, @files) =   $src eq 'local' ? ('', "$CGI_DATA_DIR/buoy.txt")
                     : $src eq 'isu'   ? files_isu('boy',  $hours, $end_date)
                     : $src eq 'nws'   ? files_nws('buoy', $hours, $end_date)
                     :                   ();
    return
        unless defined $url;
    return ''
        if $src eq 'nws' && $url eq '';

    @prev_msgs = ();
    for my $file (@files) {
        my ($cmd, $fh, @curr_msgs, %msg_found, $msg_seq);

        if ($url) {
            add_searched_file "$url/$file";
            $cmd = "$CURL $CURL_PROXY_OPTS $url/$file";
            start_file("buoy_$src", 1, 1);
            start_fh;
            pipe_open \$fh, 1, split(' ', "$CURL $CURL_PROXY_OPTS"),"$url/$file"
                or return '';
        } else {
            add_searched_file $file;
            $cmd = $file;
            open $fh, '<:bytes', $file ## no critic (RequireBriefOpen)
                or return error "ERROR: open '$cmd' for $wmoid failed: $!.\n";
        }

        @curr_msgs = ();
        $msg_seq  = '';
        while (my $line = $url ? next_fh $fh : <$fh>) {
            my ($report_type, $msg, $id);

            chomp $line
                unless $url;

            ($report_type, $msg, $id) = $line =~ /^($re_report_type_buoy) ($re_header_buoy (\d{5}).*)/o;
            next
                unless defined $id && index(" $wmoid ", " $id ") > -1;

            # include station
            $line = "$report_type $id $msg";

            # looking for all messages (with correct id)?
            if ($mode eq 'summary') {
                # prepend unless it already exists
                unshift @curr_msgs, $line
                    unless grep { /^\Q$line\E$/ } (@prev_msgs, @curr_msgs);
                next;
            }

            # skip if message was in previous (= more recent) files
            next
                if grep { /^$report_type $id / } @prev_msgs;

            if ($msg_found{$id}) {
                $msg_seq =~ s/ $id //;
            } else {
                $no_wmoid--;
            }
            $msg_seq = " $id " . $msg_seq;   # prepend new entry
            $msg_found{$id} = $line;
        }
        (close $fh && !$?)
            or return exit_info $!, $?, "$wmoid - $cmd";

        # fill curr_msgs if not done yet
        @curr_msgs = map { $msg_found{$_} } $msg_seq =~ /[^ ]+/g
            if $mode ne 'summary';

        # append data of this file to data of previous (= more recent) files
        @prev_msgs = (@prev_msgs, @curr_msgs);

        last if $mode eq 'latest' && $no_wmoid == 0;
    }

    return @prev_msgs;
}

#
# scan files with AMDAR data for $acid
#
sub search_files_amdar {
    my ($acid, $mode, $src, $hours, $end_date) = @_;
    my ($url, @files, $no_acid, @prev_msgs);

    # these files have no BUFR, and UD-X is the same as X
    $acid = mk_uniq $acid =~ /\G(?:IUA-[^ ]+ ?)*+(?:UD-)?([^ ]+) ?/g;
    return
        if $acid eq '';

    $no_acid = $acid =~ tr/ // + 1;

    $hours = 7 unless $mode eq 'summary'; # overwrite for mode latest
    ($url, @files) =   $src eq 'local' ? ('', "$CGI_DATA_DIR/amdar.txt")
                     : $src eq 'isu'   ? files_isu('upa',  $hours, $end_date)
                     : $src eq 'nws'   ? files_nws('amdar', $hours, $end_date)
                     :                   ();
    return
        unless defined $url;
    return ''
        if $src eq 'nws' && $url eq '';

    @prev_msgs = ();
    for my $file (@files) {
        my ($cmd, $fh, @curr_msgs, %msg_found, $msg_seq);

        if ($url) {
            add_searched_file "$url/$file";
            $cmd = "$CURL $CURL_PROXY_OPTS $url/$file";
            start_file("amdar_$src", 1, 1);
            start_fh;
            pipe_open \$fh, 1, split(' ', "$CURL $CURL_PROXY_OPTS"),"$url/$file"
                or return '';
        } else {
            add_searched_file $file;
            $cmd = $file;
            open $fh, '<:bytes', $file ## no critic (RequireBriefOpen)
                or return error "ERROR: open '$cmd' for $acid failed: $!.\n";
        }

        @curr_msgs = ();
        $msg_seq  = '';
        while (my $line = $url ? next_fh $fh : <$fh>) {
            my ($report_type, $msg, $id);

            chomp $line
                unless $url;

            ($report_type, $msg, $id) = $line =~ m{^($re_report_type_amdar) ($re_header_amdar (?:[A-Z]{3}|///) ($re_aircraft).*)}o;
            next
                unless defined $id && index(" $acid ", " $id ") > -1;

            # include station
            $line = "$report_type $id $msg";

            # looking for all messages (with correct id)?
            if ($mode eq 'summary') {
                # does it already exist?
                for (@prev_msgs, @curr_msgs) {
                    my ($rt_line, $msg_line, $msg2);

                    # disregard "AMDAR YYGG" for comparison, YYGG may vary
                    ($rt_line, $msg_line) = /^($re_report_type_amdar) $re_aircraft $re_header_amdar (.*)/o;
                    ($msg2) = $msg =~ /^$re_header_amdar (.*)/o;
                    if ("$rt_line $msg_line" eq "$report_type $msg2") {
                        $msg = undef;
                        last;
                    }
                }
                # prepend unless it already exists
                unshift @curr_msgs, $line
                    if $msg;
                next;
            }

            # skip if message was in previous (= more recent) files
            next
                if grep { /^$report_type $id / } @prev_msgs;

            if ($msg_found{$id}) {
                $msg_seq =~ s/ $id //;
            } else {
                $no_acid--;
            }
            $msg_seq = " $id " . $msg_seq;   # prepend new entry
            $msg_found{$id} = $line;
        }
        (close $fh && !$?)
            or return exit_info $!, $?, "$acid - $cmd";

        # fill curr_msgs if not done yet
        @curr_msgs = map { $msg_found{$_} } $msg_seq =~ /[^ ]+/g
            if $mode ne 'summary';

        # append data of this file to data of previous (= more recent) files
        @prev_msgs = (@prev_msgs, @curr_msgs);

        last if $mode eq 'latest' && $no_acid == 0;
    }

    return @prev_msgs;
}

#
# scan files with METAR/TAF data for $icao
#
sub search_files_metaf {
    my ($icao, $mode, $src, $hours, $type_metaf, $end_date) = @_;
    my ($url, @files, @prev_msgs, $fc_found, $ft_found);

    # NOAA/IWS cycle files: keyword SPECI not provided
    # (TODO?: should we search for SA instead?)
    $icao =~ s/ SP-$re_ICAO|SP-$re_ICAO ?//og
        if $src eq 'cycle_iws';

    return
        if $icao eq '';

    $hours = 7 unless $mode eq 'summary'; # overwrite for mode latest
    ($url, @files) =   $src eq 'local'     ? ('', "$CGI_DATA_DIR/metaf.txt")
                     : $src eq 'cod'       ? files_cod('sao',  $hours,$end_date)
                     : $src eq 'isu'       ? files_isu('sao',  $hours,$end_date)
                     : $src eq 'nws'       ? files_nws($type_metaf, $hours,
                                                                      $end_date)
                     : $src eq 'cycle_iws' ? files_cycle_iws($type_metaf, $icao,
                                                     $hours, gettime, $end_date)
                     :                       ();
    return ()
        unless defined $url;
    return ''
        if $src eq 'nws' && $url eq '';

    @prev_msgs = ();
    $fc_found = 0;
    $ft_found = 0;
    for my $ii (0 .. $#files) {
        my ($cmd, $format, $fh, @curr_msgs, %msg_found, $msg_seq);

        # if src_metaf=noaa, type_metaf=taf, mode=latest and all FC/FT found:
        #    skip staf/taf file
        if ($src eq 'cycle_iws' && $type_metaf eq 'taf' && $mode eq 'latest') {
            if (index($files[$ii], '/shorttaf/') > -1) {
                next
                    if $fc_found;
            } else {
                next
                    if $ft_found;
            }
        }

        if ($url) {
            add_searched_file "$url/$files[$ii]";
            $cmd = "$CURL $CURL_PROXY_OPTS $url/$files[$ii]";
            $format =   $src ne 'cycle_iws'
                      ? 'metaf'
                      : index($files[$ii], '/shorttaf/') > -1
                      ? 'staf'
                      : $type_metaf;
            start_file("${format}_$src", 1, 1);
            start_fh;
            pipe_open \$fh, 1,
                      split(' ', "$CURL $CURL_PROXY_OPTS"), "$url/$files[$ii]"
                or return '';
        } else {
            add_searched_file $files[$ii];
            $cmd = $files[$ii];
            open $fh, '<:bytes', $files[$ii] ## no critic (RequireBriefOpen)
                or return error "ERROR: open '$cmd' for $icao failed: $!.\n";
        }

        @curr_msgs = ();
        $msg_seq  = '';
        while (my $line = $url ? next_fh $fh : <$fh>) {
            my ($report_type, $report_type_allowed, $msg, $id);
            my ($report_type_id, $report_type_matches, $report_type_any);

            chomp $line
                unless $url;

            ($report_type, $msg, $id) = $line =~ /^($re_report_type_metaf) ($re_header_metaf($re_ICAO).*)/o;
            next
                unless defined $id;

            $report_type_id = "$report_type-$id";
            $report_type_matches = index($icao, $report_type_id) > -1;
            $report_type_any     = index(" $icao", " $id") > -1;

            # - skip if not required
            # - prepare check if message was in previous (= more recent) files
            #   for mode=latest
            if ($report_type_matches) {
                # skip if report_type_id was in previous files
                $report_type_allowed = $report_type;
            } elsif ($report_type_any) {
                # skip FC,FT if FC|FT was in previous files
                # skip SA    if SA    was in previous files
                # skip SP    if SA|SP was in previous files
                # no distinction by type_metaf required as long as files don't
                # have both TAF and METAR, or we have only one file (src=local)
                $report_type_allowed = $report_type eq 'SP' ? '(?:S.|F.)'
                                                            : '(?:SA|F.)';
            } else {
                next;
            }

            # looking for all messages (with correct id)?
            if ($mode eq 'summary') {
                # prepend unless it already exists
                unshift @curr_msgs, $line
                    unless grep { /^\Q$line\E$/ } (@prev_msgs, @curr_msgs);
                next;
            }

            # skip if message was in previous (= more recent) files
            next
                if grep { /^$report_type_allowed $re_header_metaf$id / } @prev_msgs;

            $msg_seq =~ s/ $report_type_id //        # remove entry if it exists
                if $msg_found{$report_type_id};
            $msg_seq = " $report_type_id " . $msg_seq;       # prepend new entry
            $msg_found{$report_type_id} = $line;             # save new value

            # remove SP (found in this file previously) if
            #   not required and SA just found
            # remove FC (found in this file previously) if
            #   not required and FT just found
            # remove FT (found in this file previously) if
            #   not required and FC just found
            for ({ SA => 'SP', FT => 'FC', FC => 'FT' }->{$report_type} // ()) {
                $_ .= "-$id";
                if ($msg_found{$_} && index($icao, $_) == -1) {
                    $msg_seq =~ s/ $_ //;
                    delete $msg_found{$_};
                }
            }
        }
        (close $fh && !$?)
            or return exit_info $!, $?, "$icao - $cmd";

        # fill curr_msgs if not done yet
        @curr_msgs = map { $msg_found{$_} } $msg_seq =~ /[^ ]+/g
            if $mode ne 'summary';

        # append data of this file to data of previous (= more recent) files
        @prev_msgs = (@prev_msgs, @curr_msgs);

        # check if all messages were found (unless this is the last file)
        if ($mode eq 'latest' && $ii < $#files) {
            # stop searching if found messages:
            # - with type: all requested, and
            # - without type: for type_metaf==metar: SA
            #                 for type_metaf==taf:   either FC or FT
            my ($all_found, $rt, $id);

            $all_found = 1;
            $fc_found = 1;
            $ft_found = 1;
            for (split / /, $icao) {
                ($rt, $id) =   length == 4
                             ? ($type_metaf eq 'metar' ? 'SA' : 'F.', $_)
                             : split /-/;
                next
                   if grep { /^$rt $re_header_metaf$id / } @prev_msgs;

                $all_found = 0;
                $fc_found = 0
                    if $rt eq 'F.' || $rt eq 'FC';
                $ft_found = 0
                    if $rt eq 'F.' || $rt eq 'FT';
            }
            last
                if $all_found;
        }
    }

    return @prev_msgs;
}

#
# query Ogimet
#
sub query_ogimet {
    my ($id, $mode, $which, $hours, $rep_type, $end_date) = @_;
    my ($data, $url, $cmd, $fh, $response);

    $hours = 7 unless $mode eq 'summary'; # overwrite for mode latest
    $end_date = gettime unless defined $end_date;
    $data =   '-Gdtipo=' . $rep_type . ' -dord=REV -dnil=SI -dfmt=txt'
            . fmtTime($end_date - $hours * 3600 + 60,
                       ' -dano=%d -dmes=%02d -dday=%02d -dhora=%02d -dmin=%02d')
            . fmtTime($end_date,
                  ' -danof=%d -dmesf=%02d -ddayf=%02d -dhoraf=%02d -dminf=%02d')
            . ' -d' . ($which eq 'synopsc2' ? 'estado' : 'lugar') . "=$id";
    $url = $data_url // 'http://www.ogimet.com/display_' . $which . '.php';
    $cmd = "$CURL $CURL_PROXY_OPTS $data $url";
    pipe_open \$fh, 1, split(' ', "$CURL $CURL_PROXY_OPTS $data"), $url
        or return '';

    {
        local $/;
        $response = <$fh>;
    }
    if (!close $fh || $?) {
        $response =~ s/[\n\r]+/ /g;
        return exit_info $!, $?, $cmd . ($response ? ":\n$response" : '.');
    }
    return error "ERROR: $1\n"
        if $response =~ /(.*quota.*reached.*)/i;
    return $response;
}

#
# query Ogimet for $msg_synop
#
sub query_synop_ogimet {
    my ($msg_synop, $mode, $type_synop, $hours, $end_date) = @_;
    my ($wmoid_uniq, $rep_type, $response, @msgs, %msg_found, $is_by_ctry);

    # this server has no BUFR
    $msg_synop =~ s/ IS[IMN]-[^ ]+|IS[IMN]-[^ ]+ ?//g;

    $wmoid_uniq = mk_uniq $msg_synop =~ /(?:$re_report_type_synop-)?([^ ]+)/og;
    return
        if $wmoid_uniq eq '';

    $wmoid_uniq =~ s/ /+/g;
    $is_by_ctry = $type_synop eq 'wmo' && $wmoid_uniq =~ /[A-Z]$/ ? 'c' : '';
    $rep_type =   $msg_synop =~ /^($re_report_type_synop)-[^ ]+(?: \1-[^ ]+)*$/o
                ? $1 : 'ALL';
    $response = query_ogimet $wmoid_uniq, $mode,
                ($type_synop eq 'ship' ? 'ships' : "synops${is_by_ctry}2"),
                $hours, $rep_type, $end_date;

    $msg_synop =~ s/ /|/g;

    $_ = (split /<[\/]?pre>/, $response)[1];     # extract answer
    return error "ERROR $wmoid_uniq: no data\n"
        unless defined;

    s/^#.*//mg;                                  # make comment lines empty
    s/\n+/\n/g;                                  # delete consecutive newlines
    s/\n +/ /g;                                  # delete newlines in messages
    s/=//g;                                      # delete EOM
    s/^.{13}//mg;                                # remove time stamps
    for (split /\n/) {                           # process each message
        my ($id, $report_time, $report_type);
        my ($report_type_id, $report_type_matches, $report_type_any);

        if (m{^AAXX $re_day($re_hour)[0134/] ([A-Z\d]{3,}) }o) {
            $report_time = $1;
            $id = $2;
        } elsif (m{^(?:BB|OO)XX ([A-Z\d]{3,}) $re_day($re_hour)[0134/] }o) {
            $id = $1;
            $report_time = $2;
        } else {
            next;
        }

        if ($report_time =~ /00|06|12|18/) {
            $report_type = 'SM';
        } elsif ($report_time =~ /03|09|15|21/) {
            $report_type = 'SI';
        } else {
            $report_type = 'SN';
        }

        $report_type_id = "$report_type-$id";
        if ($is_by_ctry) {
            $report_type_matches = index($msg_synop, $report_type) > -1;
            $report_type_any     = $msg_synop !~ /^$re_report_type_synop-/;
        } else {
            $report_type_matches = index("|$msg_synop|", "|$report_type_id|") > -1;
            $report_type_any     = index("|$msg_synop|", "|$id|") > -1;
        }

        # skip if not required
        next
            unless $report_type_matches || $report_type_any;

        if ($mode ne 'summary') {
            my $is_required;

            # request  : input             => output
            # id       : Sy-id,Sx-id,Sz-id => Sy-id
            # Sx-id    : Sy-id,Sx-id,Sz-id => Sx-id
            # Sx-id,id : Sx-id,Sy-id       => Sx-id
            # Sx-id,id : Sy-id,Sx-id       => Sy-id,Sx-id
            if ($report_type_matches && !exists $msg_found{$report_type_id}) {
                $msg_found{$report_type_id} = undef;
                $is_required = 1;
            }
            if ($report_type_any && !exists $msg_found{$id}) {
                $msg_found{$id} = undef;
                $is_required = 1;
            }
            next
                unless $is_required;
        }
        # include station
        push @msgs, "$report_type $id $_";
    }
    return @msgs;
}

#
# query Ogimet for $msg_buoy
#
sub query_buoy_ogimet {
    my ($msg_buoy, $mode, $hours, $end_date) = @_;
    my ($response, @msgs, %msg_found);

    # this server has no BUFR, and SS-X is the same as X
    $msg_buoy = mk_uniq $msg_buoy =~ /\G(?:IOB-[^ ]+ ?)*+(?:SS-)?([^ ]+) ?/g;
    return
        if $msg_buoy eq '';

    $msg_buoy =~ s/ /+/g;
    $response = query_ogimet $msg_buoy, $mode,
                             'buoys', $hours, 'ALL', $end_date;

    $_ = (split /<[\/]?pre>/, $response)[1];     # extract answer
    return error "ERROR $msg_buoy: no data\n"
        unless defined;

    s/^#.*//mg;                                  # make comment lines empty
    s/\n+/\n/g;                                  # delete consecutive newlines
    s/\n +/ /g;                                  # delete newlines in messages
    s/=//g;                                      # delete EOM
    s/^.{13}//mg;                                # remove time stamps
    for (split /\n/) {                           # process each message
        /^$re_header_buoy (\d{5}) /o
            or next;
        if ($mode eq 'summary' || !exists $msg_found{$1}) {
            # include station
            push @msgs, "SS $1 $_";
            $msg_found{$1} = undef;
        }
    }
    return @msgs;
}

#
# query NOAA AWC Text Data Server API for METAR or TAF for $icao
#
# - does not provide METAR/TAF with NIL
# - does not provide all data in NOAA/NWS
# - the same data is returned with /cgi-bin/data/{metar,taf}.php?ids=...
#   or /api/data/{metar,taf}?ids=...&date=... (and hours=... for METAR)
#   but only as text, TAF is splitted into multiple lines, only one TAF per
#   station, "TAF " is sometimes missing in message, SPECI is not marked
sub query_metaf_addsds_get {
    my ($icao, $type, $hours, $end_date) = @_;
    my ($data, $url, $cmd, $fh, $response);

    return '' if $icao eq '';

    $icao =~ s/ /,/g;

    $data = '-GdrequestType=retrieve -dformat=xml -ddataSource=';
    if ($type eq 'metar') {
        $data .= 'metars -dfields=raw_text,station_id,observation_time,metar_type';
    } else {
        $data .= 'tafs -dfields=raw_text,station_id,issue_time,valid_time_from,valid_time_to -dtimeType=issue';
    }

    $end_date = gettime unless defined $end_date;
    $data .=   ' -dstartTime=' . ($end_date - $hours * 3600)
             . ' -dendTime=' . $end_date
             . " -dstationString=$icao";

    $url = $data_url // 'https://aviationweather.gov/cgi-bin/data/dataserver.php';
    $cmd = "$CURL $CURL_PROXY_OPTS $data $url";
    pipe_open \$fh, 1, split(' ', "$CURL $CURL_PROXY_OPTS $data"), $url
        or return '';

    {
        local $/;
        $response = <$fh>;
    }
    $response =~ s/[\n\r]+/ /g;
    (close $fh && !$?)
        or return exit_info $!, $?, $cmd . ($response ? ":\n$response" : '.');
    return $response;
}

#
# query NOAA AWC Text Data Server API for $icao
#
sub query_metaf_addsds {
    my ($icao_metar, $icao_taf, $mode, $hours, $end_date, $msg_metaf) = @_;
    my ($response, %msg_found, @msgs, %metar, %taf);

    $hours = 7 unless $mode eq 'summary'; # overwrite for mode latest

    $response = query_metaf_addsds_get
                mk_uniq($icao_metar =~ /(?:$re_report_type_metaf-)?([^ ]+)/og),
                'metar', $hours, $end_date;

    $response =~ s/ +/ /g;
    %metar = ();
    %msg_found = ();
    # use closing tag for METAR, metar_type is sometimes missing
    for ($response =~ m{<((?:raw_text|station_id|observation_time|metar_type|/METAR)>[^<]+)}ig)
    {
        my $report_type_id;

        /^([^>]+).(.*)/;
        if ($1 ne '/METAR') {
            $metar{$1} = $2;
            next;
        }

        # check for missing fields
        $metar{metar_type} = $metar{metar_type} && $metar{metar_type} eq 'SPECI'
                             ? 'SP' : 'SA';
        if (scalar keys %metar != 4) {
            %metar = ();
            next;
        }
        $report_type_id = "$metar{metar_type}-$metar{station_id}";

        # skip if not required
        if  (   index(" $msg_metaf ", " $metar{station_id} ") == -1
             && index(" $msg_metaf ", " $report_type_id ") == -1)
        {
            %metar = ();
            next;
        }

        if ($mode ne 'summary') {
            # skip if found previously
            if (exists $msg_found{$report_type_id}) {
                %metar = ();
                next;
            }

            $msg_found{$report_type_id} = undef;

            # mark SP found if SA found and SP not required
            $msg_found{"SP-$metar{station_id}"} = undef
                if    $metar{metar_type} eq 'SA'
                   && index(" $msg_metaf ", " SP-$metar{station_id} ") == -1;
        }

        push @msgs, "@metar{qw(observation_time metar_type raw_text)}";
        %metar = ();
    }

    if ($response =~ /(<error>[^<]+)/ig) {
        /^[^>]+.(.*)/;
        return error "ERROR $icao_metar: $1\n";
    }

    $response = query_metaf_addsds_get
                mk_uniq($icao_taf =~ /(?:$re_report_type_metaf-)?([^ ]+)/og),
                'taf', $hours, $end_date;

    %taf = ();
    %msg_found = ();
    for ($response =~ /<((?:raw_text|station_id|issue_time|valid_time_from|valid_time_to)>[^<]+)/ig)
    {
        my $report_type_id;

        /^([^>]+).(.*)/
            or next;
        $taf{$1} = $2;
        next
            if $1 ne 'valid_time_to';

        # check for missing fields
        if (scalar keys %taf != 5) {
            %taf = ();
            next;
        }

        $taf{valid_time_from} = mkTime split /[-TZ:]/, $taf{valid_time_from};
        $taf{valid_time_to} =   mkTime split /[-TZ:]/, $taf{valid_time_to};
        $hours = ($taf{valid_time_to} - $taf{valid_time_from}) / 3600;
        $taf{msg_type} = $hours >= 12 ? 'FT' : 'FC';
        $report_type_id = "$taf{msg_type}-$taf{station_id}";

        # skip if not required
        if (   index(" $msg_metaf ", " $taf{station_id} ") == -1
            && index(" $msg_metaf ", " $report_type_id ") == -1)
        {
            %taf = ();
            next;
        }

        if ($mode ne 'summary') {
            # skip if found previously
            if (exists $msg_found{$report_type_id}) {
                %taf = ();
                next;
            }

            $msg_found{$report_type_id} = undef;

            # mark FT found if FC found (before FT) and FT not required
            # mark FC found if FT found (before FC) and FC not required
            for ({ FT => 'FC', FC => 'FT' }->{$taf{msg_type}} // ()) {
                $report_type_id = "$_-$taf{station_id}";
                $msg_found{$report_type_id} = undef
                    if index(" $msg_metaf ", " $report_type_id ") == -1;
            }
        }

        push @msgs, "@taf{qw(issue_time msg_type raw_text)}";
        %taf = ();
    }

    if ($response =~ /(<error>[^<]+)/ig) {
        /^[^>]+.(.*)/;
        return error "ERROR $icao_taf: $1";
    }

    # sort by date
    return map { substr $_, 21 } reverse sort @msgs;
}

#
# query Ogimet for $icao
#
sub query_metaf_ogimet {
    my ($icao, $mode, $hours, $end_date, $msg_metaf) = @_;
    my ($rep_type, $response, @msgs, @msgs2, %msg_found);

    $icao =~ s/ /+/g;
    $rep_type =   $msg_metaf =~ /^($re_report_type_metaf)-[^ ]+(?: \1-[^ ]+)*$/o
                ? $1 : 'ALL';
    # also returns TAF!
    $response = query_ogimet $icao, $mode, 'metars2', $hours, $rep_type, $end_date;

    $_ = (split /<[\/]?pre>/, $response)[1];  # extract answer
    return error "ERROR $icao: no data\n"
        unless defined;

    s/^#  (?=TAF|MET)/x            /mg;      # preserve comments w. message type
    s/^#.*//mg;                              # make comment lines empty
    s/\n+/\n/g;                              # delete consecutive newlines
    s/\n +/ /g;                              # delete newlines in messages
    s/=//g;                                  # delete EOM

    # insert message type after time stamp
    @msgs = ();
    for (split /^x/m) {
        if (/^ +METAR/) {
            s/.*\n//;
            for (split /\n/) {
                if (/^.{12} ($re_header_metaf)$re_header_sao/o) {
                    # prepend 'C' to 3-letter station code
                    substr $_, 13 + length $1, 0, 'C'
                        if length $2 == 3;
                    substr $_, 12, 0,
                             ' ' . { SA => 'SA', SP => 'SP', RS => 'SP' }->{$3};
                } elsif (/^.{12} ($re_header_metaf)$re_ICAO /o) {
                    substr $_, 12, 0, index($1, 'SPECI') > -1 ? ' SP' : ' SA';
                } else {
                    next;
                }
                push @msgs, $_;
            }
        } elsif (/^ +TAF LAR/) {
            s/.*\n//;
            s/^.{12}\K/ FT/mg;
            push @msgs, split /\n/;
        } elsif (/^ +TAF COR/) {
            s/.*\n//;
            s/^.{12}\K/ FC/mg;
            push @msgs, split /\n/;
        }
    }

    # copy required messages to msgs2
    @msgs2 = ();
    for (reverse sort @msgs) {
        my ($rt, $id, $report_type_id);

        s/^.{13}//;                             # remove time stamps

        /^($re_report_type_metaf) $re_header_metaf($re_ICAO) /o;
        ($rt, $id) = ($1, $2);
        $report_type_id = "$rt-$id";

        # skip if not required
        next
            if    index(" $msg_metaf ", " $id ") == -1
               && index(" $msg_metaf ", " $report_type_id ") == -1;

        if ($mode ne 'summary') {
            # skip if found previously
            next
                if exists $msg_found{$report_type_id};

            $msg_found{$report_type_id} = undef;

            # mark SP found if SA found and SP not required
            # mark FT found if FC found (before FT) and FT not required
            # mark FC found if FT found (before FC) and FC not required
            for ({ SA => 'SP', FT => 'FC', FC => 'FT' }->{$rt} // ()) {
                $report_type_id = "$_-$id";
                $msg_found{$report_type_id} = undef
                    if index(" $msg_metaf ", " $report_type_id ") == -1;
            }
        }
        push @msgs2, $_;
    }
    return @msgs2;
}

{
    my %stationInfo;

    sub query_db_station {
        my ($db, $station, $station_type) = @_;

        if (!exists $stationInfo{$station_type}{$station}) {
            my ($is_fixed, $prepared, $data);

            $is_fixed = $station_type eq 'icao' || $station_type eq 'wmo';
            if ($is_fixed) {
                $prepared = $db->prepare(
                      'SELECT name, ctry, loc'
                    .  ' FROM station_fixed'
                    . ' WHERE ' . $station_type . ' = "' . $station . '"');
            } else {
                $prepared = $db->prepare(
                      'SELECT name, ctry'
                    .  ' FROM station_coords'
                   # some buoys report BBXX: lookup ship call sign in buoy table
                    . ' WHERE stationType ' . { ship  => q{IN('S','B')},
                                                mobil => q{='M'},
                                                buoy  => q{='B'},
                                                ac    => q{='A'},
                                              }->{$station_type}
                    .   ' AND station = "' . $station . '"');
            }
            $prepared->execute();
            $data = $prepared->fetchrow_hashref();
            $prepared->finish;
            $stationInfo{$station_type}{$station} = $prepared->rows == 1
                ? { ''   => $data->{name},
                    ctry => $data->{ctry},
                    ($is_fixed ? (loc => $data->{loc}) : ())
                  }
                : {};
        }
        return $stationInfo{$station_type}{$station};
    }
}

sub query_db_msgs {
    my ($db, $station_uniq, $mode, $table,
        $hours, $rep_type, $end_date, $station, $station_type) = @_;
    my ($from, $till, $prepared, @msgs, %msg_found);

    $station_uniq =~ s/^/'/;
    $station_uniq =~ s/$/'/;
    $station_uniq =~ s/ /','/g;
    if (defined $end_date) {
        $from = $end_date - $hours * 3600;
        $till = $end_date;
    } else {
        $from = "(SELECT MAX(time) FROM ${table}_get) - $hours * 3600";
        $till = "(SELECT MAX(time) FROM ${table}_get)";
    }
    if ($table eq 'synop_fixed' && $station_type) {
        $prepared = $db->prepare(
              'SELECT repType, sb.country || "-" || sb.natStation station,'
            . '       r.msg msg, r.obsTime time'
            . '  FROM synop_fixed r'
            .  ' LEFT JOIN station_bufr sb'
            .    ' ON r.s_b_id != -1'
            .   ' AND r.s_b_id = sb.id'
            . " WHERE sb.country || '-' || sb.natStation IN ($station_uniq)"
            .   " AND r.obsTime >  $from"
            .   " AND r.obsTime <= $till"
            . ' ORDER BY 4 DESC, 1');
    } else {
        $prepared = $db->prepare(
              'SELECT repType, station, msg, time'
            .  " FROM ${table}_get"
            . " WHERE station IN ($station_uniq)"
            .   " AND time >  $from"
            .   " AND time <= $till"
            .     (defined $rep_type ? " AND repType = '$rep_type'" : '')
            .     ($station_type ? " AND stationType = '$station_type'" : '')
            . ($table eq 'metar'
               ?    ' UNION ALL'
                 . ' SELECT repType, station, msg, time'
                 .   ' FROM taf_get'
                 .  " WHERE station IN ($station_uniq)"
                 .    " AND time >  $from"
                 .    " AND time <= $till"
                 .      (defined $rep_type ? " AND repType = '$rep_type'" : '')
               : '')
            . ' ORDER BY 4 DESC, 1');
    }
    $prepared->execute();
    while (my $data = $prepared->fetchrow_hashref()) {
        if ($table eq 'metar' || $table eq 'taf') {
            my $report_type_id;

            $report_type_id = "$data->{repType}-$data->{station}";

            # skip if not required
            next
                if    index(" $station ", " $data->{station} ") == -1
                   && index(" $station ", " $report_type_id ") == -1;

            if ($mode ne 'summary') {
                # skip if found previously
                next
                    if exists $msg_found{$report_type_id};

                $msg_found{$report_type_id} = undef;

                # mark SP found if SA found and SP not required
                # mark FT found if FC found (before FT) and FT not required
                # mark FC found if FT found (before FC) and FC not required
                for ({ SA => 'SP',
                       FT => 'FC', FC => 'FT' }->{$data->{repType}} // ())
                {
                    $report_type_id = "$_-$data->{station}";
                    $msg_found{$report_type_id} = undef
                        if index(" $station ", " $report_type_id ") == -1;
                }
            }
            push @msgs, $data->{repType} . ' ' . $data->{msg};
        } else {
            my ($report_type_id, $report_type_matches, $report_type_any);

            $report_type_id = "$data->{repType}-$data->{station}";
            $report_type_matches = index(" $station ", " $report_type_id ") > -1;
            $report_type_any     = index(" $station ", " $data->{station} ") > -1;

            # skip if not required
            next
                unless $report_type_matches || $report_type_any;

            if ($mode ne 'summary') {
                my $is_required;

                if ($report_type_matches && !exists $msg_found{$report_type_id})
                {
                    $msg_found{$report_type_id} = undef;
                    $is_required = 1;
                }
                if ($report_type_any && !exists $msg_found{$data->{station}}) {
                    $msg_found{$data->{station}} = undef;
                    $is_required = 1;
                }
                next
                    unless $is_required;
            }

            # include station
            push @msgs,
                 $data->{repType} . ' ' . $data->{station} . ' ' . $data->{msg};
        }
    }
    $prepared->finish;
    return @msgs;
}

sub clean_msg {
    $_ = param shift;
    return ''
        unless defined;

    $_ = uc(substr $_, 0, 3000);
    s/[^ -~]/?/g; # replace non-ASCII characters
    s/  +/ /g;
    s/ $//;
    return $_;
}

sub usage {
    pod2usage -exitval => 'NOEXIT', -output => \*STDERR, -noperldoc => 1;
    error <<"EOF";
Options may be merged together.  -- stops processing of options.
Space is not required between options and their arguments.

For more details run
        perldoc -F $0
EOF
    exit 1;
}

sub HELP_MESSAGE { usage; return; } # for Getopt::Std

# don't run if loaded as a library
return 1
    if caller;

{
    my %opts;

    # get and check options
    $Getopt::Std::STANDARD_HELP_VERSION = 1;
    usage unless getopts 'v', \%opts;
    if ($opts{v}) {
        print "metaf.pl version $VERSION\n" or exit 1;
        exit 0;
    }
}

# capture STDERR unless called from command line
{
    my $stderr = '';

    if (!$is_cmdline) {
        close STDERR;
        open STDERR, '>', \$stderr;
    }

    sub get_stderr {
        my ($format) = @_;

        return   !defined $format ? $stderr
               : $format eq 'xml' ? ($stderr =~ s/</&lt;/gr)
               : '<pre><b>' . ($stderr =~ s/</&lt;/gr) . '</b></pre>';
    }
}

=pod

If there are any CGI errors the type of error is reported and the script
terminates.

=cut

charset('UTF-8');

error cgi_error()
    if cgi_error();

sub check_param {
    my ($name, $default, @allowed) = @_;
    my $val;

    $val = param $name;
    return   defined $val && { map { $_ => 1 } @allowed }->{$val} ? $val
           :                                                        $default;
}

sub mk_radio {
    my ($name, $value, $checked_val, $check_local) = @_;

    return ("${name}_$value" =>
            "<input type=radio name=$name value=$value"
           . ($checked_val eq $value ? ' checked' : '')
           . ($check_local && $checked_val eq 'local'
                ? ' disabled' : '' )
           . '>');
}

#
# determine language: check CGI parameter, then browser header. default: de
#
sub check_lang {
    my @allowed = @_; # full ('en-GB') or short ('en') names
    my ($lang, $first_short);

    # if parameter is set: check full match, then short name match
    $lang = param 'lang';
    if (defined $lang) {
        return $lang
            if { map { $_ => 1 } @allowed }->{$lang};
        $lang =
              { map { substr($_, 0, 2) => $_ } @allowed }->{substr $lang, 0, 2};
        return $lang || $allowed[0]; # TODO: or continue with browser header?
    }

    # browser header: check full match while remembering first short name match
    $first_short = '';
    for (split /,/, lc(http('Accept-Language') || '')) {
        s/ //g;
        s/;.*//;
        return $_
            if { map { $_ => 1 } @allowed }->{$_};
        $first_short =
                  { map { substr($_, 0, 2) => $_ } @allowed }->{substr $_, 0, 2}
            unless $first_short;
    }
    return $first_short || $allowed[0];
}
my $lang = check_lang qw(de en es);

my $is_gui = defined param 'gui';

sub readline_data_source {
    my ($fh, $expect) = @_;
    my ($line, $expect_len);

    return
        unless defined ($line = <$fh>);

    chomp $line;
    return $line
        unless defined $expect;

    $expect_len = length $expect;

    return substr $line, $expect_len
        if substr($line, 0, $expect_len) eq $expect;
    return;
}

#
# determine format: check CGI parameter, then browser header. default: html
#
my $format = check_param qw(format html text xml msgs);
if ($format eq 'xml') {
    my $accept = http('Accept');
    if ($accept) {
        my $accept_all = ",$accept," =~ m{[ ,]\*/\*[;,]};
        my $accept_xml = ",$accept," =~ m{[ ,](?:text|application)/(?:[a-z+-]*xml[a-z+-]*|\*)[;,]};
        my $accept_xsl = ",$accept," =~ m{[ ,](?:text|application)/(?:[a-z+-]*x[sm]l[a-z+-]*|\*)[;,]};
        $format = 'html' unless $accept_all || ($accept_xml && $accept_xsl);
    }
}

my $need_html = !$is_cmdline || $format eq 'html';

my $unit_temp = check_param qw(unit_temp C F);

=pod

The parameters C<msg_>* are truncated to 3000 characters.
Letters are converted to upper case. Any characters that are not allowed in a
message are removed.

=cut

my $msg_metaf = clean_msg 'msg_metaf';
my $msg_synop = clean_msg 'msg_synop';
my $msg_buoy  = clean_msg 'msg_buoy';
my $msg_amdar = clean_msg 'msg_amdar';

#
# determine message types and if messages need to be fetched
#
my $type_metaf = check_param qw(type_metaf metar taf icao);
my $type_synop = check_param qw(type_synop synop wmo nat ship mobil);
my $type_buoy  = check_param qw(type_buoy  buoy wmo);
my $type_amdar = check_param qw(type_amdar amdar ac);
my ($fetch_metaf, $fetch_synop, $fetch_buoy, $fetch_amdar, $db);

if ($msg_metaf ne '') {
    if ($msg_metaf =~ /^(?:METAR|SPECI) /) {
        $type_metaf = 'metar';
    } elsif ($msg_metaf =~ /^TAF /) {
        $type_metaf = 'taf';
    } elsif ($type_metaf eq 'icao') {
        if (" $msg_metaf" !~ /^(?: (?:$re_report_type_metaf-)?$re_ICAO)+$/o) {
            $type_metaf = 'metar';
        } else {
            $fetch_metaf = 1;
        }
    }
}

if ($msg_synop ne '' && $type_synop ne 'synop') {
    # test for start of SYNOP message, it also matches a list of identifiers
    if (   $msg_synop =~ /^(?:AA|BB|OO)XX /
        || (   " $msg_synop" !~ /^(?: (?:$re_report_type_synop-)?[A-Z\d]{3,})+$/o
            && " $msg_synop" !~ /^(?: (?:$re_report_type_synop-)?\d+-\d+)+$/o
            && $msg_synop !~ /^(?:$re_report_type_synop-)?[A-Z]+$/o))
    {
        $type_synop = 'synop';
    } else {
        $fetch_synop = 1;
    }
}

if ($msg_buoy ne '' && $type_buoy ne 'buoy') {
    if (" $msg_buoy" !~ /^(?: (?:$re_report_type_buoy-)?\d{5,})+$/o) {
        $type_buoy = 'buoy';
    } else {
        $fetch_buoy = 1;
    }
}

if ($msg_amdar ne '' && $type_amdar ne 'amdar') {
    # test for start of AMDAR message, it also matches a list of identifiers
    if (   $msg_amdar =~ m{^$re_header_amdar (?:[A-Z]{3}|///) }o
        || " $msg_amdar" !~ /^(?: (?:$re_report_type_amdar-)?$re_aircraft)+$/o)
    {
        $type_amdar = 'amdar';
    } else {
        $fetch_amdar = 1;
    }
}

#
# if data needs to be fetched from a DB: try to open it
# (avoid lookup of DB unless required)
# hide errors if used as CGI and the use of DB is not forced
#
if ($fetch_metaf || $fetch_synop || $fetch_buoy || $fetch_amdar) {
    my $had_dbi;

    for (split / /, $DB_DATA_SOURCES) {
        my ($rc, $is_forced, $dsn, $user, $pass);

        if (!$had_dbi) {
            $rc = eval { require DBI; };
            if (!$rc || $@) {
                error $@
                    if $is_cmdline || $DB_DATA_SOURCES =~ m{/f\b};
                last;
            }
            $had_dbi = 1;
        }

        $is_forced = s{/f$}{};

        if ($_ eq 'sqlite') {
            # avoid creating the file
            if (!-f "$CGI_DATA_DIR/$_.db") {
                error "ERROR: DB '$_' does not exist.\n"
                    if $is_cmdline || $is_forced;
                $is_forced ? last : next;
            }
            $dsn = "DBI:SQLite:$CGI_DATA_DIR/sqlite.db";
        } else {
            if (open my $fh, '<:bytes', "$CGI_DATA_DIR/$_.db") {
                $dsn = readline_data_source $fh;
                $user = readline_data_source $fh, 'user='
                    if defined $dsn;
                $pass = readline_data_source $fh, 'pass='
                    if defined $user;
                close $fh;
                ($is_forced ? last : next)
                    unless defined $dsn;
            } else {
                error "ERROR: open DB '$_' failed: $!.\n"
                    if $is_cmdline || $is_forced;
                $is_forced ? last : next;
            }
        }

        $rc = eval {
            $db = DBI->connect($dsn, $user, $pass,
                         { AutoCommit => 1, RaiseError => 1, PrintError => 0 });
            1;
        };
        last
            if $rc && !$@;
        error "ERROR: connect to DB '$_' failed" . ($is_cmdline ? ": $@" : ".\n")
            if $is_cmdline || $is_forced;
        last
            if $is_forced;
    }
}

#
# determine src_synop, src_metaf, src_buoy, src_amdar
# set src_* to 'local' if data needs to be fetched and local data is available
# (src_* will not be 'local' if local files exist but are not needed)
#
my ($src_metaf, $src_synop, $src_buoy, $src_amdar);

if ($db || $DB_DATA_SOURCES =~ m{/f\b}) {
    $src_metaf = $src_synop = $src_buoy = $src_amdar = 'local';
} else {
    $src_metaf =   $fetch_metaf && -f "$CGI_DATA_DIR/metaf.txt"
                 ? 'local' : check_param qw(src_metaf noaa nws addsds
                                                      ogimet cod isu);
    $src_synop =   $fetch_synop && -f "$CGI_DATA_DIR/synop.txt"
                 ? 'local' : check_param qw(src_synop nws ogimet cod isu);
    $src_buoy  =   $fetch_buoy && -f "$CGI_DATA_DIR/buoy.txt"
                 ? 'local' : check_param qw(src_buoy  nws ogimet isu);
    $src_amdar =   $fetch_amdar && -f "$CGI_DATA_DIR/amdar.txt"
                 ? 'local' : check_param qw(src_amdar nws isu);
}

#
# determine mode: check CGI parameter. default: latest
#
my $mode = check_param qw(mode latest summary);

#
# determine number of hours (for mode summary): check CGI parameter. default: 24
#
my $hours = param('hours') || '';
$hours = 24
    unless $hours =~ /^(?:[1-9]\d?|0\.\d+)$/ && $hours <= 72;

#
# determine end date (for some data sources): check CGI parameter. default: now
#
my $end_date;
my $end_date_str = param 'end_date';
if (   defined $end_date_str
    && $end_date_str =~ /^(1999|20\d\d)-(0[1-9]|1[012])-($re_day) ($re_hour):($re_min)$/o)
{
    $end_date = mkTime $1, $2, $3, $4, $5, 0;
} else {
    $end_date = undef;
    $end_date_str = '';
}

#
# log CGI request unless called from the command line
#
if (!$is_cmdline && open my $log, '>>', $CGI_LOG_FILE) { ## no critic (RequireBriefOpen)
    # garble IP address, but still allow to group the requests
    my $addr;

    if (defined http('DNT') && substr(http('DNT'), 0, 1) eq '1') {
        $addr = 'xx.xx.xx.xx';
    } else {
        $addr = $ENV{'HTTP_X_REMOTE_ADDR'} || remote_addr();
        $addr =~ s/\.\d+/.x/;
        $addr =~ s/\d+$/x/;
    }
    print {$log} gmtime()
              . " $addr "
              . request_method()
              . " $type_metaf $type_synop $type_buoy $type_amdar $lang $unit_temp $format $src_metaf $src_synop $src_buoy $src_amdar $mode $hours $end_date_str"
              . ($is_gui ? ' gui' : '')
              . " '$msg_metaf' '$msg_synop' '$msg_buoy' '$msg_amdar'\n";
    close $log;
}

#
# translated strings
#
my $trans = {
    de => {
        title => 'metaf2xml: analysiere und dekodiere METAR, TAF, SYNOP, BUOY und AMDAR',
        infosrc => 'Eine gute Quelle für aktuelle METAR- und TAF-Meldungen ist',
        infofetch => 'Wird eine Option mit "Id(s)" gewählt und eine oder mehrere Id(s) angegeben, werden Meldungen von der ausgewählten Quelle (oder, falls diese Option gesperrt ist, von dem Server, von dem diese Web-Seite kommt) geholt.',
        options => 'Optionen',
        example => 'Beispiele',
        type1 => 'METAR-Meldung oder',
        type2 => 'TAF-Meldung oder',
        type3 => 'Meldung oder',
        type4 => 'ICAO Id(s)',
        type5 => 'WMO Id(s)',
        type7 => 'WMO Id(s) oder',
        type8 => 'SHIP Id(s) oder',
        type9 => 'MOBIL Id(s)',
        type11 => 'Flugzeug Id(s)',
        source => 'hole Meldungen von',
        source1 => 'NOAA/IWS',
        source3 => 'NOAA/API',
        source4 => 'NOAA/NWS',
        source5 => 'College of DuPage',
        source6 => 'Ogimet',
        language => 'Sprache',
        unit_temp => 'Temperatur in',
        format => 'gib aus als',
        format1 => 'HTML',
        format2 => 'Text',
        format3 => 'XML (+XSL -> HTML)',
        mode => 'hole/zeige',
        mode1 => 'letzte/detailliert',
        mode2 => 'alle/Übersicht seit',
        hours => 'Stunde(n)',
        decode => 'Übersetzen',
    },
    en => {
        title => 'metaf2xml: parse and decode METAR, TAF, SYNOP, BUOY and AMDAR',
        infosrc => 'A good source for up-to-date METAR and TAF messages is',
        infofetch => 'If an option with "id(s)" is chosen and one or more id(s) are given, messages will be fetched from the selected source (or if this option is disabled from the server where this web page comes from).',
        options => 'Options',
        example => 'examples',
        type1 => 'METAR message or',
        type2 => 'TAF message or',
        type3 => 'message or',
        type4 => 'ICAO id(s)',
        type5 => 'WMO id(s)',
        type7 => 'WMO id(s) or',
        type8 => 'SHIP id(s) or',
        type9 => 'MOBIL id(s)',
        type11 => 'aircraft id(s)',
        source => 'fetch messages from',
        source1 => 'NOAA/IWS',
        source3 => 'NOAA/API',
        source4 => 'NOAA/NWS',
        source5 => 'College of DuPage',
        source6 => 'Ogimet',
        language => 'language',
        unit_temp => 'temperature in',
        format => 'output as',
        format1 => 'HTML',
        format2 => 'text',
        format3 => 'XML (+XSL -> HTML)',
        mode => 'fetch/display',
        mode1 => 'latest/detailed',
        mode2 => 'all/summary since',
        hours => 'hour(s)',
        decode => 'decode',
    },
    es => {
        title => 'metaf2xml: analiza y decodifica METAR, TAF, SYNOP y BUOY',
        infosrc => 'Una buena fuente por mensajes actuales de METAR y TAF es',
        infofetch => 'Si es elegido una opción con "código(s)" y son dado uno o más código(s), los mensajes serán traído de la fuente elegida (o si esta opción es minusválida del servidor de donde viene esta página web).',
        options => 'Opciones',
        example => 'ejemplos',
        type1 => 'mensaje METAR o',
        type2 => 'mensaje TAF o',
        type3 => 'mensaje o',
        type4 => 'código(s) OACI',
        type5 => 'código(s) OMM',
        type7 => 'código(s) OMM o',
        type8 => 'código(s) SHIP o',
        type9 => 'código(s) MOBIL',
        type11 => 'código(s) aeronave',
        source => 'consultar mensajes de',
        source1 => 'NOAA/IWS',
        source3 => 'NOAA/API',
        source4 => 'NOAA/NWS',
        source5 => 'College of DuPage',
        source6 => 'Ogimet',
        language => 'lengua',
        unit_temp => 'temperatura en',
        format => 'presentar como',
        format1 => 'HTML',
        format2 => 'texto',
        format3 => 'XML (+XSL -> HTML)',
        mode => 'consultar/presentar',
        mode1 => 'último/detallado',
        mode2 => 'todos/resumen desde',
        hours => 'hora(s)',
        decode => 'traducir',
    },
}->{$lang};

#
# save message parameter
#
my $msg_metaf_orig = $msg_metaf;
my $msg_synop_orig = $msg_synop;
my $msg_buoy_orig  = $msg_buoy;
my $msg_amdar_orig = $msg_amdar;

sub mk_icao_uniq_metar { return mk_uniq " $_[0]" =~ / ((?:S[AP]-)?....\b)/g; }
sub mk_icao_uniq_taf   { return mk_uniq " $_[0]" =~ / ((?:F[CT]-)?....\b)/g; }

local $SIG{PIPE} = sub {
    error "ERROR: Caught SIGPIPE while reading from $CURL.\n";
};

#
# process METAR/TAF messages if they need to be fetched
#
if ($fetch_metaf) {
    my ($icao_uniq, @metaf);

    $icao_uniq = mk_uniq $msg_metaf =~ /(?:$re_report_type_metaf-)?([^ ]{4})/og;

    if ($src_metaf eq 'noaa') {
        # SP is filtered out by search_files_metaf()
        @metaf = (search_files_metaf(mk_icao_uniq_metar($msg_metaf), $mode,
                                     'cycle_iws', $hours, 'metar', $end_date),
                  search_files_metaf(mk_icao_uniq_taf($msg_metaf), $mode,
                                     'cycle_iws', $hours, 'taf', $end_date));
    } elsif ($src_metaf eq 'addsds') {
        @metaf = query_metaf_addsds(mk_icao_uniq_metar($msg_metaf),
                                    mk_icao_uniq_taf($msg_metaf),
                                    $mode, $hours, $end_date, $msg_metaf);
    } elsif ($src_metaf eq 'ogimet') {
        @metaf = query_metaf_ogimet $icao_uniq, $mode, $hours,
                                    $end_date, $msg_metaf;
    } elsif ({ cod => 1, isu => 1 }->{$src_metaf}) {
        @metaf = search_files_metaf mk_icao_uniq_metar($msg_metaf),
                                    $mode, $src_metaf, $hours, 'metar',
                                    $end_date;
    } elsif ($src_metaf eq 'nws') {
        @metaf = (search_files_metaf(mk_icao_uniq_metar($msg_metaf),
                                     $mode, 'nws', $hours, 'metar', $end_date),
                  search_files_metaf(mk_icao_uniq_taf($msg_metaf),
                                     $mode, 'nws', $hours, 'taf', $end_date));
    } elsif ($db) {
        my $rep_type;

        $rep_type = $1
            if $msg_metaf =~ /^($re_report_type_metaf)-[^ ]+(?: \1-[^ ]+)*$/o;
        @metaf = query_db_msgs($db, $icao_uniq, $mode, 'metar',
                               $hours, $rep_type, $end_date, $msg_metaf);
    } elsif ($DB_DATA_SOURCES !~ m{/f\b}) { # local
        @metaf = search_files_metaf $msg_metaf, $mode, $src_metaf, $hours;
    }

    $msg_metaf = join "\n",
        map {
            my $id = $_;
            grep { /^$re_report_type_metaf $re_header_metaf$id / } @metaf;
        } split / /, $icao_uniq;
    if ($msg_metaf) {
        $msg_metaf =~ s/^SA (?:METAR )?//mg;
        $msg_metaf =~ s/^SP (?:SPECI )?/SPECI /mg;
        $msg_metaf =~ s/^F[CT] (?:TAF )*/TAF /mg;
    }
}

# process SYNOP messages if they need to be fetched
if ($fetch_synop) {
    my ($wmoid_uniq, @synop);

    $wmoid_uniq = mk_uniq $msg_synop =~ /(?:$re_report_type_synop-)?([^ ]+)/og;
    if ($src_synop eq 'ogimet' && $type_synop ne 'nat') {
        @synop = query_synop_ogimet $msg_synop, $mode, $type_synop,
                                    $hours, $end_date;
    } elsif ($db) {
        my $rep_type;

        $rep_type = $1
            if $msg_synop =~ /^($re_report_type_synop)-[^ ]+(?: \1-[^ ]+)*$/o;
        if ($type_synop eq 'wmo') {
            @synop = query_db_msgs $db, $wmoid_uniq, $mode, 'synop_fixed',
                                   $hours, $rep_type, $end_date, $msg_synop;
        } elsif ($type_synop eq 'nat') {
            @synop = query_db_msgs $db, $wmoid_uniq, $mode, 'synop_fixed',
                                   $hours, $rep_type, $end_date, $msg_synop,'N';
        } elsif ($type_synop eq 'ship') {
            @synop = query_db_msgs $db, $wmoid_uniq, $mode, 'synop_coords',
                                   $hours, $rep_type, $end_date, $msg_synop,'S';
        } else {
            @synop = query_db_msgs $db, $wmoid_uniq, $mode, 'synop_coords',
                                   $hours, $rep_type, $end_date, $msg_synop,'M';
        }
    } elsif (   $type_synop ne 'nat'          # cod, isu, nws, local
             && $DB_DATA_SOURCES !~ m{/f\b})
    {
        @synop = search_files_synop $msg_synop, $mode, $src_synop, $hours,
                                    $type_synop, $end_date;
    }

    if ($wmoid_uniq =~ /[A-Z]$/) {
        $msg_synop = join "\n", @synop;
    } else {
        $msg_synop = join "\n",
            map {
                my $id = $_;
                grep { /^$re_report_type_synop $id / } @synop;
            } split / /, $wmoid_uniq;
    }
    $msg_synop =~ s/^$re_report_type_synop [^ ]+ //omg
        if $msg_synop;
}

# process BUOY messages if they need to be fetched
if ($fetch_buoy) {
    my ($wmoid_uniq, @buoy);

    $wmoid_uniq = mk_uniq $msg_buoy =~ /(?:$re_report_type_buoy-)?([^ ]+)/og;
    if ($src_buoy eq 'ogimet') {
        @buoy = query_buoy_ogimet $msg_buoy, $mode, $hours, $end_date;
    } elsif ($db) {
        my $rep_type;

        $rep_type = $1
            if $msg_buoy =~ /^($re_report_type_buoy)-[^ ]+(?: \1-[^ ]+)*$/o;
        @buoy = query_db_msgs $db, $wmoid_uniq, $mode, 'buoy',
                              $hours, $rep_type, $end_date, $msg_buoy;
    } elsif ($DB_DATA_SOURCES !~ m{/f\b}) { # nws, isu, local
        @buoy = search_files_buoy $msg_buoy, $mode, $src_buoy, $hours,
                                  $end_date;
    }

    $msg_buoy = join "\n",
        map {
            my $id = $_;
            grep { /^$re_report_type_buoy $id / } @buoy;
        } split / /, $wmoid_uniq;
    $msg_buoy =~ s/^$re_report_type_buoy \d{5,} //omg
        if $msg_buoy;
}

# process AMDAR messages if they need to be fetched
if ($fetch_amdar) {
    my ($acid_uniq, @amdar);

    $acid_uniq = mk_uniq $msg_amdar =~ /(?:$re_report_type_amdar-)?([^ ]+)/og;
    if ($db) {
        my $rep_type;

        $rep_type = $1
            if $msg_amdar =~ /^($re_report_type_amdar)-[^ ]+(?: \1-[^ ]+)*$/o;
        @amdar = query_db_msgs $db, $acid_uniq, $mode, 'amdar',
                               $hours, $rep_type, $end_date, $msg_amdar;
    } elsif ($DB_DATA_SOURCES !~ m{/f\b}) { # nws, isu, local
        @amdar = search_files_amdar $msg_amdar, $mode, $src_amdar, $hours,
                                    $end_date;
    }

    $msg_amdar = join "\n",
        map {
            my $id = $_;
            grep { /^$re_report_type_amdar $id / } @amdar;
        } split / /, $acid_uniq;
    $msg_amdar =~ s/^$re_report_type_amdar $re_aircraft //omg
        if $msg_amdar;
}

my $msg = join "\n", $msg_metaf, $msg_synop, $msg_buoy, $msg_amdar, '';
$msg =~ s/[^\n -~]/?/g; # replace non-ASCII characters
$msg =~ s/  +/ /g;
$msg =~ s/ $//mg;
$msg =~ s/^\n//mg;

if ($format eq 'msgs') {
    print header('text/plain')
        unless $is_cmdline;
    print $msg . get_stderr;
    exit 0;
}

$metaf2xml::parser::_mk_station_info = sub { query_db_station $db, @_; }
    if $db && !$is_gui;

if ($format eq 'xml') {
    my $xsl = $mode eq 'summary' ? 'sum-ui' : 'fullhtml';
    start_xml { o => '-',
                $is_cmdline ? () : (I => header('application/xml')),
                S => "../metaf-$xsl.xsl",
                O => "$type_metaf $type_synop $type_buoy $type_amdar $lang $unit_temp $format $src_metaf $src_synop $src_buoy $src_amdar $mode $hours $msg_metaf_orig  $msg_synop_orig  $msg_buoy_orig  $msg_amdar_orig"
    };
    for (split /\n/, $msg . get_stderr) {
        parse_report $_, $type_metaf eq 'taf' ? 'TAF' : 'METAR';
    }
    finish;
    exit 0;
}

print header()
    unless $is_cmdline;

print <<"EOF"
<!DOCTYPE html>
<html lang=$lang>
<head>
<meta content="text/html; charset=UTF-8" http-equiv=Content-Type>
<title>$trans->{title}</title>
<meta name=keywords content=metaf2xml,parse,decode,METAR,SPECI,SAO,TAF,SYNOP,BUOY,AMDAR,BUFR,XML,aviation,weather,report,forecast>
<meta name=robots content=noindex,nofollow>
<meta name=viewport content="width=device-width, initial-scale=1">
<meta name=format-detection content=telephone=no>
<!-- link href=../images/metaf2xml.ico rel="shortcut icon" type=image/x-icon -->
<link rel=stylesheet type=text/css href=../metaf2xml.css>
<body>
<header>
<h1>$trans->{title}</h1>
EOF
    if $need_html;

if ($is_cmdline) {
    print "</header>\n"
        if $need_html;
} else {
    my (%radios, @m2x_text, %content);

    $msg_metaf_orig =~ s/"/&quot;/g;
    $msg_synop_orig =~ s/"/&quot;/g;
    $msg_buoy_orig  =~ s/"/&quot;/g;
    $msg_amdar_orig =~ s/"/&quot;/g;

    %radios = (
        (map { mk_radio('lang',       $_, $lang      )} qw(de en es)),
        (map { mk_radio('unit_temp',  $_, $unit_temp )} qw(C F)),
        (map { mk_radio('format',     $_, $format    )} qw(html text xml msgs)),
        (map { mk_radio('mode',       $_, $mode      )} qw(latest summary)),
        (map { mk_radio('type_metaf', $_, $type_metaf)} qw(metar taf icao)),
        (map { mk_radio('src_metaf',  $_, $src_metaf, 1)}
                                          qw(noaa nws addsds ogimet)),
        (map { mk_radio('type_synop', $_, $type_synop)}
                                          qw(synop wmo ship mobil)),
        (map { mk_radio('src_synop',  $_, $src_synop, 1)}
                                          qw(nws cod ogimet)),
        (map { mk_radio('type_buoy',  $_, $type_buoy)} qw(buoy wmo)),
        (map { mk_radio('src_buoy',   $_, $src_buoy, 1)}
                                          qw(nws ogimet)),
        (map { mk_radio('type_amdar', $_, $type_amdar)} qw(amdar ac)),
        (map { mk_radio('src_amdar',  $_, $src_amdar, 1)} qw(nws)),
    );
    # avoid false positives from Perl::Critic
    %content = (
        lang_de        => 'Deutsch',
        lang_en        => 'English',
        lang_es        => 'Español',
        example_metaf1 => 'METAR QUDO 090600Z 00000KT CAVOK 22/15 Q1021 NOSIG',
        example_metaf2 => 'TAF QUDO 090600Z 0907/0916 VRB03KT 8000 SKC PROB40 TEMPO 0907/0908 0200 +TSRA FM090800 CAVOK',
        example_metaf3 => 'TAF QUDO 090600Z 090716 VRB03KT 8000 SKC PROB40 TEMPO 0708 0200 +TSRA FM0800 CAVOK',
        example_synop1 => 'AAXX 09004 08495 11459 30714 10147 20136 30151 40159 58005 60001 70511 83500 92350',
        example_synop2 => 'BBXX UIDS 18061 99613 70284 41698 61806 10000 40107 57020 70121 86//8 22271 00064',
        example_synop3 => 'OOXX AAERC 10104 99815 31639 60713 00451 46/// /1511 11185 39739 49798 5//// 90940',
        example_buoy1  => 'ZZYY 47503 02031 2206/ 769820 097173 6112/ 444 20120 02031 2201/ 500// 9/000',
        example_amdar1 => 'AMDAR 1605 ASC AFZA51 2624S 02759E 160414 F156 MS048 278/034 TB0 S031 333 F/// VG010',
    );

    if ($ENV{M2X_TEST}) {
        @m2x_text = ('', "<td>$radios{format_msgs}raw ",
                     "before <input type=text name=end_date size=16 maxlength=16 class=mono value=\"$end_date_str\"> (YYYY-MM-DD hh:mm)");
    } else {
        @m2x_text = (' colspan=2', '', '');
    }

    print <<"EOF";
<div id=info>
   <p>$trans->{infosrc}:
   <kbd>https://aviationweather.gov/dataserver</kbd>
   <p>$trans->{infofetch}
</div>
</header>
<form method=get action=metaf.pl>
<table>
   <tr><td colspan=5><strong>$trans->{options}</strong>
   <tr>
      <td class=nowrap>$trans->{language}:
      <td>$radios{lang_de}$content{lang_de}
      <td>$radios{lang_en}$content{lang_en}
      <td>$radios{lang_es}$content{lang_es}
   <tr>
      <td class=nowrap>$trans->{format}:
      <td>$radios{format_html}$trans->{format1}
      <td>$radios{format_text}$trans->{format2}
      <td$m2x_text[0]>$radios{format_xml}$trans->{format3}
      $m2x_text[1]
   <tr>
      <td class=nowrap>$trans->{mode}:
      <td class=nowrap>$radios{mode_latest}$trans->{mode1}
      <td class=nowrap colspan=3>$radios{mode_summary}$trans->{mode2}
         <input type=text name=hours size=3 maxlength=3 class=mono value="$hours">
         $trans->{hours},
         $trans->{unit_temp}$radios{unit_temp_C}°C $radios{unit_temp_F}°F
      $m2x_text[2]
</table>
<table>
   <tr><td colspan=2><hr>
   <tr>
      <td class=top rowspan=3>
         <strong>METAR, TAF </strong>$trans->{example}:
      <td class=mono>$content{example_metaf1}
   <tr>
      <td class=mono>$content{example_metaf2}
   <tr>
      <td class=mono>$content{example_metaf3}
   <tr>
      <td class=nowrap>
         $radios{type_metaf_metar}$trans->{type1}
         <br>
         $radios{type_metaf_taf}$trans->{type2}
         <br>
         $radios{type_metaf_icao}$trans->{type4}:
      <td class=bottom>
         <input type=text name=msg_metaf size=80 maxlength=3000 class=mono value="$msg_metaf_orig">
   <tr>
      <td class="nowrap right">
         $trans->{source}:
      <td>
         $radios{src_metaf_noaa}$trans->{source1}
         $radios{src_metaf_nws}$trans->{source4}
         $radios{src_metaf_addsds}$trans->{source3}
         $radios{src_metaf_ogimet}$trans->{source6}
   <tr><td colspan=2><hr>
   <tr>
      <td class=top rowspan=3>
         <strong>SYNOP </strong>$trans->{example}:
      <td class=mono>$content{example_synop1}
   <tr>
      <td class=mono>$content{example_synop2}
   <tr>
      <td class=mono>$content{example_synop3}
   <tr>
      <td class=nowrap>
         $radios{type_synop_synop}$trans->{type3}
         <br>
         $radios{type_synop_wmo}$trans->{type7}
         <br>
         $radios{type_synop_ship}$trans->{type8}
         <br>
         $radios{type_synop_mobil}$trans->{type9}:
      <td class=bottom>
         <input type=text name=msg_synop size=80 maxlength=3000 class=mono value="$msg_synop_orig">
   <tr>
      <td class="nowrap right">$trans->{source}:
      <td>
         $radios{src_synop_nws}$trans->{source4}
         $radios{src_synop_cod}$trans->{source5}
         $radios{src_synop_ogimet}$trans->{source6}
   <tr><td colspan=2><hr>
   <tr>
      <td class=top>
         <strong>BUOY </strong>$trans->{example}:
      <td class=mono>$content{example_buoy1}
   <tr>
      <td class=nowrap>
         $radios{type_buoy_buoy}$trans->{type3}
         <br>
         $radios{type_buoy_wmo}$trans->{type5}:
      <td class=bottom>
         <input type=text name=msg_buoy size=80 maxlength=3000 class=mono value="$msg_buoy_orig">
   <tr>
      <td class="nowrap right">
         $trans->{source}:
      <td>
         $radios{src_buoy_nws}$trans->{source4}
         $radios{src_buoy_ogimet}$trans->{source6}
   <tr><td colspan=2><hr>
   <tr>
      <td class=top>
         <strong>AMDAR </strong>$trans->{example}:
      <td class=mono>$content{example_amdar1}
   <tr>
      <td class=nowrap>
         $radios{type_amdar_amdar}$trans->{type3}
         <br>
         $radios{type_amdar_ac}$trans->{type11}:
      <td class=bottom>
         <input type=text name=msg_amdar size=80 maxlength=3000 class=mono value="$msg_amdar_orig">
   <tr>
      <td class="nowrap right">
         $trans->{source}:
      <td>
         $radios{src_amdar_nws}$trans->{source4}
   <tr><td colspan=2><hr>
   <tr>
      <td colspan=2>
         <input type=submit class=button value=$trans->{decode}>
</table>
</form>
EOF
}

if ($msg_metaf or $msg_synop or $msg_buoy or $msg_amdar) {
    my ($sum, @xsltproc, $fh);

    $sum = $mode eq 'summary' ? 'sum-' : '';
    print '<p><div class=output><pre>'
        if !$is_cmdline && $format eq 'text';
    @xsltproc = ($XSLTPROC_PATH, qw(--nonet --nowrite --nomkdir --stringparam lang), $lang, qw(--stringparam unit_temp), $unit_temp, qw(--stringparam no_links), $is_cmdline, qw(--stringparam escape4text), (!$is_cmdline && $format eq 'text') ? 1 : 0, "$METAF2XML_SHARE/metaf-$sum$format.xsl", '-');
    if (pipe_open \$fh, 0, @xsltproc) {
        local $SIG{PIPE} = sub {
            error "ERROR: Caught SIGPIPE while writing to $XSLTPROC_PATH.\n";
        };
        start_xml { o => '&' . fileno $fh,
                    O => "$type_metaf $type_synop $type_buoy $type_amdar $lang $unit_temp $format $src_metaf $src_synop $src_buoy $src_amdar $mode $hours $msg_metaf_orig  $msg_synop_orig  $msg_buoy_orig  $msg_amdar_orig"
        };
        for (split /\n/, $msg) {
            parse_report $_, $type_metaf eq 'taf' ? 'TAF' : 'METAR';
        }
        finish;
        (close $fh && !$?)
            or exit_info $!, $?, "'@xsltproc'";
    }
    print '</pre></div>'
        if !$is_cmdline && $format eq 'text';
} elsif ($need_html) {
    print "<!-- Perl: $^V -->\n" .
          '<!-- ' . (scalar gmtime) . "-->\n";
}

print get_stderr($need_html ? 'html' : undef);

if ($need_html) {
    print <<'EOF';
<!-- metaf.pl: 2.9 -->
<p>copyright (c) 2006-2023 metaf2xml @
<a href=https://metaf2xml.sourceforge.io/>
    SourceForge.net
</a>
EOF
} else {
    print 'metaf.pl: 2.9 copyright (c) 2006-2023 metaf2xml @ SourceForge.net' .
          "\n";
}

exit 0;

=head1 EXIT STATUS

If an invalid option was provided, the script will exit with status 1, otherwise
it will exit with status 0.

=head1 EXAMPLES

Parse a METAR message and print the result as HTML:

 metaf.pl msg_metaf="QUDO 090600Z 00000KT CAVOK 22/15 Q1021"

Fetch up-to-date METAR and TAF messages (from NOAA/IWS), SYNOP messages (from
NOAA/NWS land) for Rio (SBGL, 83746), New York (KJFK, 74486), and Tokyo
(RJTT, 47662) and BUOY messages from a buoy nearby (from NOAA/NWS) and print
the result as XML:

 metaf.pl format=xml type_metaf=icao type_synop=wmo type_buoy=wmo msg_metaf="SBGL KJFK RJTT" msg_synop="83746 74486 47662" msg_buoy="31550 41997 53596"

=head1 SEE ALSO

L<metaf2xml::parser|metaf2xml::parser>(3pm),
L<metaf2xml::src2raw|metaf2xml::src2raw>(3pm),
L<metafsrc2raw|metafsrc2raw>(1),
L<https://metaf2xml.sourceforge.io/>

=head1 COPYRIGHT and LICENSE

copyright (c) 2006-2023 metaf2xml @ L<https://metaf2xml.sourceforge.io/>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see L<https://www.gnu.org/licenses/>.

=cut
