########################################################################
# metaf2xml/bufr.pm 2.9
#   process a decoded BUFR message
#
# copyright (c) 2012-2023 metaf2xml
#
# This file is part of metaf2xml.
#
# metaf2xml is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# metaf2xml is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with metaf2xml.  If not, see <https://www.gnu.org/licenses/>.
########################################################################

package metaf2xml::bufr 2.009;

########################################################################
# some things strictly Perl
########################################################################
use warnings;
use 5.014_004;

########################################################################
# export the functions provided by this module
########################################################################
BEGIN {
    require Exporter;

    our @ISA       = qw(Exporter);
    our @EXPORT_OK = qw();
}

END {}

sub VERSION {
    my ($p, $v) = @_;

    die "$p version $v required--this is version $metaf2xml::bufr::VERSION\n"
        if $#_ == 1 && $v != $metaf2xml::bufr::VERSION;
    return $metaf2xml::bufr::VERSION;
}

=encoding utf8

=head1 NAME

metaf2xml::bufr - process a decoded BUFR message

=head1 SYNOPSIS

 use metaf2xml::bufr 2.009;

=head1 DESCRIPTION

This Perl module can be used to process decoded BUFR messages. It has
no public functions and should not be used directly.

=cut

# convert a value (and associated field(s)) to pseudo-BUFR (text) string
sub _val2s {
    my ($val, $assoc) = @_;

    # replace multiple blanks, remove trailing blank
    if (defined $val) {
        $val =~ s/ \K +//g;
        $val =~ s/ $//;
    }
    return   ($assoc
                 ? join('/', '', map { "$_:" . $assoc->{$_} } sort keys %$assoc)
                 : ''
             )
           . (!defined $val
               ? '-'
               : (($val =~ m{ 0\d{5}[/:-]} ? '/' . length($val) : '') . ":$val")
             );
}

# return strings for unprocessed, defined values
sub _not_processed {
    my ($data, $s, $start, $end) = @_;

    return join '',
                map { defined $data->[$_] ? " $s->[$_]" : '' } $start .. $end;
}

sub _str2arr {
    my (@data, @desc, @assoc);

    pos = 0;
    while (m{\G(0\d{5})((?:/\d+:\d+)*)(?:/(\d+))?([:-])}gc) {
        push @desc, $1;
        push @assoc, $2 ? { split m{[/:]}, substr $2, 1 } : undef;
        if (defined $3) {
            push @data, substr $_, pos, $3;
            pos() += $3 + 1;
        } elsif ($4 eq '-') {
            push @data, undef;
            pos()++;
        } else {
            m{\G(?:(.*?) )??(?=0\d{5}[/:-])}gc || /\G(.*)/;
            push @data, $1;
        }
    }
    return \@data, \@desc, \@assoc;
}

sub _K2C {
    my ($temp, $precision) = @_;
    return {
        v => sprintf('%.2f', $temp - 273.15), u => 'C',
        defined $precision ? (rn => $precision, rpi => $precision) : ()
    };
}

sub _salinity2percent {
    my ($v) = @_;

    return defined $v ? (salinity => sprintf '%.3f', $v / 10) : ();
}

# create assoc array, mark invalid groups, filter out repeater groups
sub _prepare {
    my ($data_old, $desc_old) = @_;
    my (@data, @desc, @assoc, $ii, $assoc_signif);

    $ii = 0;
    while ($ii <= $#$data_old) {
        if ($desc_old->[$ii] eq '031021') {
            $assoc_signif = $data_old->[$ii];
        } elsif ($desc_old->[$ii] eq '999999') {
            push @assoc,
                 defined $assoc_signif ? { $assoc_signif => $data_old->[$ii] }
                                       : undef;
            $ii++;
            push @data, $data_old->[$ii];
            push @desc, $desc_old->[$ii];
        } elsif (!{ map { $_ => 1 } qw(031000 031001 031002 031011 031012) }->{$desc_old->[$ii]})
        {
            push @data, $data_old->[$ii];
            push @desc, $desc_old->[$ii];
            push @assoc, undef;
        }

        $ii++;
    }

    return \@data, \@desc, \@assoc;
}

sub _addRadiationSun {
    my ($report, $rad) = @_;
    my $tag;

    $tag = exists $report->{isSynop} ? 'section3' : 'buoyOtherArr';

    for (exists $report->{$tag} ? @{$report->{$tag}} : ()) {
        if (exists $_->{radiationSun}) {
            if (   $_->{radiationSun}{sunshinePeriod}{v}
                == $rad->{sunshinePeriod}{v})
            {
                $_->{radiationSun}{s} .= ' ' . $rad->{s};
                delete $rad->{s};
                $_ = { radiationSun => { %{$_->{radiationSun}}, %$rad }};
                delete $_->{radiationSun}{sunshineNotAvailable}
                    if exists $_->{radiationSun}{sunshine};
                return;
            }
        }
    }
    push @{$report->{$tag}}, { radiationSun => $rad };
    return;
}

sub _mkWind {
    my ($dir, $speed) = @_;

    return { notAvailable => undef }
        if !defined $dir && !defined $speed;

    return { isCalm => undef }
        if defined $dir && defined $speed && $dir == 0 && $speed == 0;

    return {
        (defined $dir ? (dir => $dir) : (dirNotAvailable => undef)),
        (defined $speed ? (speed => { v => $speed, u => 'MPS' })
                        : (speedNotAvailable => undef))
    };
}

sub _mkSensorHeight {
    my ($have_ground, $have_water, $data, $ii) = @_;

    return (sensorHeight => { v => $data->[$ii], u => 'M' })
        if $have_ground && defined $data->[$ii];
    $ii++
        if $have_ground;
    return (sensorHeight => { v => $data->[$ii], u => 'M', above => 'water' })
        if $have_water && defined $data->[$ii];
    return ();
}

sub _fixMeasurePeriod {
    my ($report, $val, $s, $allow_0) = @_;

    if (defined $$val && ($$val > 0 || (!$allow_0 && $$val == 0))) {
        push @{$report->{warning}}, {
            warningType => 'periodInvalid',
            s           => "$$val: $s"
        };
        $$val = -$$val;
    }
    return;
}

# did they just copy CL,CM,CH to 020012,020012,020012?
sub _fixCloudType {
    my ($report, $cloudTypeLow, $cloudTypeMiddle, $cloudTypeHigh, $s) = @_;

    return
        if    !defined $$cloudTypeLow
           && !defined $$cloudTypeMiddle
           && !defined $$cloudTypeHigh;

    if (   (   !defined $$cloudTypeLow
            || ($$cloudTypeLow >= 0 && $$cloudTypeLow <= 9))
        && (   !defined $$cloudTypeMiddle
            || ($$cloudTypeMiddle >= 0 && $$cloudTypeMiddle <= 9))
        && (   !defined $$cloudTypeHigh
            || ($$cloudTypeHigh >= 0 && $$cloudTypeHigh <= 9)))
    {
        push @{$report->{warning}}, {
            warningType => 'cloudTypeInvalid',
            s           => $s
        };
        $$cloudTypeLow += 30
            if defined $$cloudTypeLow;
        $$cloudTypeMiddle += 20
            if defined $$cloudTypeMiddle;
        $$cloudTypeHigh += 10
            if defined $$cloudTypeHigh;
    }
    return;
}

sub _getPeriod {
    my ($r, $val1, $val2, $unit, $s) = @_;

    _fixMeasurePeriod $r, \$val1, $s;
    _fixMeasurePeriod $r, \$val2, $s, 1;

    if (defined $val1) {
        return { $unit => -$val1 }
            if !defined $val2 || $val2 == 0;
        if ($unit eq 'minutes') { # cannot happen yet
            $val1 /= 60;
            $val2 /= 60;
        }
        return { hoursFrom => -$val2, hoursTill => -$val1 };
    }
    return { notAvailable => undef };
}

sub _codeTable001003 {
    my ($v) = @_;

    return   defined $v && $v =~ /^[0-6]$/
           ? (region => qw(Antarctic I II III IV V VI)[$v])
           : ();
}

sub _codeTable020054 {
    my ($type, $v) = @_;

    return   !defined $v ? ("cloudType${type}NA" => undef)
           : $v == 0     ? ("cloudType${type}None" => undef)
           : $v >= 500   ? ("cloudType${type}Invisible" => undef)
           :               ("cloudType${type}Dir" => $v);
}

sub _codeTable020012 {
    my ($v) = @_;

    return   ($v // 59) == 59
                      ? { cloudTypeNotAvailable => undef }
           : $v < 10  ? { cloudType => qw(CI CC CS AC AS NS SC ST CU CB)[$v] }
           : $v < 20  ? { cloudTypeHigh     => $v - 10 }
           : $v < 30  ? { cloudTypeMiddle   => $v - 20 }
           : $v < 40  ? { cloudTypeLow      => $v - 30 }
           : $v == 60 ? { cloudTypeHighNA   => undef }
           : $v == 61 ? { cloudTypeMiddleNA => undef }
           : $v == 62 ? { cloudTypeLowNA    => undef }
           :            { invalidFormat     => $v };
}

########################################################################
# _bufr2report
########################################################################
sub _bufr2report {
    my ($data, $desc, $assoc) = @_;
    my (@reports, $r, %report_base, $not_processed_s, $ii, $is_processed);
    my (@s, $ii_base, $ii_report, $val0, $s0, $sall, $is_auto, $val, $amdarObs);
    my ($temperature, $tempPrecision);
    local $_;

    ($data, $desc, $assoc) = _prepare $data, $desc
        unless defined $assoc;

    for (0 .. $#$desc) {
        push @s, $desc->[$_] . _val2s $data->[$_], $assoc->[$_];
    }

    $_ = "@$desc ";
    $ii = 0;
    $not_processed_s = '';
    push @reports, {};
    $r = $reports[-1];
    $ii_base = -1;
    $ii_report = 0;
    while ($ii <= $#$desc) {
        $is_processed = 1;
        pos   = 7 * $ii;
        $val0 = $data->[$ii];
        $s0   = $s[$ii];

        if (/\G001011 001012 001013/ && defined $val0 && !exists $r->{callSign})
        {
            # replace non-IA5 characters
            $val0 =~ s/[^ -\x7e]/?/g;
            $s[$ii] =~ s/[^ -\x7e]/?/g;

            $r->{callSign} = { s => $s[$ii], id => $val0 };
            $r->{isSynop} = undef;
            @{$r->{obsStationType}}{qw(s stationType)} = qw(BBXX BBXX);

            if (($data->[$ii + 1] // 509) != 509 || defined $data->[$ii + 2]) {
                $r->{displacement} = {
                    s             => "@s[$ii + 1, $ii + 2]",
                    timeBeforeObs => { notAvailable => undef }
                };
                if (defined $data->[$ii + 2] && $data->[$ii + 2] == 0) {
                    $r->{displacement}{isStationary} = undef;
                } else {
                    if (defined $data->[$ii + 2]) {
                        $r->{displacement}{speed} = [
                            { v => $data->[$ii + 2], u => 'MPS' },
                            { v => $data->[$ii + 2], u => 'MPS' }
                        ];
                    } else {
                        $r->{displacement}{speedNotAvailable} = undef;
                    }
                    if (($data->[$ii + 1] // 509) == 509) {
                        $r->{displacement}{dirVarAllUnk} = undef;
                    } else {
                        $r->{displacement}{dir} = $data->[$ii + 1];
                    }
                }
            }
            $ii += 3;
        } elsif (/\G001011 001003/ && defined $val0 && !exists $r->{callSign}) {
            # replace non-IA5 characters
            $val0 =~ s/[^ -\x7e]/?/g;
            $s[$ii] =~ s/[^ -\x7e]/?/g;

            $r->{callSign} = {
                s  => "@s[$ii, $ii + 1]",
                id => $val0,
                _codeTable001003 $data->[$ii + 1]
            };
            $r->{isSynop} = undef;
            @{$r->{obsStationType}}{qw(s stationType)} = qw(OOXX OOXX);
            $ii +=2;
        } elsif (   /\G001001 001002/
                 && (   !exists $r->{obsStationId}
                     || !exists $r->{obsStationId}{id}))
        {
            $r->{obsStationId}{id} = sprintf '%02d%03d', @$data[$ii .. $ii + 1]
                if defined $val0 && defined $data->[$ii + 1];
            if (exists $r->{obsStationId}{s}) {
                $r->{obsStationId}{s} .= " @s[$ii, $ii + 1]";
            } else {
                $r->{obsStationId}{s} = "@s[$ii, $ii + 1]";
            }
            $r->{isSynop} = undef;
            @{$r->{obsStationType}}{qw(s stationType)} = qw(AAXX AAXX);
            $ii += 2;
        } elsif (   /\G001101 001102/
                 && (   !exists $r->{obsStationId}
                     || !exists $r->{obsStationId}{natId})
                 && defined $val0
                 && defined $data->[$ii + 1])
        {
# TODO?
# code table 001101 418: 19 C tables have "CURACAO AND SINT MAARTEN"
#                        all others have "NETHERLANDS ANTILLES AND ARUBA"
            $r->{obsStationId}{natId} = {
                countryId    => $val0,
                natStationId => $data->[$ii + 1]
            };
            if (exists $r->{obsStationId}{s}) {
                $r->{obsStationId}{s} .= " @s[$ii, $ii + 1]";
            } else {
                $r->{obsStationId}{s} = "@s[$ii, $ii + 1]";
            }
            $r->{isSynop} = undef;
            @{$r->{obsStationType}}{qw(s stationType)} = qw(AAXX AAXX);
            $ii += 2;
        } elsif (/\G00101[589]/) {
            my $tag;

            $tag = exists $r->{isBuoy} ? 'buoyId' : 'obsStationId';

            if (defined $val0) {
                # replace non-IA5 characters
                $val0 =~ s/[^ -\x7e]/?/g;
                $s[$ii] =~ s/[^ -\x7e]/?/g;

                $r->{$tag}{obsStationName} = $val0;
            }
            if (exists $r->{$tag}{s}) {
                $r->{$tag}{s} .= " $s[$ii]";
            } else {
                $r->{$tag}{s} = $s[$ii];
            }
            $ii++;
        } elsif (   /\G001003 001020 001005/
                 && defined $data->[$ii + 1]
                 && defined $data->[$ii + 2]
                 && !exists $r->{buoyId})
        {
            $r->{buoyId} = {
                s  => "@s[$ii .. $ii + 2]",
                id => sprintf('%d%d%03d', @$data[$ii .. $ii + 2]),
                _codeTable001003 $val0
            };
            $r->{isBuoy} = undef;
            @{$r->{obsStationType}}{qw(s stationType)} = qw(ZZYY ZZYY);
            $ii += 3;
        } elsif (/\G001087/ && defined $val0 && !exists $r->{buoyId}) {
            # https://community.wmo.int/en/rules-allocating-wmo-numbers:
            #   Numbers in the form A1bwnnn are equivalent to the form A1bw00nnn
            $val0 =~ s/^\d\d\K00(?=\d{3}$)//;
            $r->{buoyId} = { s  => $s0, id => $val0 };
            $r->{isBuoy} = undef;
            @{$r->{obsStationType}}{qw(s stationType)} = qw(ZZYY ZZYY);
            $ii++;
        } elsif (/\G001008/ && defined $val0 && !exists $r->{aircraftId}) {
            # replace non-IA5 characters
            $val0 =~ s/[^ -\x7e]/?/g;
            $s[$ii] =~ s/[^ -\x7e]/?/g;

            $r->{aircraftId} = { s  => $s[$ii], id => $val0 };
            $r->{isAmdar} = undef;
            @{$r->{obsStationType}}{qw(s stationType)} = qw(AMDAR AMDAR);
            $ii++;
        } elsif (/\G001023/ && defined $val0 && exists $r->{isAmdar}) {
            @{$r->{obsSequence}}{qw(s v)} = ($s0, $val0);
            $ii++;
        } elsif (/\G001006/ && defined $val0 && exists $r->{isAmdar}) {
            # replace non-IA5 characters
            $val0 =~ s/[^ -\x7e]/?/g;
            $s[$ii] =~ s/[^ -\x7e]/?/g;

            @{$r->{flightNo}}{qw(s v)} = ($s[$ii], $val0)
                if $val0 ne '';
            $ii++;
        } elsif (/\G001110/ && defined $val0 && exists $r->{isAmdar}) {
            # replace non-IA5 characters
            $val0 =~ s/[^ -\x7e]/?/g;
            $s[$ii] =~ s/[^ -\x7e]/?/g;

            @{$r->{aircraftTailNo}}{qw(s v)} = ($s[$ii], $val0)
                if $val0 ne '';
            $ii++;
        } elsif (/\G001111/ && defined $val0 && exists $r->{isAmdar}) {
            # replace non-IA5 characters
            $val0 =~ s/[^ -\x7e]/?/g;
            $s[$ii] =~ s/[^ -\x7e]/?/g;

            @{$r->{airportDep}}{qw(s v)} = ($s[$ii], $val0)
                if $val0 ne '';
            $ii++;
        } elsif (/\G001112/ && defined $val0 && exists $r->{isAmdar}) {
            # replace non-IA5 characters
            $val0 =~ s/[^ -\x7e]/?/g;
            $s[$ii] =~ s/[^ -\x7e]/?/g;

            @{$r->{airportDest}}{qw(s v)} = ($s[$ii], $val0)
                if $val0 ne '';
            $ii++;
        } elsif (/\G002001/ && ($val0 // -1) < 3) {
            if (($val0 // -1) == 0) {
                $r->{reportModifier} = { s => $s0, modifierType => 'AUTO' };
                $is_auto = 'AUTO';
            } elsif (($val0 // -1) == 2) {
                $r->{reportModifier} = { s => $s0, modifierType => 'MANAUTO' };
            }
            $ii++;
        } elsif (/\G0020(05|6[12])/ && exists $r->{isAmdar}) {
            if (exists $r->{amdarInfo}) {
                $r->{amdarInfo}{s} .= " $s0";
            } else {
                $r->{amdarInfo}{s} = $s0;
            }
            if (defined $val0) {
                $r->{amdarInfo}{ { '05' => 'tempPrecision',
                                    61  => 'navSystem',
                                    62  => 'amdarSystem' }->{$1} } = $val0;
                $tempPrecision = $val0 if $1 eq '05';
            }
            $ii++;
        } elsif (/\G002064/ && $amdarObs) {
            $$amdarObs->{rollAngleQuality} = {
                s                   => $s0,
                rollAngleQualityVal => $val0
            }
                if defined $val0;
            $ii++;
        } elsif (/\G011036/ && exists $r->{isAmdar}) {
            push @{$r->{section3}}, { maxDerivedVerticalGust => {
                s    => $s0,
                wind => { speed => { v => $val0, u => 'MPS' }}
            }}
                if defined $val0;
            $ii++;
        } elsif (   /\G00500([12]) 00600\1/
                 && (   $amdarObs
                     || (   !exists $r->{stationPosition}
                         && !exists $r->{aircraftLocation})))
        {
            if (defined $val0 && defined $data->[$ii + 1]) {
                my ($tag, $pos, $q);

                $tag = exists $r->{isAmdar} ? 'aircraftLocation'
                                            : 'stationPosition';
                $pos->{s} = "@s[$ii, $ii + 1]";
                $q = $1 == 1 ? 100000 : 100;
                @$pos{qw(lat lon)} = ( { v => $val0,            q => $q },
                                       { v => $data->[$ii + 1], q => $q });
                ($amdarObs ? $$amdarObs->{$tag} : $r->{$tag}) = $pos;
            }
            $ii += 2;
        } elsif (/\G007010/ && exists $r->{isAmdar}) {
            push @{$r->{amdarObs}}, {};
            $amdarObs = \$r->{amdarObs}[-1];
            $$amdarObs->{flightLvl} = {
                s     => $s0,
                level => $val0 / 0.3048 / 100 # given in [M]
            }
                if defined $val0;
            $ii++;
        } elsif (/\G011031/ && exists $r->{isAmdar} && $amdarObs) {
            $$amdarObs->{turbulenceAtPA} = {
                s             => $s0,
                turbulenceVal => $val0
            }
                if defined $val0;
            $ii++;
        } elsif (   /\G00800([49])/
                 && defined $val0
                 && (              ($val0 >= 2 && $val0 <= 6)
                     || ($1 == 9 && $val0 >= 0 && $val0 <= 14)))
        {
            $r->{phaseOfFlight} = { s => $s0, phaseOfFlightVal => $val0 };
            $ii++;
        } elsif (   /\G007030/        # ignore elevation if lat/lon are missing
                 && exists $r->{stationPosition}
                 && !exists $r->{stationPosition}{elevation})
        {
            if (defined $val0) {
                $r->{stationPosition}{s} .= " $s0";
                $r->{stationPosition}{elevation} = { v => $val0, u => 'M' };
            }
            $ii++;
        } elsif (/\G007031/ && !exists $r->{barometerElev}) {
            @{$r->{barometerElev}}{qw(s v)} = ($s0, $val0);
            $ii++;
        } elsif (/\G004025 008023 010004 011001 011002 012101 013003 008023/) {
            # 302083 First-order statistics of P, W, T, U data
            $is_processed = 0;
            $not_processed_s .= _not_processed $data, \@s, $ii, $ii + 7;
            $ii += 8;
        } elsif (   /\G(008021 )?004001 004002 004003 004004 004005 /gc
                 && (($val = $ii + ($1 ? 1 : 0)) || 1)
                 && defined $val0
                 && defined $data->[$val]
                 && defined $data->[$val + 1]
                 && defined $data->[$val + 2]
                 && defined $data->[$val + 3]
                 && defined $data->[$val + 4])
        {
# TODO?
# code table 008021 30: 1044 C tables have "RESERVED"
#                       all others have "TIME OF OCCURRENCE"
            if (   (!defined $1 || $val0 == 25)
                || (   defined $1
                    && ($val0 == 26 || $val0 == 0)
                    && !exists $r->{lastKnownPosTime}
                    && exists $r->{isBuoy}))
            {
                my $tag = defined $1 && ($val0 == 26 || $val0 == 0)
                              ? 'lastKnownPosTime'
                              : 'obsTime';

                if ($tag eq 'obsTime') {
                    if (keys %report_base == 0) {        # no report yet?
                        %report_base = %$r;
                        $ii_base = $ii - 1;              # end of base
                        $ii_report = $ii;                # start of first report
                    }

                    if (exists $r->{obsTime}) {          # another report?
                        # complete this report
                        $r->{msg} = "@s[0 .. $ii_base, $ii_report .. $ii - 1]";
                        if ($not_processed_s ne '') {
                            push @{$r->{warning}}, {
                                warningType => 'notProcessed',
                                s           => substr $not_processed_s, 1
                            };
                            $not_processed_s = '';
                        }

                        # make new report
                        push @reports, { %report_base };
                        $r = $reports[-1];
                        $ii_report = $ii;                # start of this report
                    }
                }

                $r->{$tag} = {
                    s      => "@s[$ii .. $val + 4]",
                    timeAt => {
                        year   => $data->[$val],
                        month  => $data->[$val + 1],
                        day    => $data->[$val + 2],
                        hour   => sprintf('%02d', $data->[$val + 3]),
                        minute => sprintf('%02d', $data->[$val + 4])
                    }
                };
                $ii = $val + 5;
                if (/\G004006/) {
                    $r->{$tag}{timeAt}{second} = sprintf '%02d', $data->[$ii]
                        if defined $data->[$ii];
                    $r->{$tag}{s} .= " $s[$ii]";
                    $ii++;
                }
            } else {
                # replace non-IA5 characters
                if (defined $val0) {
                    $s[$ii] =~ s/[^ -\x7e]/?/g;
                    $not_processed_s .= ' ' . $s[$ii];
                }
                $is_processed = 0;
                $ii++;
            }
        } elsif (   /\G(002005 )?002032 002033 007062 022043 022062/
                 && (   !exists $r->{salinityMeasurement}
                     || !defined $data->[$ii + 2 - !defined $1]
                     || $data->[$ii + 2 - !defined $1] eq '0'))
        {
            if (defined $1) {
                $s0 .= ' ';
            } else {
                $ii--;
                $s0 = '';
                $val0 = undef;
            }

            $r->{salinityMeasurement} = {
                s => $s[$ii + 2],
                  $data->[$ii + 2] =~ /^[0-3]$/
                                  ? (salinityMeasurementInd => $data->[$ii + 2])
                                  : (invalidFormat => $data->[$ii + 2])
            }
                if defined $data->[$ii + 2] && $data->[$ii + 2] ne '0';
            $ii += 3;
            pos = 7 * $ii;
            while (/\G007062 022043 022062/) {
                push @{$r->{waterTempSalDepth}}, {
                    s     => "$s0@s[$ii .. $ii + 2]",
                    depth => { v => $data->[$ii], u => 'M' },
                    defined $data->[$ii + 1]
                                  ? (temp => _K2C $data->[$ii + 1], $val0) : (),
                    _salinity2percent $data->[$ii + 2]
                }
                    if    defined $data->[$ii]
                       && (defined $data->[$ii + 1] || defined $data->[$ii + 2]);
                $ii += 3;
                pos = 7 * $ii;
            }
        } elsif (/\G002038 007063 022043 007063/) {
            if (defined $data->[$ii + 2]) {
                if (   ($data->[$ii + 1] // 0) <= 10
                    && !exists $r->{seaSurfaceTemp})
                {
                    $r->{seaSurfaceTemp} = {
                        s    => "@s[$ii .. $ii + 3]",
                        temp => _K2C($data->[$ii + 2]),
                        defined $val0 ? (waterTempSalMeasurement => $val0) : ()
                    };
                } else {
                    push @{$r->{waterTempSalDepth}}, {
                        s     => "@s[$ii .. $ii + 3]",
                        depth => { v => ($data->[$ii + 1] // 0), u => 'CM' },
                        temp  => _K2C($data->[$ii + 2]),
                        defined $val0 ? (waterTempSalMeasurement => $val0) : ()
                    };
                }
            }
            $ii += 4;
        } elsif (   /\G002005 022043 002033 022059/
                 && (   !exists $r->{salinityMeasurement}
                     || !defined $data->[$ii + 2]
                     || $data->[$ii + 2] eq '0'))
        {
            $r->{salinityMeasurement} = {
                s => $s[$ii + 2],
                  !defined $data->[$ii + 2]
                ? (notAvailable => undef)
                : $data->[$ii + 2] =~ /^[0-3]$/
                ? (salinityMeasurementInd => $data->[$ii + 2])
                : (invalidFormat => $data->[$ii + 2])
            }
                if defined $data->[$ii + 2] && $data->[$ii + 2] ne '0';
            if (defined $data->[$ii + 1] && !defined $data->[$ii + 3]) {
                unshift @{$r->{section2}}, { seaSurfaceTemp => {
                    s => "@s[$ii, $ii + 1]",
                    temp => _K2C $data->[$ii + 1], $val0
                }};
            } elsif (defined $data->[$ii + 1] || defined $data->[$ii + 3]) {
                push @{$r->{waterTempSalDepth}}, {
                    s     => "@s[$ii, $ii + 1, $ii + 3]",
                    depth => { v => 0, u => 'M' },
                    defined $data->[$ii + 1]
                                  ? (temp => _K2C $data->[$ii + 1], $val0) : (),
                    _salinity2percent $data->[$ii + 3]
                };
            }
            $ii += 4;
        } elsif (   /\G002005 007063 022049/
                 && !exists $r->{seaSurfaceTemp}
                 && ($data->[$ii + 1] // 0) <= 10)
        {
            $r->{seaSurfaceTemp} = {
                s    => "@s[$ii .. $ii + 2]",
                temp => _K2C $data->[$ii + 2], $val0
            }
                if defined $data->[$ii + 2];
            $ii += 3;
        } elsif (   /\G02200([12]) 02201\1 02202\1/
                 && (exists $r->{isSynop} || exists $r->{isBuoy})
                 && !exists $r->{qw(waves windWaves)[$1 - 1]})
        {
            if (   defined $val0
                || defined $data->[$ii + 1]
                || defined $data->[$ii + 2])
            {
                $r->{qw(waves windWaves)[$1 - 1]} = {
                    s => "@s[$ii .. $ii + 2]",
                         ($data->[$ii + 1] // -1) == 0
                      || ($data->[$ii + 2] // -1) == 0
                    ? (noWaves => undef)
                    : (defined $val0 ? (dir => $val0) : (),
                       defined $data->[$ii + 1]
                           ? (wavePeriod => $data->[$ii + 1]) : (),
                       defined $data->[$ii + 2]
                           ? (height => { v => $data->[$ii + 2], u => 'M'}) : ()
                      )
                };
            }
            $ii += 3;
        } elsif (   (exists $r->{isSynop} || exists $r->{isBuoy})
                 && (!exists $r->{swell} || $#{$r->{swell}{swellData}} == 0)
                 && /\G022003 022013 022023/)
        {
            if (   defined $val0
                || defined $data->[$ii + 1]
                || defined $data->[$ii + 2])
            {
                if (exists $r->{swell}) {
                    $r->{swell}{s} .= " @s[$ii .. $ii + 2]";
                } else {
                    $r->{swell}{s} = "@s[$ii .. $ii + 2]";
                }
                push @{$r->{swell}{swellData}}, {
                         ($data->[$ii + 1] // -1) == 0
                      || ($data->[$ii + 2] // -1) == 0
                    ? (noWaves => undef)
                    : (defined $val0 ? (dir => $val0) : (),
                       defined $data->[$ii + 1]
                           ? (wavePeriod => $data->[$ii + 1]) : (),
                       defined $data->[$ii + 2]
                           ? (height => { v => $data->[$ii + 2], u => 'M'}) : ()
                      )
                };
            }
            $ii += 3;
        } elsif (   exists $r->{isSynop}
                 && /\G007004 010009/
                 && !exists $r->{gpSurface})
        {
            if (defined $val0 && defined $data->[$ii + 1]) {
                $r->{gpSurface}{s} = "@s[$ii, $ii + 1]";

                if ({ map { $_ => 1 } qw(1000 925 500 700 850) }->{$val0 / 100})
                {
                    @{$r->{gpSurface}}{qw(surface geopotential)} =
                                                ($val0 / 100, $data->[$ii + 1]);
                } else {
                    $r->{gpSurface}{invalidFormat} = $r->{gpSurface}{s};
                }
            }
            $ii += 2;
        } elsif (   /\G(007032 )?(007033 )?012101 /gc
                 && ($amdarObs || !exists $r->{temperature}))
        {
            $val = $ii;
            $val++
                if $1;
            $val++
                if $2;
            ($amdarObs ? $$amdarObs->{temperature} : $r->{temperature}) =
            $temperature = {
                s   => "@s[$ii .. $val]",
                air => defined $data->[$val] ? { temp => _K2C $data->[$val] }
                                             : { notAvailable => undef },
                _mkSensorHeight $1, $2, $data, $ii
            };
            $ii = $val + 1;

            if (/\G002039 012102 /gc) {
                $not_processed_s .= _not_processed $data, \@s, $ii, $ii + 1;
                $ii += 2;
            }

            if (/\G012103 /gc) {
                $temperature->{s} .= " $s[$ii]";
                $temperature->{dewpoint} = { temp => _K2C $data->[$ii] }
                    if defined $data->[$ii] && $data->[$ii] > 0;
                $ii++;
            }
            if (/\G013003/gc) {
                $temperature->{s} .= " $s[$ii]";
                $temperature->{relHumid1} = $data->[$ii]
                    if defined $data->[$ii];
                $ii++;
            }
            if (   !exists $temperature->{air}{temp}
                && !exists $temperature->{dewpoint}
                && !exists $temperature->{relHumid1})
            {
                if ($amdarObs) {
                    delete $$amdarObs->{temperature};
                } else {
                    delete $r->{temperature};
                }
                $temperature = undef;
            }
        } elsif (   /\G(004025 )?010004/
                 && !exists $r->{stationPressure}
                 && (!defined $1 || (defined $1 && ($val0 // 0) == 0)))
        {
            if (defined $1) {
                $ii++;
                $val0 = $data->[$ii];
                $s0 .= " $s[$ii]";
            }
            $r->{stationPressure} = {
                s => $s0,
                $val0 < 20000
                    ? (invalidFormat => '<200 hPa')
                    : (pressure =>
                              { v => sprintf('%.1f', $val0 / 100), u => 'hPa' })
            }
                if defined $val0;
            $ii++;
        } elsif (/\G010051/ && !exists $r->{SLP}) {
            $r->{SLP} = {
                s => $s0,
                $val0 < 80000
                    ? (invalidFormat => '<800 hPa')
                    : (pressure =>
                              { v => sprintf('%.1f', $val0 / 100), u => 'hPa' })
            }
                if defined $val0;
            $ii++;
        } elsif (   /\G01006([12])( 010063)?/
                 && (!exists $r->{pressureChange} || exists $r->{isSynop}))
        {
            $val = $ii + ($2 ? 1 : 0);
            if (defined $val0) {
                my $pc;

                $pc = {
                    s                 => "@s[$ii .. $val]",
                    timeBeforeObs     => { hours => (3, 24)[$1 - 1] },
                    pressureChangeVal =>
                               { v => sprintf('%.1f', $val0 / 100), u => 'hPa' }
                };
                if ($2 && defined $data->[$val]) {
                    if ($data->[$val] >= 0 && $data->[$val] <= 8) {
                        if (   $data->[$val] >= 6
                            && $pc->{pressureChangeVal}{v} > 0)
                        {
                            $pc->{invalidFormat} = "$val0<=>$data->[$val]";
                            delete $pc->{pressureChangeVal};
                        } else {
                            $pc->{pressureTendency} = $data->[$val];
                        }
                    } else {
                        $pc->{invalidFormat} = $s[$val];
                        delete $pc->{pressureChangeVal};
                    }
                }
                if (!exists $r->{pressureChange} && $1 == 1) {
                    $r->{pressureChange} = $pc;
                } else {
                    push @{$r->{section3}}, { pressureChange => $pc };
                }
            }
            $ii = $val + 1;
        } elsif (/\G(007032 )?013023/) {
            $val = $1 ? $ii + 1 : $ii;
            push @{$r->{section3}}, { precipitation => {
                s             => "@s[$ii .. $val]",
                timeBeforeObs => { hours => 24 },
                $data->[$val] == -0.1
                   ? (precipTraces => undef)
                   : (precipAmount => { v => $data->[$val], u => 'MM' }),
                $1 && defined $val0
                   ? (sensorHeight => { v => $val0, u => 'M' })
                   : ()
            }}
                if defined $data->[$val];
            $ii = $val + 1;
        } elsif (/\G(007032 )?(002175 002178 )?(?=00402[45] 013011 )/gc) {
            my ($sensorHeight, $s, $tag);

            if ($1) {
                $sensorHeight = $val0;
                $ii++;
            }
            if ($2) {
                $not_processed_s .= _not_processed $data, \@s, $ii, $ii + 1;
                $ii += 2;
            }
            $tag = exists $r->{isSynop} ? 'section3' : 'buoyOtherArr';
            while (/\G00402([45]) 013011 /gc) {
                if (defined $data->[$ii + 1]) {
                    $s = "$s0 @s[$ii, $ii + 1]";
                    push @{$r->{$tag}}, { precipitation => {
                            s             => $s,
                            timeBeforeObs => _getPeriod($r, $data->[$ii], undef,
                                                 qw(hours minutes)[$1 - 4], $s),
                            $data->[$ii + 1] == -0.1
                                 ? (precipTraces => undef)
                                 : (precipAmount => { v => $data->[$ii + 1],
                                                      u => 'MM' }),
                            defined $sensorHeight
                                 ? (sensorHeight => { v => $val0, u => 'M' })
                                 : ()
                    }};
                }
                $ii += 2;
            }
        } elsif (   /\G(007032 )?(007033 )?(033041 )?020001/
                 && !exists $r->{visPrev})
        {
            $val = $ii;
            $val++
                if $1;
            $val++
                if $2;
            $val++
                if $3;
            $r->{visPrev} = {
                s        => "@s[$ii .. $val]",
                distance => { v => $data->[$val], u => 'M',
                              $3
                           && defined $data->[$val - 1]
                           && ($data->[$val - 1] == 1 || $data->[$val - 1] == 2)
                         ? (q => qw(isGreater isLess)[$data->[$val - 1] - 1])
                         : ()
               },
               _mkSensorHeight $1, $2, $data, $ii
            }
                if defined $data->[$val];
            $ii = $val + 1;
        } elsif (/\G020010/ && !exists $r->{totalCloudCover}) {
            $r->{totalCloudCover} = {
                    s => $s0,
                      $val0 > 100 && $val0 != 113 ? (invalidFormat => $s0)
                    : $val0 == 113                ? (skyObscured => undef)
                    :                    (oktas => sprintf '%.0f', $val0 / 12.5)
            }
                if defined $val0;
            $ii++;
        } elsif (   /\G008002 020011 020013 020012 020012 020012/
                 && !exists $r->{cloudTypes})
        {
            my $s;

            $r->{baseLowestCloud} = {
                s        => $s[$ii + 2],
                distance => { v => $data->[$ii + 2], u => 'M' }
            }
                if defined $data->[$ii + 2];

            $s = "@s[$ii, $ii + 1, $ii + 3 .. $ii + 5]";
            if (!defined $data->[$ii + 1] && !defined $data->[$ii + 3]) {
                # ignore
            } elsif (($data->[$ii + 1] // -1) == 9) {
                $r->{cloudTypes}{skyObscured} = undef;
            } else {
                _fixCloudType $r, \@$data[$ii + 3 .. $ii + 5], $s;
                if (   (   defined $data->[$ii + 3]
                        && (   $data->[$ii + 3] < 30
                            || (   $data->[$ii + 3] >= 40
                                && $data->[$ii + 3] != 62)))
                    || (   defined $data->[$ii + 4]
                        && (   $data->[$ii + 4] < 20
                            || (   $data->[$ii + 4] >= 30
                                && $data->[$ii + 4] != 61)))
                    || (   defined $data->[$ii + 5]
                        && (   $data->[$ii + 5] < 10
                            || (   $data->[$ii + 5] >= 20
                                && $data->[$ii + 5] != 60))))
                {
                    $r->{cloudTypes}{invalidFormat} = $s;
                } else {
                    if (($data->[$ii + 3] // 62) == 62) {
                        $r->{cloudTypes}{cloudTypeLowNA} = undef;
                    } else {
                        $r->{cloudTypes}{cloudTypeLow} = $data->[$ii + 3] - 30;
                    }
                    if (($data->[$ii + 4] // 61) == 61) {
                        $r->{cloudTypes}{cloudTypeMiddleNA} = undef;
                    } else {
                        $r->{cloudTypes}{cloudTypeMiddle} =
                                                          $data->[$ii + 4] - 20;
                    }
                    if (($data->[$ii + 5] // 60) == 60) {
                        $r->{cloudTypes}{cloudTypeHighNA} = undef;
                    } else {
                        $r->{cloudTypes}{cloudTypeHigh} = $data->[$ii + 5] - 10;
                    }
                    if (($data->[$ii + 1] // 9) < 9) {
                        if (($val0 // 0) == 0) {
                            if (($r->{cloudTypes}{cloudTypeLow} // 0) > 0) {
                                $r->{cloudTypes}{oktasLow} = $data->[$ii + 1];
                            } elsif (   exists $r->{cloudTypes}{cloudTypeLow}
                                     && (($r->{cloudTypes}{cloudTypeMiddle} // 0) > 0))
                            {
                                $r->{cloudTypes}{oktasMiddle} =
                                                               $data->[$ii + 1];
                            } else {
                                $r->{cloudTypes}{oktas} = $data->[$ii + 1];
                            }
                        } elsif ($val0 == 7 || $val0 == 8) {
                            $r->{cloudTypes}{qw(oktasLow oktasMiddle)[$val0 -7]}
                                                             = $data->[$ii + 1];
                        }
                    }
                }
            }
            $r->{cloudTypes}{s} = $s
                if exists $r->{cloudTypes};

            $ii += 6;
        } elsif (   exists $r->{isSynop}
                 && /\G008002 020011 020012 (033041 )?020013/
                 && (   !defined $val0
                     || (   ($val0 >= 0 && $val0 <= 5)
                         || ($val0 >= 20 && $val0 <= 24))))
        {
            $val = $ii + ($1 ? 4 : 3);

            if (   defined $data->[$val]
                || ($data->[$ii + 1] // 10) <= 9
                || ($data->[$ii + 2] // 59) != 59)
            {
                my ($cl, $noClouds);

                $cl->{s} = "@s[$ii .. $val]";
                if (defined $val0 && $val0 == 20) {
                    push @{$cl->{sortedArr}},{ noClouds => 'NCD' };
                    $noClouds = 1;
                } elsif (!defined $data->[$val]) {
                    push @{$cl->{sortedArr}},{ cloudBaseNotAvailable => undef };
                } else {
                    push @{$cl->{sortedArr}}, { cloudBase => {
                        v => $data->[$val],
                        u => 'M',
                             $1
                          && defined $data->[$val - 1]
                          && ($data->[$val - 1] == 1 || $data->[$val - 1] == 2)
                        ? (q => qw(isGreater isLess)[$data->[$val - 1] - 1])
                        : ()
                    }};
                }
                if (!$noClouds) {
                    push @{$cl->{sortedArr}}, _codeTable020012 $data->[$ii + 2];
                    push @{$cl->{sortedArr}}, { cloudOktas => {
                          ($data->[$ii + 1] // 10) > 9
                        ? (oktasNotAvailable => undef)
                        : $data->[$ii + 1] == 9
                        ? (skyObscured => undef)
                        : (oktas => $data->[$ii + 1])
                    }};
                }
                push @{$r->{section3}}, { cloudInfo => $cl };
            }
            $ii = $val + 1;
        } elsif (   /\G008002 020011 020012 020014 020017/
                 && exists $r->{isSynop}
                 && ($val0 // 0) == 11)
        {
            push @{$r->{section4}}, { cloudBelowStation => {
                s          => "@s[$ii .. $ii + 4]",
                cloudOktas =>
                      ($data->[$ii + 1] // 10) > 9
                    ? { oktasNotAvailable => undef }
                    : $data->[$ii + 1] == 9
                    ? { skyObscured => undef }
                    : { oktas => $data->[$ii + 1] },
                cloudTopDescr => $data->[$ii + 4],
                cloudTops     => { v => $data->[$ii + 3], u => 'M' },
                      ($data->[$ii + 2] // 10) > 9
                    ? (cloudTypeNotAvailable => undef)
                    : (cloudType =>
                            qw(CI CC CS AC AS NS SC ST CU CB)[$data->[$ii + 2]])
            }}
                if defined $data->[$ii + 3] && defined $data->[$ii + 4];
            $ii += 5;
        } elsif (/\G011001 011002/ && ($amdarObs || !exists $r->{sfcWind})) {
            ($amdarObs ? $$amdarObs->{windAtPA} : $r->{sfcWind}) = {
                s    => "@s[$ii .. $ii + 1]",
                wind => _mkWind $val0, $data->[$ii + 1]
            }
                if defined $val0 || defined $data->[$ii + 1];
            $ii += 2;
        } elsif (   /\G(007032 )?(007033 )?(002002 )?008021 004025 011001 011002 /gc
                 && !exists $r->{sfcWind}
                 && (($val = $ii + ($1 ? 1 : 0) + ($2 ? 1 : 0) + ($3 ? 1 : 0)) || 1)
                 && defined $data->[$val]
                 && $data->[$val] == 2)
        {
# TODO?
# code table 008021 30: 1044 C tables have "RESERVED"
#                       all others have "TIME OF OCCURRENCE"
            my $have_wind;

            $have_wind = defined $data->[$val + 2] || defined $data->[$val + 3];
            if ($have_wind) {
                $r->{sfcWind} = {
                    s => "@s[$ii .. $val + 3]",
                    _mkSensorHeight $1, $2, $data, $ii
                };

                if (defined $data->[$val + 1]) {
                    _fixMeasurePeriod $r, \$data->[$val + 1], $r->{sfcWind}{s};
                    $r->{sfcWind}{measurePeriod} =
                                        { v => -$data->[$val + 1], u => 'MIN' };
                }

                $r->{sfcWind}{wind} = _mkWind $data->[$val + 2],
                                              $data->[$val + 3];
            }
            $ii = $val + 4;

            if (/\G008021/ && !defined $data->[$ii]) {
# TODO?
# code table 008021 30: 1044 C tables have "RESERVED"
#                       all others have "TIME OF OCCURRENCE"
                $r->{sfcWind}{s} .= " $s[$ii]"
                    if $have_wind;
                $ii++;
            }
        } elsif (   /\G004025 011016 011017/
                 && (   !(defined $data->[$ii + 1] && defined $data->[$ii + 2])
                     || (   defined $val0
                         && exists $r->{sfcWind}
                         && exists $r->{sfcWind}{measurePeriod}
                         && $r->{sfcWind}{measurePeriod}{v} == -$val0)))
        {
            if (defined $data->[$ii + 1] && defined $data->[$ii + 2]) {
                $r->{sfcWind}{s} .= " @s[$ii .. $ii + 2]";
                @{$r->{sfcWind}{wind}}{qw(windVarLeft windVarRight)}
                                                     = @$data[$ii + 1, $ii + 2];
            }
            $ii += 3;
        } elsif (/\G004025( 011043)? 011041/) {
            my ($have_dir, $dir, $speed);

            $have_dir = defined $1 ? 1 : 0;
            if ($have_dir) {
                $dir = $data->[$ii + 1];
                $speed = $data->[$ii + 2];
            } else {
                $speed = $data->[$ii + 1];
            }
            if (defined $dir || defined $speed) {
                my $tag;

                $tag = exists $r->{isSynop} ? 'section3' : 'buoyOtherArr';
                _fixMeasurePeriod $r, \$val0, "@s[$ii .. $ii + 1 + $have_dir]";
                push @{$r->{$tag}}, { highestGust => {
                    s             => "@s[$ii .. $ii + 1 + $have_dir]",
                    defined $val0
                        ? (measurePeriod => { v => -$val0, u => 'MIN' })
                        : (),
                    wind          => {
                         defined $dir ? (dir => $dir)
                                                  : (dirNotAvailable => undef),
                         defined $speed
                             ? (speed => { v => $speed, u => 'MPS' })
                             : (speedNotAvailable => undef)
                    }
                }};
            }
            $ii += 2 + $have_dir;
        } elsif (/\G004025 011042/) {
            if (defined $data->[$ii + 1]) {
                my $s;

                $s = "@s[$ii, $ii + 1]";
                push @{$r->{section3}}, { highestMeanSpeed => {
                    s             => $s,
                    measurePeriod => { v => 10, u => 'MIN' },
                    wind          => { speed => { v => $data->[$ii + 1],
                                                  u => 'MPS'}},
                    timeBeforeObs => _getPeriod($r, $val0, undef, 'minutes', $s)
                }};
            }
            $ii += 2;
        } elsif (/\G007032 004025 012112/) {
            if (defined $val0 && defined $data->[$ii + 2]) {
                my $s;

                $s = "@s[$ii .. $ii + 2]";
                push @{$r->{section3}}, {
                    'tempMin' . ($val0 <= 0.1 ? 'Ground' : '') => {
                        s             => $s,
                        sensorHeight  => { v => $val0, u => 'M' },
                        timeBeforeObs => _getPeriod($r, $data->[$ii + 1], undef,
                                                    'minutes', $s),
                        temp          => _K2C $data->[$ii + 2]
                }};
            }
            $ii +=3;
        } elsif (/\G007032 007033 004025 012111 012112/) {
            if (defined $data->[$ii + 3] || defined $data->[$ii + 4]) {
                my $s;

                $s = "@s[$ii .. $ii + 4]";
                _fixMeasurePeriod $r, \$data->[$ii + 2], $s; # fix only once
                push @{$r->{section3}}, { tempExtreme => {
                    s => $s,
                    _mkSensorHeight(1, 1, $data, $ii),
                    tempExtremeMax => {
                        defined $data->[$ii + 3]
                        ? (temp          => _K2C($data->[$ii + 3]),
                           timeBeforeObs => _getPeriod($r, $data->[$ii + 2],
                                                       undef, 'minutes', $s))
                        : (notAvailable => undef)
                    },
                    tempExtremeMin => {
                        defined $data->[$ii + 4]
                        ? (temp          => _K2C($data->[$ii + 4]),
                           timeBeforeObs => _getPeriod($r, $data->[$ii + 2],
                                                       undef, 'minutes', $s))
                        : (notAvailable => undef)
                    }
                }};
            }
            $ii += 5;
        } elsif (/\G007032 (007033 )?004024 004024 012111 004024 004024 012112/)
        {
# TODO?
# 004024: B tables 2000, 98000000, 98002001 have reference -1024, width 11
#         all others have reference -2048, width 12
            $val = $ii + ($1 ? 2 : 1);
            if (defined $data->[$val + 2] || defined $data->[$val + 5]) {
                my $s;

                $s = "@s[$ii .. $val + 5]";
                push @{$r->{section3}}, { tempExtreme => {
                    s => $s,
                    _mkSensorHeight(1, $1, $data, $ii),
                    tempExtremeMax => {
                        defined $data->[$val + 2]
                           ? (temp          => _K2C($data->[$val + 2]),
                              timeBeforeObs => _getPeriod($r, $data->[$val],
                                                          $data->[$val + 1],
                                                          'hours', $s))
                           : (notAvailable => undef)
                    },
                    tempExtremeMin => {
                        defined $data->[$val + 5]
                           ? (temp          => _K2C($data->[$val + 5]),
                              timeBeforeObs => _getPeriod($r, $data->[$val + 3],
                                                          $data->[$val + 4],
                                                          'hours', $s))
                           : (notAvailable => undef)
                }}};
            }
            $ii = $val + 6;
        } elsif (/\G012113/) {
            push @{$r->{section3}}, { tempMinGround => {
                s             => $s0,
                temp          => _K2C($val0),
                timeBeforeObs => { hours => 12 }
            }}
                if defined $val0;
            $ii++;
        } elsif (!exists $r->{weatherSynop} && /\G020003 /gc) {
           #if (defined $val0) {
           #    my $wxIndVal;

           #    $wxIndVal =   $val0 < 100  ? 1
           #                : $val0 < 200  ? 7
           #                : $val0 == 508 ? ($is_auto ? 5 : 2)
           #                : $val0 == 509 ? ($is_auto ? 6 : 3)
           #                :                0;
           #    $r->{wxInd} = { s => $wxIndVal, wxIndVal => $wxIndVal }
           #        if $wxIndVal;
           #}

            if (!defined $val0) {
            } elsif ($val0 == 508) {
                $r->{weatherSynop}{NSW} = undef;
            } elsif ($val0 >= 300) {
                $r->{weatherSynop}{weatherPresentNotAvailable} = undef;
            } elsif ($val0 < 100) {
                $r->{weatherSynop}{weatherPresent} = $val0;
            } elsif ($val0 < 200) {
                $r->{weatherSynop}{weatherPresentSimple} = $val0 - 100;
            } else {
                # code table 4687, must be post-processed by metaf2xml::parser
                push @{$r->{weatherSynopAdd}}, {
                    s              => $s0,
                    weatherPresent => $val0 - 200,
                };
            }
            $r->{weatherSynop}{s} = $s0
                if exists $r->{weatherSynop};
            $ii++;

            while (/\G004024 020004 020005 /gc) {
                if (defined $data->[$ii + 1] || defined $data->[$ii + 2]) {
                    my $w;

                    push @{$r->{warning}},{ warningType => 'weatherNotOmitted' }
                        if    ($val0 // -1) == 508
                           && (   (   defined $data->[$ii + 1]
                                   && (   $data->[$ii + 1] % 10 != 0
                                       || $data->[$ii + 1] > 19))
                               || (   defined $data->[$ii + 2]
                                   && (   $data->[$ii + 2] % 10 != 0
                                       || $data->[$ii + 2] > 19)));

                    $w->{s} = "@s[$ii .. $ii + 2]";
                    $w->{timeBeforeObs} = _getPeriod($r, $data->[$ii], undef,
                                                     'hours', $w->{s});

                    for (1, 2) {
                        $val = $data->[$ii + $_];
                        if (($val // 20) > 19) {
                            $w->{"weatherPast${_}NotAvailable"} = undef;
                        } else {
                            my $simple;

                            $simple = $val > 9 ? 'Simple' : '';
                            $w->{"weatherPast$_$simple"} = $val % 10;
                        }
                    }
                    push @{$r->{weatherPast}}, $w;
                }
                $ii += 3;
            }
        } elsif (   exists $r->{isSynop}
                 && /\G004025 004025 020003/
                 && defined $data->[$ii + 2])
        {
            my $s;

            $s = "@s[$ii .. $ii + 2]";
            push @{$r->{section3}}, { weatherSynopAmpl => {
                s              => $s,
                timeBeforeObs  => _getPeriod($r, $val0, $data->[$ii + 1],
                                             'minutes', $s),
                weatherPresent => $data->[$ii + 2]
            }};
            $ii += 3;
        } elsif (/\G020003/ && defined $val0 && $val0 < 100) {
            push @{$r->{section3}}, { weatherSynopAdd => {
                s              => $s0,
                weatherPresent => $val0
            }};
            $ii++;
        } elsif (   /\G(020023 020024 |020054 020023 |020054 020025 020026 |020040 020066 |020021 020067 )020027/
                 && ($data->[$ii + length($1) / 7] // 0) == 0)
        {
            # dangerous weather phenomena, but phenomena occurrence = 0
            $ii += length($1) / 7 + 1;
        } elsif (   exists $r->{isSynop}
                 && /\G020021 020067 020027/
                 && defined $val0            && $val0 == 1024
                 && defined $data->[$ii + 1]
                 && defined $data->[$ii + 2] && $data->[$ii + 2] == 256)
        {
            push @{$r->{section3}}, { wetsnowDeposit => {
                s => "@s[$ii .. $ii + 2]",
                diameter => { v => $data->[$ii + 1] * 1000, u => 'MM' }
            }};
            $ii += 3;
        } elsif (/\G004024 014031/) {
            # TODO: sunshinePeriod/@v=24 != sunshinePeriod/@v=p !?!
            if (defined $val0 && defined $data->[$ii + 1]) {
                _fixMeasurePeriod $r, \$val0, "@s[$ii, $ii + 1]";
                _addRadiationSun $r, {
                    s              => "@s[$ii, $ii + 1]",
                    sunshine       => { v => $data->[$ii + 1], u => 'MIN' },
                    sunshinePeriod => { v => -$val0, u => 'H' }
                };
            }
            $ii += 2;
        } elsif (/\G00402([45]) 014002 014004 014016 014028 014029 014030/) {
# TODO?
# 014002, 014004: 1310 B tables have reference -2048, width 12
#         all others have reference -65536, width 17
# 014028. 014029, 014030: 1146 B tables have width 16
#         all others have width 20
            my $p;

            _fixMeasurePeriod $r, \$val0, "@s[$ii .. $ii + 6]";
            $p = { v => -$val0, u => qw(H MIN)[$1 - 4] }
                if defined $val0;
            _addRadiationSun $r, {
                s => "@s[$ii .. $ii + 5 + (defined $data->[$ii + 6] ? 0 : 1)]",
                defined $data->[$ii + 1]
                    ? (rad4DownwardLongWave => { radiationValue =>
                                 { v => $data->[$ii + 1] / 1000, u => 'kJm2' }})
                    : (),
                defined $data->[$ii + 2]
                    ? (rad6ShortWave => { radiationValue =>
                                 { v => $data->[$ii + 2] / 1000, u => 'kJm2' }})
                    : (),
                defined $data->[$ii + 3] && $data->[$ii + 3] >= 0
                    ? (rad0PosNet => { radiationValue =>
                                 { v => $data->[$ii + 3] / 1000, u => 'kJm2' }})
                    : (),
                defined $data->[$ii + 3] && $data->[$ii + 3] < 0
                    ? (rad1NegNet => { radiationValue =>
                                { v => -$data->[$ii + 3] / 1000, u => 'kJm2' }})
                    : (),
                defined $data->[$ii + 4]
                    ? (rad2GlobalSolar => { radiationValue =>
                                 { v => $data->[$ii + 4] / 1000, u => 'kJm2' }})
                    : (),
                defined $data->[$ii + 5]
                    ? (rad3DiffusedSolar => { radiationValue =>
                                 { v => $data->[$ii + 5] / 1000, u => 'kJm2' }})
                    : (),
                radiationPeriod      => $p,
                sunshinePeriod       => $p,
                sunshineNotAvailable => undef,
            }
                if    defined $val0
                   && (   defined $data->[$ii + 1]
                       || defined $data->[$ii + 2]
                       || defined $data->[$ii + 3]
                       || defined $data->[$ii + 4]
                       || defined $data->[$ii + 5]);
            push @{$r->{section3}}, { radDirectSolar => {
                s               => "$s0 $s[$ii + 6]",
                radiationPeriod => $p,
                radiationValue  => { v => $data->[$ii + 6], u => 'kJm2' }
            }}
                if defined $val0 && defined $data->[$ii + 6];
            $ii += 7;
        } elsif (/\G002149/) {
            $r->{buoyType} = { s => $s0, buoyTypeInd => $val0 }
                if defined $val0;
            $ii++;
        } elsif (/\G020062/ && defined $val0) {
            my $tag;

            if ($val0 < 10) {
                $tag = 'stateOfGround';
            } else {
                $tag = 'stateOfGroundSnow';
                $val0 -= 10;
            }
            push @{$r->{section3}}, { $tag => {
                s           => $s0,
                "${tag}Val" => $val0
            }};
            $ii++;
        } elsif (/\G013013/ && defined $val0) {
# TODO?
# 013013: 98000000 has reference 0
#         all others have reference -2
            push @{$r->{section3}}, { snowDepth => {
                s            => $s0,
                precipAmount => { v => $val0 * 100, u => 'CM' }
            }};
            $ii++;
        } elsif (exists $r->{isSynop} && /\G((?:007061 012130 )+)007061/) {
            for (1..length($1) / 14) {
                my $len = $_ == length($1) / 14 ? 3 : 2;

                if (defined $data->[$ii] && defined $data->[$ii + 1]) {
                    push @{$r->{section5}}, { soilTemp => {
                        s     => "@s[$ii .. $ii + $len - 1]",
                        depth => { v => $data->[$ii], u => 'M' },
                        temp  => _K2C $data->[$ii + 1]
                    }};
                } else {
                    $not_processed_s .= " @s[$ii .. $ii + $len - 1]"
                        if defined $data->[$ii + 1];
                }
                $ii += $len;
            }
        } elsif (exists $r->{isSynop} && /\G004024 002004 013033/) {
            if (defined $data->[$ii + 1] && defined $data->[$ii + 2]) {
                my $s;

                $s = "@s[$ii .. $ii + 2]";
                push @{$r->{section3}}, { evapo => {
                    s              => $s,
                    timeBeforeObs  => _getPeriod($r, $val0, undef, 'hours', $s),
                    evapoIndicator => $data->[$ii + 1],
                    evapoAmount    => { v => $data->[$ii + 2], u => 'kgm2' }
                }};
            }
            $ii += 3;
        } elsif (exists $r->{isSynop} && /\G004024 004024 012049/) {
            if (defined $data->[$ii + 2]) {
                my $s;

                $s = "@s[$ii .. $ii + 2]";
                push @{$r->{section3}}, { tempChange => {
                    s             => $s,
                    timeBeforeObs => _getPeriod($r, $val0, $data->[$ii + 1],
                                                'hours', $s),
                    tempChangeVal => { v => $data->[$ii + 2], u => 'C' }
                }};
            }
            $ii += 3;
        } elsif (   exists $r->{isSynop}
                 && /\G(?:008002 020054 ){3}/
                 && defined $val0            && $val0 == 7
                 && defined $data->[$ii + 2] && $data->[$ii + 2] == 8
                 && defined $data->[$ii + 4] && $data->[$ii + 4] == 9)
        {
            push @{$r->{section3}}, { cloudTypesDrift => {
                s => "@s[$ii .. $ii + 5]",
                _codeTable020054('Low', $data->[$ii + 1]),
                _codeTable020054('Middle', $data->[$ii + 3]),
                _codeTable020054('High', $data->[$ii + 5])
            }}
                if    defined $data->[$ii + 1]
                   || defined $data->[$ii + 3]
                   || defined $data->[$ii + 5];
            $ii += 6;
        } elsif (exists $r->{isSynop} && /\G004024 013012 /gc) {
# TODO?
# 013012: 98000000 has reference 0
#         all others have reference -2
            $val = /\G004024/ && ($data->[$ii + 2] // -1) == 0 ? 2 : 1;
            if (   defined $val0
                && defined $data->[$ii + 1]
                && defined $data->[$ii + $val])
            {
                _fixMeasurePeriod $r, \$val0, "@s[$ii, $ii + $val]";
                push @{$r->{section3}}, { snowFall => {
                    s             => "@s[$ii .. $ii + $val]",
                    timeBeforeObs => { hours => -$val0 },
                    precipAmount  => { v => $data->[$ii + 1] / 100, u => 'CM' }
                }};
            }
            $ii += 1 + $val;
        } elsif (   exists $r->{isBuoy}
                 && !exists $r->{qualityGroup2}
                 && /\G033022 033023 033027/
                 && defined $val0
                 && defined $data->[$ii + 1]
                 && defined $data->[$ii + 2])
        {
# TODO?
# code table 033027 0: 778 C tables have "RADIUS > 1500 M"
#                      all others have "RADIUS >= 1500 M"
            $r->{qualityGroup2} = {
                s                   => "@s[$ii .. $ii + 2]",
                qualityTransmission => $val0,
                qualityLocation     => $data->[$ii + 1],
                qualityLocClass     => $data->[$ii + 2]
            };
            $ii += 3;
        } elsif (exists $r->{isBuoy} && !exists $r->{drogueType} && /\G002034/){
            $r->{drogueType} = {
                s => $s0,
                $val0 =~ /^[0-5]$/ ? (drogueTypeInd => $val0)
                                   : (invalidFormat => $val0)
            }
                if defined $val0;
            $ii++;
        } elsif (   exists $r->{isBuoy}
                 && !exists $r->{cableLengthDrogue}
                 && /\G007070/)
        {
            $r->{cableLengthDrogue} = {
                s      => $s0,
                length => { v => $val0, u => 'M' }
            }
                if defined $val0;
            $ii++;
        } elsif (   exists $r->{isBuoy}
                 && !exists $r->{batteryVoltage}
                 && /\G025026/)
        {
            $r->{batteryVoltage} = {
                s                 => $s0,
                batteryVoltageVal => sprintf '%.1f', $val0
            }
                if defined $val0;
            $ii++;
        } elsif (   exists $r->{isBuoy}
                 && !exists $r->{submergence}
                 && /\G002190/)
        {
            $r->{submergence} = {
                s              => $s0,
                submergenceVal => $val0
            }
                if defined $val0;
            $ii++;
        } else {
            # replace non-IA5 characters
            if (defined $val0) {
                $s[$ii] =~ s/[^ -\x7e]/?/g;
                $not_processed_s .= ' ' . $s[$ii];
            }
            $is_processed = 0;
            $ii++;
        }
    }

    if ($temperature && $tempPrecision) {
        @{$temperature->{air}{temp}}{qw(rn rpi)} =
                                                ($tempPrecision, $tempPrecision)
            if exists $temperature->{air} && exists $temperature->{air}{temp};
        @{$temperature->{dewpoint}{temp}}{qw(rn rpi)} =
                                                ($tempPrecision, $tempPrecision)
            if    exists $temperature->{dewpoint}
               && exists $temperature->{dewpoint}{temp};
    }

    $r->{msg} = "@s[0 .. $ii_base, $ii_report .. $ii - 1]";
    push @{$r->{warning}}, {
        warningType => 'notProcessed',
        s           => substr $not_processed_s, 1
    }
        if $not_processed_s ne '';

    $ii = 0;
    $sall = "@s"
        if $#reports > 0;
    for my $r (@reports) {
        $ii++;

        $r->{multi} = { part => $ii, parts => $#reports + 1, s_all => $sall }
            if $#reports > 0;

        # force order for amdarObs
        for my $amdarObs (@{$r->{amdarObs}}) {
            for (qw(flightLvl aircraftLocation windAtPA rollAngleQuality
                    temperature turbulenceAtPA))
            {
                if (exists $amdarObs->{$_}) {
                    push @{$amdarObs->{sortedArr}}, { $_ => $amdarObs->{$_} };
                    delete $amdarObs->{$_};
                }
            }
        }

        # complete / force order for amdarInfo
        if (exists $r->{amdarInfo}) {
            if (keys(%{$r->{amdarInfo}}) == 1) {
                $r->{amdarInfo}{notAvailable} = undef;
            } else {
                for (qw(navSystem amdarSystem tempPrecision)) {
                    if (exists $r->{amdarInfo}{$_}) {
                        push @{$r->{amdarInfo}{sortedArr}},
                                                  { $_ => $r->{amdarInfo}{$_} };
                        delete $r->{amdarInfo}{$_};
                    }
                }
            }
        }
    }

    return @reports;
}

sub _bufrstr2report {
    return _bufr2report _str2arr;
}

=head1 SEE ALSO

L<https://metaf2xml.sourceforge.io/>

=head1 COPYRIGHT and LICENSE

copyright (c) 2012-2023 metaf2xml @ L<https://metaf2xml.sourceforge.io/>

This file is part of metaf2xml.

metaf2xml is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

metaf2xml is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metaf2xml.  If not, see L<https://www.gnu.org/licenses/>.

=cut

1;
