########################################################################
# metaf2xml/src2raw.pl 2.9
#   convert data from different sources to standard format
#
# copyright (c) 2011-2023 metaf2xml
#
# This file is part of metaf2xml.
#
# metaf2xml is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# metaf2xml is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with metaf2xml.  If not, see <https://www.gnu.org/licenses/>.
########################################################################

package metaf2xml::src2raw 2.009;

########################################################################
# some things strictly Perl
########################################################################
use warnings;
use 5.014_004;

########################################################################
# export the functions provided by this module
########################################################################
BEGIN {
    require Exporter;

    our @ISA       = qw(Exporter);
    our @EXPORT_OK = qw(start_file next_line start_bufr next_bufr);
}

END {}

sub VERSION {
    my ($p, $v) = @_;

    die "$p version $v required--this is version $metaf2xml::src2raw::VERSION\n"
        if $#_ == 1 && $v != $metaf2xml::src2raw::VERSION;
    return $metaf2xml::src2raw::VERSION;
}

use metaf2xml::bufr 2.009;

=encoding utf8

=head1 NAME

metaf2xml::src2raw - convert data from different sources to standard format

=head1 SYNOPSIS

 use metaf2xml::src2raw 2.009;

 metaf2xml::src2raw::start_file($format,$suppressNIL,$wmo_prefix);
 while (1) {
   my $line = <>;
   @msgs = metaf2xml::src2raw::next_line($line);
   # process contents of @msgs ...
   last unless defined $line;
 }

 metaf2xml::src2raw::start_bufr($wmo_prefix,$file_name,$filter_cb);
 while (1) {
   my ($hdr, $data, $desc, $assoc, $msg) = next_bufr(1);
   last unless defined $hdr;
   # process returned data ...
 }
 next_bufr();

=head1 DESCRIPTION

This Perl module contains functions to convert METAR, TAF, SYNOP, BUOY or AMDAR
information provided in files on various public Internet servers to messages
delimited by a newline and in the format specified by the I<WMO Manual No.
306>, without modifications due to distribution like providing the initial part
of messages only once for several messages or appending an "=" (equal sign) to
terminate a message.
It prepends parts of the WMO header according to the argument C<wmo_prefix>.
The output can then be used by the module C<metaf2xml::parser>.

=head1 ABBREVIATIONS

=over

=item noaa

National Oceanic and Atmospheric Administration

=item nws

the National Weather Service department of the NOAA

=item iws

the Internet Weather Source of the NWS

=item addsds

the Text Data Server API service of the Aviation Weather Center of the NOAA

=item cod

College of DuPage

=item isu

Iowa State University

=back

=cut

########################################################################
# define some regular expressions
########################################################################
my $re_id_synop = '(?:S[IMN][A-Z]{2}\d{2})';
my $re_id_buoy = '(?:SS[A-Z]{2}\d{2})';
my $re_id_amdar = '(?:UD[A-Z]{2}\d{2})';
my $re_id_metaf = '(?:(?:S[AP]|F[CT])[A-Z]{2}\d{2})';
my $re_day  = '(?:0[1-9]|[12]\d|3[01])';
my $re_hour = '(?:[01]\d|2[0-3])';
my $re_min  = '[0-5]\d';
my $re_header_synop = '(?:AAXX {0,2}'. $re_day . $re_hour .'[0134/]|BBXX|OOXX)';
my $re_header_buoy = '(?:ZZYY (?://///|\d{5}))';
my $re_header_amdar = '(?:(?:AMDAR|ARP) ' . $re_day . $re_hour . ')';
my $re_ICAO = '(?:[A-Z][A-Z\d]{3})';
my $re_header_metaf = "(?:(?:COR +|AMD +|ADV +|METAR +|TAF +|SPECI +)*$re_ICAO )";
my $re_header_sao = "(?:[A-Z]{3,4} (NIL|(?:S[AP]|RS)(?: COR)? $re_hour$re_min (?:AUTO|AWOS|SAWR)))";
my $re_start_amdar = '[A-Z/]{3} [A-Z]{2}[A-Z0-9]{2,6}';

########################################################################
# variables to be preserved over function calls
########################################################################
my ($format, $suppressNIL, $wmo_prefix,
    $state, $wmo_header, $wmo_header_fix, $msg_header, $msg, $line_no,
    $report_id, $dissem, $date, @return_msg, $bufr);

sub _add_msg {
    my $m = $wmo_header . shift;

    $m =~ s/  +/ /g;
    $m =~ s/[ =]*$//;
    push @return_msg, $m;
    return;
}

# special cases:
#   TAFPA TAF PHLI 292334Z ...
#   TAFPQ TAF PGUM 292347Z ...
#   ONTAF VTSP 300330Z ...
#   TFFR 300600Z METAR VRB01KT ...
sub _add_metaf_msg {
    my $m = shift;

    $m =~ s/  +/ /g;
    $m =~ s/^ +//;
    $m =~ s/ +$//;
    $m =~ s/^(?:PART \d+ OF \d+ PARTS |ONTAF |TAFP[AQ] )+//;
    $m =~ s/^(METAR|SPECI|TAF(?: AMD| COR)?) \1 /$1 /;
    $m =~ s/^(TAF (?:AMD |COR )?)(?:TAF |COR |AMD )*/$1/;
    $m =~ s/^(?=[A-Z]{3} (?:NIL|(?:S[AP]|RS)(?: COR)? $re_hour$re_min (?:AUTO|AWOS|SAWR)))/C/;

    _add_msg $m
        unless $suppressNIL
           && $m =~ /^$re_header_metaf(?:$re_day$re_hour${re_min}Z (?:$re_day(?:$re_hour|24){2} )?)?(RMK )?NIL[= ]*$/;
    return;
}

sub _add_synop_msg {
    my $m = shift;

    _add_msg $m
        if $m =~ /^$re_header_synop +[^ ]{3,}/o;
    return;
}

sub _add_buoy_msg {
    my $m = shift;

    _add_msg $m
        if $m =~ /^$re_header_buoy +[^ ]{3,}/o;
    return;
}

# WMO header: T1T2A1A2[ii] CCCC YYGGgg [BBB]
# https://community.wmo.int/en/ahl-table-definitions
# https://community.wmo.int/en/activity-areas/wis/wis-manuals:
#   Manual on the Global Telecommunication System, WMO-No. 386, Attachment II-5:
#   T1T2:   data type (e.g. SM: main synoptic hour reports, SA: METAR)
#   A1A2:   geographical and/or data type and/or time designators
#   ii:     optional. 2 digits to differentiate bulletins with the same
#           T1T2A1A2 and CCCC
#   CCCC:   station or centre originating or compiling the bulletin
#   YYGGgg: date-time group
#           YY: day of month
#           GGgg: standard time reports:
#                     standard time of observation (UTC)
#                 aeronautical forecasts:
#                     GG: hour (UTC) of transmission time, gg = 00
#                 other forecasts:
#                     standard time of observation (UTC) on which forecast
#                     is based
#                 other messages: time of compilation (UTC)
#   BBB:    optional. indicator for additional/corrected/amended bulletins
#
#   $opt_H: 0 -> (empty)
#           1 -> T1T2
#           2 -> T1T2A1A2[ii]
#           3 -> T1T2A1A2[ii] CCCC
#           4 -> T1T2A1A2[ii] CCCC YYGGgg
#           5 -> T1T2A1A2[ii] CCCC YYGGgg [BBB]
sub _format_wmo_header {
    my $hdr = shift;

    return ''
        if $wmo_prefix <= 0;
    return ($hdr =~ /(((((..)[^ ]+) [^ ]+) \d+)(?: COR| (?:RR|CC|AA)[A-Z]| P[A-Z]{2})?)/)[5 - $wmo_prefix] . ' ';
}

#
# parse METAR/TAF file
#
sub _process_metaf_file {
    my $re_SOM;

    local $_ = shift;
    $re_SOM = shift;

    if (defined $_) {
        s/[\n\r\cA\cC\cZ]+//g;
        s/\xC2\xA0//g;

        # check for trailing encoded(!) '=' or ' '
        s/[^ ]\K=3D(?:=20)?=? *$/=/;
        s/[^ ]\K=20 *$//;

        return
            if $_ eq 'TX OPMET V1';

        # * -> SOM
        if (/$re_SOM/) {
            my $have_prev = exists $+{prev};

            # final '=' missing in previous block or missing new line?
            if (   ($state eq 'STARTED' || $state eq 'Sx')
                && ($msg ne '' || exists $+{prev}))
            {
                $msg .= $+{prev}
                    if exists $+{prev};
                $have_prev = 0;
                if (   $msg =~ /^$re_header_metaf *[^ ]{3,}/o
                    || ($msg ne 'TAF NIL' && $msg =~ /^$re_header_sao/o))
                {
                    _add_metaf_msg "$msg_header$msg";
                } else {
                    #print STDERR "IGNORED 1: $line_no: $msg\n" if $msg ne '';
                }
                $msg = '';
            }
            #print STDERR "IGNORED 2: $line_no: $msg\n" if $msg ne '';
            #print STDERR "IGNORED 2: $line_no: $+{prev}\n" if $have_prev;
            $state = 'SOM';

        # SOM -> Sx
        # SOM: KAWN uses SA(GL|EU|EW|UK|..), FT(MX), ..
        # SOM: SAST uses SAAG
        } elsif (   $state eq 'SOM'
                 && (   /^$re_id_metaf (?<dissem>[A-Z\d]{4}) \d{6}/o
                     || /^(?:S[AP]|F[CT])[A-Z]{2} (?<dissem>KAWN|SAST) \d{6}/))
        {
            $state      = 'Sx';
            $dissem     = $+{dissem};
            $wmo_header = _format_wmo_header $_;
            $msg_header = '';

        # Sx: K*, PANC, PGUM, PHFO, TJSJ use (MTR|TAF)[A-Z\d]{3}
        # Sx: UHHH, WMKK use 'METAR YYGGggZ'
        } elsif (   $state eq 'Sx'
                 && (/^(?:MTR|TAF)[A-Z\d]{3}$/ || /^METAR \d{6}Z$/))
        {
            # skip this line
            #print STDERR "IGNORED 3: $line_no: $_\n";

        # Sx, STARTED: (MSGSTART +) MSGEND -> Sx
        } elsif (($state eq 'STARTED' || $state eq 'Sx') && /= *$/) {
            if ($state eq 'STARTED' && /^(?:METAR |SPECI |TAF )?(?:ADV )?$re_ICAO (?:$re_day$re_hour${re_min}Z |NIL ?(?:= ?)?$)/)
            {
                # new message starts, previous didn't end with =
                if (   $msg =~ /^$re_header_metaf *[^ ]{3,}/o
                    || ($msg ne 'TAF NIL' && $msg =~ /^$re_header_sao/o))
                {
                    _add_metaf_msg "$msg_header$msg";
                } else {
                    #print STDERR "IGNORED 4: $line_no: $msg\n" if $msg ne '';
                }
                $msg = '';
            }
            /^(?<all>.*)= *$/;
            $msg .= $+{all};
            $msg_header = '' if /^(?:METAR|SPECI|TAF) /;
            if (   $msg =~ /^$re_header_metaf *[^ ]{3,}/o
                || ($msg ne 'TAF NIL' && $msg =~ /^$re_header_sao/o))
            {
                _add_metaf_msg "$msg_header$msg";
            } else {
                #print STDERR "IGNORED 5: $line_no: $msg\n" if $msg ne '';
            }
            $msg   = '';
            $state = 'Sx';

        # Sx, STARTED: MSGSTART -> STARTED
        } elsif (   (   ($state eq 'STARTED' || $state eq 'Sx')
                     && /^(?<all>(?:METAR|SPECI|TAF)(?: AMD| COR)*(?: +(?<msg>[^ ].*))?)/)
                 || ($state eq 'Sx' && /^ *(?<all>(?<msg>[^ ].*))/)) # METAR|SPECI|TAF missing
        {
            if (exists $+{msg}) {
                $msg = $+{all};
                # some disseminators always break msgs at column 69
                # TODO: or rather in about 95 % of the cases ...
                $msg .= ' '
                    unless    (   $dissem eq 'WAAA'
                               || $dissem eq 'HRYR'
                               || substr($dissem, 0, 2) eq 'SB')
                           && length $+{all} == 69;
                $state = 'STARTED';
                $msg_header = '' if /^(?:METAR|SPECI|TAF) /;
            } else {
                $msg_header = "$+{all} ";
            }

        # STARTED: MSGCONTD
        } elsif ($state eq 'STARTED' && /[^ ]/) {
            $msg .= $_;
            # some disseminators always break msgs at column 69
            # TODO: or rather in about 95 % of the cases ...
            $msg .= ' '
                unless    (   $dissem eq 'WAAA'
                           || $dissem eq 'HRYR'
                           || substr($dissem, 0, 2) eq 'SB')
                       && length $_ == 69;
        }
    } else {
        # end of message could be missing in last block
        _add_metaf_msg "$msg_header$msg"
            if $state eq 'STARTED' && $msg;
    }
    return;
}

#
# parse SYNOP file
#
sub _process_synop_file {
    my $re_SOM;

    local $_ = shift;
    $re_SOM = shift;

    if (defined $_) {
        s/[\n\r\cA\cC\cZ]+//g;
        s/\xC2\xA0//g;

        # check for trailing encoded(!) '=' or ' '
        s/[^ ]\K=3D(?:=20)?=? *$/=/;
        s/[^ ]\K=20 *$//;

        return
            if /^NNNN/ || ($suppressNIL && /NIL=$/);

        # * -> SOM
        if (/$re_SOM/) {
            # final '=' missing in previous block or missing new line?
            _add_synop_msg $msg_header . $msg . ($+{prev} // '')
                if    ($state eq 'STARTED' || $state eq 'Sx')
                   && ($msg ne '' || exists $+{prev});
            $msg        = '';
            $msg_header = '';
            $state      = 'SOM';

        # ? -> Sx
        } elsif (/^(?<id>$re_id_synop) (?<dissem>[A-Z\d]{4}) (?<date>\d{4})/o) {
            ($report_id, $dissem, $date) = @+{qw(id dissem date)};
            $wmo_header = _format_wmo_header $_;
            $state      = 'Sx';

        # Sx, STARTED -> STARTED
        } elsif (   ($state eq 'Sx' || $state eq 'STARTED')
                 && /$re_header_synop(?: +[^ ].*| *)$/o)
        {
            _add_synop_msg "$msg_header$msg"
                if $msg;

            s/^AAXX\K */ /;
            /(?<hdr>$re_header_synop)(?: +(?<msg>[^ ].*)| *)$/o;
            $state = 'STARTED';
            if (exists $+{msg}) {
                $msg        = "$+{hdr} $+{msg} ";
                $msg_header = '';
            } else {
                $msg        = '';
                $msg_header = "$+{hdr} ";
            }

        # Sx, STARTED -> '' for invalid header
        } elsif (($state eq 'Sx' || $state eq 'STARTED') && /AAXX/)
        {
            $state = '';

        # Sx -> STARTED
        } elsif ($state eq 'Sx') {
            if ($dissem eq 'KGYX' && /^SSMMWN/) {
                # KGYX omits 'AAXX YYGGiw' (and has no '=')
                $state      = 'STARTED';
                $msg        = '';
                $msg_header = "AAXX ${date}4 ";
            } elsif ($dissem eq 'PTYA' && /^SSYM2 (?:.{5})/) {
                # PTYA sometimes uses 'SSYM2' instead of 'AAXX' (and has no '=')
                $state      = 'STARTED';
                $msg        = '';
                $msg_header = "AAXX ${date}4 ";
            }

        # STARTED: MSGSTART SxPF80 KWBC
        # SxPF80 KWBC does not use delimiters, check for IIiii + (NIL | iRixhVV)
        } elsif (   $state eq 'STARTED'
                 && $dissem eq 'KWBC'
                 && substr($report_id, 2) eq 'PF80'
                 && m{^919(?:2[59]|4[358]|5[48]) (?:NIL|[0-4/][1-7/][\d/]{3} )})
        {
            _add_synop_msg "$msg_header$msg"
                if $msg;
            $msg = "$_ ";

        # STARTED: MSGEND
        # - MDSD uses empty lines as delimiter
        # - PTKK sometimes has no delimiters, last group is (555 .* 9YYGG | NIL)
        } elsif (   $state eq 'STARTED'
                 && (   /^(?<msg>.*)= *$/                   # end of message
                     || ($dissem eq 'MDSD' && /^ *$/)       # MDSD && empty line
                     || (   $dissem eq 'PTKK'               # PTKK && ...
                         && (   /^ *(?<msg>.* NIL)\s*$/     #  ... NIL
                             || (   "$msg $_" =~ / 555 /    #  ... 555 .* 9YYGG
                                 && /^ *(?<msg>.* 9$date)\s*$/)))))
        {
            $msg = $msg_header . $msg . ($+{msg} // '');
            _add_synop_msg $msg;
            $msg = '';

        # STARTED: MSGCONTD
        } elsif ($state eq 'STARTED' && /^ *(?<msg>.+)/) {
            $msg .= $+{msg};
            # MXBA uses fixed width format
            $msg .= ' ' unless $dissem eq 'MXBA';
        }
    } else {
        # end of message could be missing in last block
        _add_synop_msg "$msg_header$msg"
            if $state eq 'STARTED' && $msg;
    }
    return;
}

#
# parse METAR/STAF/TAF cycle file from NOAA/IWS
#
# keyword SPECI not provided
#
# special cases:
#   TAF GCLP 292300Z 300606 01016KT CAVOK= TAF GCXO 292300Z ...
#   TAF HECA 300400Z 300606 VRB03KT ... CAVOK= HELX 300400Z ...
#   DRZR 300600Z 00000KT CAVOK 24/23 Q1013 METAR DNAA 300600Z ...
sub _process_metaf_cycle_iws {
    local $_ = shift;

    if (defined $_) {
        s/[\n\r\cC\cZ]+//g;

        # check for trailing encoded(!) '=' or ' '
        s/[^ ]\K=3D(?:=20)?=? *$/=/;
        s/[^ ]\K=20 *$//;

        if (/^ *$/) {                                     # end of message?
            _add_metaf_msg $msg;
            $msg = '';
        } elsif (   /(?<prev>.*)= ?TAF (?<msg>.*)/        # TAF after '='?
                 || /(?<prev>.*)= ?(?<msg>$re_ICAO .*)/   # ICAO code after '='?
                 || /(?<prev>.*) METAR (?<msg>$re_ICAO .*)/) # METAR+ICAO code?
        {
            _add_metaf_msg "$msg $+{prev}";
            $msg = $+{msg};
        } elsif (/^[^\d]/) {                 # start or continuation of message?
            $msg .= $_;
        } elsif (m{^\d{4}/(?:0[1-9]|1[012])/(?<day>$re_day) (?<hour>$re_hour):(?<min>$re_min)$})
        {
            # time of report dissemination is N/A, this is the observation time
            $wmo_header = _format_wmo_header   $wmo_header_fix
                                             . $+{day} . $+{hour} . $+{min};
        }
    } else {
        _add_metaf_msg $msg
            if $msg;
    }
    return;
}

#
# parse BUOY file
#
sub _process_buoy_file {
    my $re_SOM;

    local $_ = shift;
    $re_SOM = shift;

    if (defined $_) {
        s/[\n\r\cA\cC\cZ]+//g;
        s/\xC2\xA0//g;

        # check for trailing encoded(!) '=' or ' '
        s/[^ ]\K=3D(?:=20)?=? *$/=/;
        s/[^ ]\K=20 *$//;

        # * -> SOM
        if (/$re_SOM/) {
            # final '=' missing in previous block or missing new line?
            if (   ($state eq 'STARTED' || $state eq 'Sx')
                && ($msg ne '' || exists $+{prev}))
            {
                _add_buoy_msg $msg . ($+{prev} // '');
                $msg = '';
            }
            $state = 'SOM';

        # ? -> Sx
        } elsif (/^(?<id>$re_id_buoy) (?<dissem>[A-Z\d]{4}) (?<date>\d{4})/o) {
            ($report_id, $dissem, $date) = @+{qw(id dissem date)};
            $wmo_header = _format_wmo_header $_;
            $state      = 'Sx';

        # Sx, STARTED -> STARTED
        } elsif (   ($state eq 'Sx' || $state eq 'STARTED')
                 && /^$re_header_buoy(?: +[^ ].*| *)$/o)
        {
            _add_buoy_msg $msg
                if $msg;
            $state = 'STARTED';
            $msg = "$_ ";

        # STARTED: MSGCONTD
        } elsif ($state eq 'STARTED' && /^ *(?<msg>[^ ].*)/) {
            $msg .= "$+{msg} ";
        }
    } else {
        # end of message could be missing in last block
        _add_buoy_msg $msg
            if $state eq 'STARTED' && $msg;
    }
    return;
}

#
# parse AMDAR file
#
sub _process_amdar_file {
    my $re_SOM;

    local $_ = shift;
    $re_SOM = shift;

    if (defined $_) {
        s/[\n\r\cA\cC\cZ]+//g;
        s/\xC2\xA0//g;

        # check for trailing encoded(!) '=' or ' '
        s/[^ ]\K=3D(?:=20)?=? *$/=/;
        s/[^ ]\K=20 *$//;

        # * -> SOM
        if (/$re_SOM/) {
            # final '=' missing in previous block or missing new line?
            if (   ($state eq 'STARTED' || $state eq 'Sx')
                && ($msg ne '' || exists $+{prev}))
            {
                $msg = $msg_header . $msg . ($+{prev} // '');
                _add_msg $msg
                    if $msg =~ m{^$re_header_amdar $re_start_amdar }o;
                $msg = '';
            }
            $state = 'SOM';

        # ? -> Sx
        } elsif (/^(?<id>$re_id_amdar) (?<dissem>[A-Z\d]{4}) (?<date>\d{4})/o) {
            ($report_id, $dissem, $date) = @+{qw(id dissem date)};
            $wmo_header = _format_wmo_header $_;
            $state      = 'Sx';

        # Sx, STARTED -> STARTED
        } elsif (   ($state eq 'Sx' || $state eq 'STARTED')
                 && /^$re_header_amdar(?: +[^ ].*| *)$/o)
        {
            $msg = "$msg_header$msg";
            _add_msg $msg
                if $msg =~ m{^$re_header_amdar $re_start_amdar }o;

            /(?<hdr>$re_header_amdar)(?: +(?<msg>[^ ].*)| *)$/o;
            $state = 'STARTED';
            if (exists $+{msg}) {
                $msg        = "$+{hdr} $+{msg} ";
                $msg_header = '';
            } else {
                $msg        = '';
                $msg_header = "$+{hdr} ";
            }

        # STARTED: MSGEND
        } elsif ($state eq 'STARTED' && /=/) { # end of message
            while (s/^(?<msg>[^=]*)= *//) {
                $msg = $msg_header . $msg . $+{msg};
                _add_msg $msg
                    if $msg =~ m{^$re_header_amdar $re_start_amdar }o;
                $msg = '';
            }

        # STARTED: MSGCONTD
        } elsif ($state eq 'STARTED' && /^ *(?<msg>[^ ].*)/) {
            $msg .= "$+{msg} ";
        }
    } else {
        # end of message could be missing in last block
        $msg = "$msg_header$msg";
        _add_msg $msg
            if    $state eq 'STARTED'
               && $msg =~ m{^$re_header_amdar $re_start_amdar }o;
    }
    return;
}

=head1 SUBROUTINES/METHODS

=cut

########################################################################
# start_file
########################################################################
sub start_file {

    ($format, $suppressNIL, $wmo_prefix) = @_;

    # init global variables
    $state      = '';
    $wmo_header = '';
    $msg_header = '';
    $msg        = '';
    $line_no    = 0;
    $report_id  = '';
    $dissem     = '';
    $date       = '';

=head2 start_file($format,$suppressNIL,$wmo_prefix)

This function must be called before processing a file.
The following arguments are expected:

=over

=item format

origin of the input data:
metar_cycle_iws, staf_cycle_iws, taf_cycle_iws,
metaf_nws, metaf_cod, metaf_isu,
synop_nws, synop_cod, synop_isu,
buoy_nws, buoy_isu,
amdar_nws, amdar_isu

=cut

    return "ERROR: invalid input data format: '$format'."
        unless $format =~
      /^(?:(?<type>metar|s?taf)_cycle_iws|(?:metaf|synop)_(?:nws|cod|isu)|buoy_(?:nws|isu)|amdar_(?:nws|isu))$/;
    if (exists $+{type}) {
        $wmo_header_fix
            =   { metar => 'SA', staf => 'FC', taf => 'FT' }->{$+{type}}
              . 'XX NOAA ';
        $wmo_header = _format_wmo_header "${wmo_header_fix}010000";
    }

=item suppressNIL

if I<true>, messages that only contain C<NIL> are not returned

=item wmo_prefix

which parts of the WMO header (C<T1T2A1A2[ii] CCCC YYGGgg [BBB]>) to prepend
to each message:

=over

=item C<0>

(no header) (default)

=item C<1>

C<T1T2> (data type):

=over

=item SA

aviation routine reports (METAR)

=item SP

special aviation weather reports (SPECI)

=item FT

aerodrome forecast (TAF) with forecast period >= 12 hours

=item FC

aerodrome forecast (TAF) with forecast period < 12 hours

=item SM

surface data (SYNOP) at main hours (00:00, 06:00, 12:00, 18:00 UTC)

=item SI

surface data (SYNOP) at intermediate (main + 3) hours

=item SN

surface data (SYNOP) at non-standard (other than main and intermediate) hours

=item SS

buoy observations (BUOY)

=item UD

aircraft reports (AMDAR)

=back

=item C<2>

C<T1T2A1A2[ii]> (like C<1> plus geographical/data type/time designators
plus 2-digit bulletin differentiator (optional))

=item C<3>

C<T1T2A1A2[ii] CCCC> (like C<2> plus station or centre originating or
compiling the bulletin)

=item C<4>

C<T1T2A1A2[ii] CCCC YYGGgg> (like C<3> plus report
observation/transmission/compilation time (UTC))

=item C<5>

C<T1T2A1A2[ii] CCCC YYGGgg [BBB]> (like C<4> plus indicator for
additional/corrected/amended bulletins (optional))

=back

=back

The function will return 0 on success, or a string describing the error which
occurred.

=cut

    $wmo_prefix = 0
        unless defined $wmo_prefix && $wmo_prefix =~ /^[0-5]$/;

    return 0;
}

########################################################################
# next_line
########################################################################
sub next_line {
    my $line = shift;

=head2 next_line($line)

This function must be called for each line in a file and once with the
argument C<undef> at the end of file.

The following arguments are expected:

=over

=item line

string that contains a line from the file, or C<undef> after the last line of
the file

=back

The return value is an array of 0, 1 or 2 messages, or an array with one
element which is C<undef> if
L<start_file()|/start_file($format,$suppressNIL,$wmo_prefix)> was not called
(never, or after L<next_line()|/next_line($line)> was called with C<undef>).

=cut

    $line_no++;

    return (undef)
        unless $format;

    @return_msg = ();
    { metar_cycle_iws => 1, taf_cycle_iws => 1, staf_cycle_iws => 1 }->{$format}
    ? _process_metaf_cycle_iws($line)
    : $format eq 'metaf_nws'
    ? _process_metaf_file($line, '(?<prev>.+)?####\d{9}####$')
    : { metaf_cod => 1, metaf_isu => 1 }->{$format}
    ? _process_metaf_file($line, '^\d{3} $')
    : $format eq 'synop_nws'
    ? _process_synop_file($line, '(?<prev>.+)?####\d{9}####$')
    : { synop_cod => 1, synop_isu => 1 }->{$format}
    ? _process_synop_file($line, '^\d{3} $')
    : $format eq 'buoy_nws'
    ? _process_buoy_file($line, '(?<prev>.+)?####\d{9}####$')
    : $format eq 'buoy_isu'
    ? _process_buoy_file($line, '^\d{3} $')
    : $format eq 'amdar_nws'
    ? _process_amdar_file($line, '(?<prev>.+)?####\d{9}####$')
    : { amdar_nws => 1, amdar_isu => 1 }->{$format}
    ? _process_amdar_file($line, '^\d{3} $')
    : do {};

    # after EOF, start_file is required again
    $format = ''
        if !defined $line;

    return @return_msg;
}

########################################################################
# start_bufr
########################################################################
sub start_bufr {
    my ($my_wmo_prefix, $file_name, $filter_cb) = @_;
    state $have_Geo_BUFR;

    $wmo_prefix = $my_wmo_prefix;

    # init global variables
    $format     = 'bufr';
    $state      = '';
    $wmo_header = '';
    $msg_header = '';
    $msg        = '';
    $report_id  = '';
    $dissem     = '';
    $date       = '';

=head2 start_bufr($wmo_prefix,$file_name,$filter_cb)

This function must be called to start processing a BUFR file.
The following arguments are expected:

=over

=item wmo_prefix

Like the argument C<wmo_prefix> for
L<start_file()|/start_file($format,$suppressNIL,$wmo_prefix)>,
except for the possible values returned for C<T1T2>:

=over

=item IS

observational surface/sea level data encoded as BUFR

if C<T1T2A1> is C<ISM>/C<ISI>/C<ISN>:
surface data (SYNOP) at main/intermediate/non-standard hours

=item IO

observational oceanographic/limnographic data encoded as BUFR

if C<T1T2A1> is C<IOB>: buoy observations (BUOY)

=item IU

observational upper-air data encoded as BUFR

if C<T1T2A1> is C<IUA>: aircraft reports (AMDAR)

=back

=item file_name

file name for a (plain) file with binary BUFR message.

=item filter_cb

reference to a filter callback function

The callback function is invoked for each BUFR message before decoding
BUFR section 4 (Data Section). The following argument is passed:

=over

=item C<bufr>

the current C<Geo::BUFR> object

=back

The callback function should return one the following values:

=over

=item I<false>

message should not be filtered out

=item I<true>

message should be filtered out

=back

=back

The function will return 0 on success, or a string describing the error which
occurred.

=cut

    $wmo_prefix = 0
        unless defined $wmo_prefix && $wmo_prefix =~ /^[0-5]$/;

    if (!$have_Geo_BUFR) {
        my $rc;

        $have_Geo_BUFR = 1;
        $rc = eval {
            local $SIG{'__DIE__'}; # don't call user's __DIE__ hooks
            require Geo::BUFR;
            Geo::BUFR->VERSION >= 1.37;
        };
        return "ERROR: cannot load Geo::BUFR 1.37+: $@\n"
            if !$rc || $@;
    }

    Geo::BUFR->set_tableformat('ECCODES');
    Geo::BUFR->set_tablepath($ENV{BUFR_TABLES});
    $bufr = Geo::BUFR->new();
    $bufr->set_strict_checking(0);
    $bufr->set_show_all_operators(0);
    $bufr->set_filter_cb($filter_cb)
        if ref $filter_cb eq 'CODE' && $bufr->can('set_filter_cb');
    $bufr->reuse_current_ahl(1)
        if $bufr->can('reuse_current_ahl');

    $bufr->fopen($file_name)
        or return "ERROR: cannot open '$file_name'\n";

    return 0;
}

########################################################################
# next_bufr
########################################################################
sub next_bufr {
    my $get_more = shift;

=head2 next_bufr($get_more)

This function must be called for each line in a file and once with an
value of I<false> to end processing.

The following arguments are expected:

=over

=item get_more

I<false> if processing of current file should be ended, I<true> otherwise

=back

The returned value is a decoded BUFR message, or C<undef> if
L<start_bufr()|/start_bufr($wmo_prefix,$file_name,$filter_cb)> was not called
(never, or after L<next_bufr()|/next_bufr($get_more)> was called with I<false>).

=cut

    return
        unless $format;

    @return_msg = ();

    if (!$get_more) {
        $bufr->fclose();
        $format = ''; # start_bufr is required again
        return;
    }

    while (1) {
        my ($rc, $data, $desc, $assoc);

        $rc = eval {
            local $SIG{'__DIE__'}; # don't call user's __DIE__ hooks
            ($data, $desc) = $bufr->next_observation();
            1;
        };
        if (!$rc || $@) {
            $wmo_header = '';
            if ($@) {
                _add_msg "ERROR: $@";
            } else {
                _add_msg 'ERROR: rc: "' . ($rc // '(undef)') . '"' . "\n";
            }
            last;
        }

        if ($bufr->eof) {
            $bufr->fclose();
            $format = ''; # start_bufr is required again
            last;
        }

        if (!$bufr->can('is_filtered') || !$bufr->is_filtered()) {
            $wmo_header = _format_wmo_header $bufr->{CURRENT_AHL};
            ($data, $desc, $assoc) = metaf2xml::bufr::_prepare $data, $desc;
            _add_msg join ' ', map { $desc->[$_] . metaf2xml::bufr::_val2s
                                                       $data->[$_], $assoc->[$_]
                                   } (0 .. $#$desc);
            last;
        }
    }

    return $return_msg[0];
}

=head1 DEPENDENCIES

If binary BUFR messages are to be processed, the Perl module C<Geo::BUFR> with
version 1.37+ and ECC BUFR table files from ECWMF are required.

=head1 SEE ALSO

L<metaf2xml::parser|metaf2xml::parser>(3pm),
L<metafsrc2raw|metafsrc2raw>(1), L<metaf|metaf>(1),
L<https://metaf2xml.sourceforge.io/>

=head1 COPYRIGHT and LICENSE

copyright (c) 2011-2023 metaf2xml @ L<https://metaf2xml.sourceforge.io/>

This file is part of metaf2xml.

metaf2xml is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

metaf2xml is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metaf2xml.  If not, see L<https://www.gnu.org/licenses/>.

=cut

1;
