#!/usr/bin/perl

########################################################################
# metafsrc2raw.pl 2.9
#   example script to convert data from different sources to standard format
#
# copyright (c) 2009-2020 metaf2xml
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
########################################################################

########################################################################
# some things strictly Perl
########################################################################
use warnings;
use 5.014_004;

=encoding utf8

=head1 NAME

metafsrc2raw.pl - convert data from different sources to standard format

=head1 SYNOPSIS

 metafsrc2raw.pl -F<format> [OPTION]... [FILE]...

=head1 DESCRIPTION

This script is an example interface to the Perl module C<metaf2xml::src2raw>.
It converts METAR, TAF, SYNOP, BUOY or AMDAR information provided in files on
various public Internet servers to messages delimited by a newline and in the
format specified by the I<WMO Manual No. 306>, without modifications due to
distribution like providing the initial part of messages only once for several
messages, or appending an "=" (equal sign) to terminate a message.
It prepends parts of the WMO header according to the option -H.
The output can then be used by the module C<metaf2xml::parser>.

If no input files are specified, then the standard input is read.

Additionally, files with binary BUFR messages can be decoded (using the Perl
module C<Geo::BUFR>) to a text format which can be processed by the module
C<metaf2xml::parser>.

=head1 DEPENDENCIES

The Perl module C<metaf2xml::src2raw> is required.

=head1 OPTIONS

=over

=item -v

print version of metafsrc2raw.pl and exit

=item -F I<format>

origin of the input data: for allowed values and their meaning see the
parameter L<metaf2xml::src2raw/format>(3pm). Additionally, I<format> can be
C<bufr>.

=item -D

suppress duplicate messages

=item -N

suppress messages that only contain C<NIL>

=item -H I<0..5>

which parts of the WMO header (C<T1T2A1A2[ii] CCCC YYGGgg [BBB]>) to prepend
to each message: for allowed values and their meaning see the parameter
L<metaf2xml::src2raw/wmo_prefix>(3pm)

=back

=head1 ARGUMENTS

Names for input files can be given as arguments. If no arguments are given,
input is read from standard input.

=head1 EXIT STATUS

If an invalid option was provided, or the option -F was not provided, or the
function L<start_file() in metaf2xml::src2raw|metaf2xml::src2raw/start_file-format-suppressNIL-wmo_prefix->(3pm) returned an error, the
script will exit with status 1, otherwise it will exit with status 0.

=head1 EXAMPLES

Parse SYNOP messages from NOAA/NWS, suppressing NIL messages and duplicates:

 metafsrc2raw.pl -DNF synop_nws << "EOF"
####018000228####
SMPY01 SGAS 060600
AAXX 06064
86134 32970 02402 10250 20235 40105 56005=
86297 NIL=
86134 32970 02402 10250 20235 40105 56005=
86218 32970 01810 10240 20209 40103 58011=
####018000158####
SMFG01 SOCA 060600
AAXX 06064
81405 01259 60505 10241 20229 30105 40114 60071 76086 86200
  333 32024 55048 60067 58002 70098 82705 86813
  555 69905=####018000118####
SMCS01 MROC 060600
AAXX 06064
78762 32970 00915 10220 20160 39094 40101 57005
333 10280 20193 30/// 59006=
EOF

results in:

 AAXX 06064 86134 32970 02402 10250 20235 40105 56005
 AAXX 06064 86218 32970 01810 10240 20209 40103 58011
 AAXX 06064 81405 01259 60505 10241 20229 30105 40114 60071 76086 86200 333 32024 55048 60067 58002 70098 82705 86813 555 69905
 AAXX 06064 78762 32970 00915 10220 20160 39094 40101 57005 333 10280 20193 30/// 59006

Parse METAR/SPECI messages from the College of DuPage with the WMO header
C<TTAAII> prepended:

 metafsrc2raw.pl -F metaf_cod -H2 << "EOF"
355 
SAUS70 KWBC 161200
METAR
KDZJ 161146Z AUTO 00000KT 10SM -DZ SCT002 BKN075 OVC100 09/09
     A3013 RMK AO2=
KLDJ 161150Z AUTO 00000KT 10SM CLR 03/M01 A3027 RMK AO2=
KMLJ 161145Z AUTO 00000KT 4SM RA BKN005 BKN012 OVC033 12/12 A3017
     RMK AO2=
351 
SPCN51 CWAO 161254
SPECI CWCL 161254Z AUTO 19008KT 9SM FEW008 FEW015 BKN027 BKN043
M04/M06 A2947=
EOF

results in:

 SAUS70 METAR KDZJ 161146Z AUTO 00000KT 10SM -DZ SCT002 BKN075 OVC100 09/09 A3013 RMK AO2
 SAUS70 METAR KLDJ 161150Z AUTO 00000KT 10SM CLR 03/M01 A3027 RMK AO2
 SAUS70 METAR KMLJ 161145Z AUTO 00000KT 4SM RA BKN005 BKN012 OVC033 12/12 A3017 RMK AO2
 SPCN51 SPECI CWCL 161254Z AUTO 19008KT 9SM FEW008 FEW015 BKN027 BKN043 M04/M06 A2947

=head1 SEE ALSO

L<metaf2xml::parser|metaf2xml::parser>(3pm),
L<metaf2xml::src2raw|metaf2xml::src2raw>(3pm),
L<https://metaf2xml.sourceforge.io/>

=head1 COPYRIGHT and LICENSE

copyright (c) 2009-2020 metaf2xml @ L<https://metaf2xml.sourceforge.io/>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see L<https://www.gnu.org/licenses/>.

=cut

use Getopt::Std;
use Pod::Usage 'pod2usage';
use File::Basename 'dirname';

our $VERSION = '2.9';

# path will be set by install
use lib dirname($0); # METAF2XML_LIB
use metaf2xml::src2raw 2.009 qw(start_file next_line start_bufr next_bufr);

sub usage {
    my $msg = shift;

    if ($msg eq '') {
        pod2usage -exitval => 'NOEXIT', -output => \*STDERR, -noperldoc => 1;
        print {*STDERR} <<"EOF";
Options may be merged together.  -- stops processing of options.
Space is not required between options and their arguments.
EOF
    } else {
        print {*STDERR} $msg;
    }

    print {*STDERR} <<"EOF";

For more details run
        perldoc -F $0
EOF
    exit 1;
}

sub HELP_MESSAGE { usage ''; return; } # for Getopt::Std

sub do_filter {
    my $self = shift;
    my ($cat, $sub);

    $cat = $self->get_data_category;
    $sub = $self->get_int_data_subcategory // 255;

    return    $cat == 2 || $cat == 3 || $cat > 4
           || { map { $_ => 1 } qw(0_14 0_20 0_30 0_40 1_20 1_30 1_31 4_1) }->{"${cat}_$sub"};
}

sub main {
    my (%opts, $rc, @msgs);

    # get and check options
    $Getopt::Std::STANDARD_HELP_VERSION = 1;
    usage '' unless getopts('vF:DNH:', \%opts);
    if ($opts{v}) {
        print "metafsrc2raw.pl version $VERSION\n" or exit 1;
        return 1;
    }
    usage "ERROR: input format must be specified with -F<FORMAT>.\n"
        unless exists $opts{F};
    $opts{H} = 0 unless exists $opts{H} && $opts{H} =~ /^[0-5]$/;

    if ($opts{F} eq 'bufr') {
        my ($msg, $data, $desc, $assoc);

        for my $file_name (@ARGV) {
            $rc = start_bufr $opts{H}, $file_name, \&do_filter;
            usage "$rc\n" if $rc;

            while (1) {
                ($msg, $data, $desc, $assoc) = next_bufr(1);
                last
                    unless defined $msg;

                if (substr($msg, 0, 6) eq 'ERROR:') {
                    print {*STDERR} "ERROR: $0:" . substr($msg, 6);
                    next;
                }

                if (!print $msg, "\n") {
                    print {*STDERR} "ERROR: $0: print failed: $!\n"
                        or exit 1;
                    next_bufr();
                    return 0;
                }
            }
            next_bufr();
        }

        return 1;
    }

    $rc = start_file($opts{F}, $opts{N}, $opts{H});
    usage "$rc\n" if $rc;
    @msgs = ();
    binmode STDIN, ':bytes'
        or print {*STDERR} "ERROR: $0: binmode STDIN: $!\n";
    while (1) {
        $_ = <>;
        for (next_line($_)) {
            if ($opts{D}) {
                my $line = $_;
                next if grep { $_ eq $line } @msgs;
                push @msgs, $_;
            }
            if (!print "$_\n") {
                print {*STDERR} "ERROR: $0: print failed: $!\n" or exit 1;
                return 0;
            }
        }
        last unless defined $_;
    }

    return 1;
}

# don't run if loaded as a library
return 1
    if caller;

exit !main;
